/*********************************************************************************\
*     File:                  A L G . C P P                                        *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * alg.cpp - image manipulation algorithms (Blur, etc.)
 *
 *  Contains:
 *         void DoAlg(int algnum);
 */

#include "ImgBasic.h"

/*
#ifndef FOO
  #define FOO
#endif
*/

#undef TIMING_TEST
//#define TIMING_TEST

static void rotXfer(int x,int y, double *rx,double *ry, double cx,double cy, double rad);
static void add2bb(int *, int *, int *, int *, int, int);



/************************/
void ImgBasic::DoAlg(int anum)
{
  /* called with a value from the algMB button.  Executes the specified
     algorithm */

  WaitCursor();

  switch (anum) {
  case ALG_NONE:      RestorePic();   break;
  case ALG_BLUR:      Blur();         break;
  case ALG_SHARPEN:   Sharpen();      break;
  case ALG_EDGE:      SobelGradient();break;
  case ALG_TINF:      TinFoil();      break;
  case ALG_OIL:       OilPaint();     break;
  case ALG_BLEND:     Blend();        break;
  case ALG_ROTATE:    FineRotate(0);  break;
  case ALG_ROTATECLR: FineRotate(1);  break;
  case ALG_PIXEL:     Pixelize();     break;
  case ALG_SPREAD:    Spread();       break;
  case ALG_MEDIAN:    MedianFilter(); break;
  case ALG_RESIZE:    ResizePic();    break;
  case ALG_CROP:      CropPic();      break;
  case ALG_GSBLUR:    GaussianBlur(); break;
  case ALG_GSNOISE:   GaussianNoise();break;
  case ALG_FASTMRES:
  case ALG_MULTIRES:  Multires(anum); break;
  case ALG_CANYMRES:  CannyMRes(0);   break;
  case ALG_CANYMRES4: CannyMRes(1);   break;
  case ALG_GRAD_ROB:  GradientRoberts();break;
  case ALG_CANNY:     AlgCanny();     break;
  case ALG_SATURATE:  Saturate();     break;
  case ALG_GAMMA:     Gamma();        break;
  }

  ArrowCursor();
}

/************************/
void ImgBasic::Blur()
{
  /* runs a n*n convolution mask (all 1's) over 'pic',
     producing a 24-bit version.  Then calls 24to8 to generate a new 8-bit
     image, and installs it. 

     Note that 'n' must be odd for things to work properly */

  byte        *thepic, *tmpPic;
  int          i, n;
  static char  buf[64] = { '3', '\0' };
  
  i = InputDialog("Difuminar(blur):\nIntroduce el tamao de la mascara"
    " (ej. 3, 5, 7, ...)", buf, 64, "0123456789", 1);
  if (i==1 || strlen(buf)==0) return;
  n = atoi(buf);

  if (n < 1 ) {
    Error("Error:  El valor debe ser positivo");
    return;
  }

  WaitCursor();

  SetISTR(ISTR_INFO, "Blurring with %dx%d convolution mask...", n,n);

  if (mono) { // GREYSCALE IMAGE
    if (startGreyAlg(&thepic, &tmpPic)) return;
    
    doBlurConvolvGrey(thepic, pWIDE,pHIGH, tmpPic, n);

    endGreyAlg(thepic, tmpPic);

  } else { //    COLOR IMAGE
    if (start24bitAlg(&thepic, &tmpPic)) return;
    
    doBlurConvolv24(thepic, pWIDE,pHIGH, tmpPic, n);

    end24bitAlg(thepic, tmpPic);
  }
}



/************************/
void ImgBasic::Sharpen()
{
  /* runs an edge-enhancment algorithm */

  byte        *pic24, *tmpPic;
  int          i, sx,sy,sw,sh, n;
  static char  buf[64] = { '7', '5', '\0' };
  
  i = InputDialog("Aumentar contrantes:\nIntroduce el factor de aumento(0-99%)",
                   buf, 64, "0123456789", 1);
  if (i==1 || strlen(buf)==0) return;
  n = atoi(buf);

  if (n>=100) {
    Error("Error:  El valor introducido debe ser menor de 100%.");
    return;
  }

  WaitCursor();

  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,pWIDE,pHIGH);

  SetISTR(ISTR_INFO, "Sharpening %s by factor of %d%%...",
	  (HaveSelection() ? "selection" : "image"), n);

  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic, 
          (size_t) (pWIDE*pHIGH*3));
    
  doSharpConvolv(pic24, pWIDE,pHIGH,
                 tmpPic, sx,sy,sw,sh, n);

  end24bitAlg(pic24, tmpPic);
}


/************************/
void ImgBasic::TinFoil()
{
  byte *pic24, *tmpPic, *tp, *tlp;
  char *str;
  int  i, j, v, sx,sy,sw,sh;
  
  WaitCursor();
  
  str = "Doing cheesy embossing effect...";
  SetISTR(ISTR_INFO, str);
  
  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,pWIDE,pHIGH);
  
  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic, 
    (size_t) (pWIDE*pHIGH*3));

  /* fill selected area of tmpPic with gray128 */
  for (i=sy; i<sy+sh; i++) {
    tp = tlp = tmpPic + (i*pWIDE + sx) * 3;
    for (j=sx; j<sx+sw; j++) {
      *tp++ = 128;  *tp++ = 128;  *tp++ = 128;
    }
  }
  
  doAngleConvolv(pic24, pWIDE, pHIGH,
                 tmpPic, sx,sy,sw,sh);
  
  /* mono-ify selected area of tmpPic */
  for (i=sy; i<sy+sh; i++) {
    tp = tlp = tmpPic + (i*pWIDE + sx) * 3;
    for (j=sx; j<sx+sw; j++, tp+=3) {
      v = MONO(tp[0], tp[1], tp[2]);
      RANGE(v,0,255);
      tp[0] = tp[1] = tp[2] = (byte) v;
    }
  }
    
  end24bitAlg(pic24, tmpPic);
}  


/************************/
void ImgBasic::OilPaint()
{
  byte *pic24, *tmpPic;
  int   sx,sy,sw,sh;

  WaitCursor();

  SetISTR(ISTR_INFO, "Doing oilpaint effect...");

  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,pWIDE,
                            pHIGH);
  
  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic, 
          (size_t) (pWIDE*pHIGH*3));

  doOilPaint(pic24, pWIDE, pHIGH, tmpPic, sx,sy,sw,sh, 3);

  end24bitAlg(pic24, tmpPic);
}



/************************/
void ImgBasic::Blend()
{
  byte *pic24, *tmpPic;
  int   sx,sy,sw,sh;

  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,
                  pWIDE,pHIGH);
  
  WaitCursor();

  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic,
    (size_t) (pWIDE*pHIGH*3));

  doBlend(pic24, pWIDE, pHIGH, tmpPic, sx,sy,sw,sh);

  end24bitAlg(pic24, tmpPic);
}


/************************/
void ImgBasic::FineRotate(int clr)
{
  byte        *pic24, *tmpPic;
  int          i,sx,sy,sw,sh;
  double       rotval;
  char         txt[128];
  static char  buf[64] = { '\0' };

  sprintf(txt, "Rotar (%s):\n\nIntroduce el angulo, en grados:  (>0 = CCW)",
	  (clr ? "Clear" : "Copy"));

  i = InputDialog(txt, buf, 64, "0123456789.-", 1);
  if (i==1 || strlen(buf)==0) return;
  rotval = atof(buf);

  if (rotval == 0.0) return;
  

  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,
                            pWIDE,pHIGH);
  
  WaitCursor();

  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic,
    (size_t) (pWIDE*pHIGH*3));

  doRotate(pic24, pWIDE, pHIGH,
           tmpPic, sx,sy,sw,sh, rotval, clr);

  end24bitAlg(pic24, tmpPic);
}


/************************/
void ImgBasic::Pixelize()
{
  byte        *pic24, *tmpPic;
  int          i,sx,sy,sw,sh, pixX,pixY,err;
  static char  buf[64] = { '4', '\0' };

  i = InputDialog( "Cuadricular:\n\nIntroudce el tamao del pixel:"
    "\n(ej. '3', '5x8')", buf, 64, "0123456789x", 1);
  if (i==1 || strlen(buf)==0) return;

  pixX = pixY = err = 0;

  if (index(buf, 'x')) {
    if (sscanf(buf, "%d x %d", &pixX, &pixY) != 2) err++;
  }
  else {
    if (sscanf(buf, "%d", &pixX) != 1) err++;
    pixY = pixX;
  }

  if (pixX<1 || pixY<1 || err) {
    Error("Error:  La cadena introducida no es vlida.");
    return;
  }

  
  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,
                   pWIDE,pHIGH);
  
  WaitCursor();

  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic, 
    (size_t) (pWIDE*pHIGH*3));

  doPixel(pic24, pWIDE, pHIGH,
          tmpPic, sx,sy,sw,sh, pixX, pixY);

  end24bitAlg(pic24, tmpPic);
}



/************************/
void ImgBasic::Spread()
{
  byte        *pic24, *tmpPic;
  int          i,sx,sy,sw,sh, pixX,pixY,err;
  static char  buf[64] = { '5', '\0' };

  i = InputDialog("Dispersin(Spread):\n\nIntroduce el factor de dispersin"
    "(o factores x,y ):  (ej. '10', '1x5')",  buf, 64, "0123456789x", 1);
  if (i==1 || strlen(buf)==0) return;

  pixX = pixY = err = 0;

  if (index(buf, 'x')) {
    if (sscanf(buf, "%d x %d", &pixX, &pixY) != 2) err++;
    if (pixX<0 || pixY<0) err++;
  }
  else {
    if (sscanf(buf, "%d", &pixX) != 1) err++;
    pixX = -pixX;
  }

  if (!pixX && !pixY) return;     /* 0x0 has no effect */

  if (err) {
    Error("Error:  La cadena introducida no es vlida");
    return;
  }

  
  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,
                            pWIDE,pHIGH);
  
  WaitCursor();

  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic,
          (size_t) (pWIDE*pHIGH*3));

  doSpread(pic24, pWIDE, pHIGH,
           tmpPic, sx,sy,sw,sh, pixX, pixY);

  end24bitAlg(pic24, tmpPic);
}



/************************/
void ImgBasic::MedianFilter()
{
  /* runs median filter algorithm (for n*n rect centered around each pixel,
     replace with median value */

  byte        *pic24, *tmpPic;
  int          i, sx,sy,sw,sh, n;
  static char  buf[64] = { '3', '\0' };

  i = InputDialog("DeSpeckle (filtro de la mediana):\nIntroduce el tamao de"
           " la mscara (ej. 3, 5, 7, ...)", buf, 64, "0123456789", 1);
  if (i==1 || strlen(buf)==0) return;
  n = atoi(buf);

  if (n < 1 || (n&1)!=1) {
    Error("Error:  El valor introducido debe ser impar y positivo.");
    return;
  }

  WaitCursor();

  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh, 0,0,
                pWIDE,pHIGH);

  SetISTR(ISTR_INFO, "DeSpeckling %s with %dx%d convolution mask...",
	  (HaveSelection() ? "selection" : "image"), n,n);

  if (start24bitAlg(&pic24, &tmpPic)) return;
  xvbcopy((char *) pic24, (char *) tmpPic, 
          (size_t) (pWIDE*pHIGH*3));
    
  doMedianFilter(pic24, pWIDE, pHIGH,
                 tmpPic, sx,sy,sw,sh, n);

  end24bitAlg(pic24, tmpPic);
}



/******************************************************************************/
void ImgBasic::doBlurConvolv24( byte *pic24, int columnas, int filas,
                                byte *results,int n)
{
  /*       U L T R A     F A S T !!!  (but requires extra memory)     */
  /* convolves with an n*n array, consisting of only 1's.  
     Operates on rectangular region 'selx,sely,selw,selh' (in pic coords)
     Region is guaranteed to be completely within pic boundaries
     'n' can be odd or even */

  int *aux,*aux2,*OutY;     /* Auxiliar image for separating components */
  int indice1,red,green,BytesPerRow=columnas*3;
  register int indice2,blue;
  byte *OutX;
  
  /* Allocate memory */
  aux = (int*) malloc( filas*BytesPerRow*sizeof(int) );
  if (!aux) {
    Error("Fallo al requerir memoria");
    return;
  }

  /* BLUR X (calculated with a "mirror image" beyond the border )*/
  for ( indice1=filas*BytesPerRow ; indice1 ;) {
     
     indice2=indice1;
     indice1-=BytesPerRow;

     /* starting value */
     OutX=pic24+indice2-3;
     //blue
     indice2--;
     blue=((int) OutX[2])*n;
     aux[indice2] = blue;
     //green
     indice2--;
     green=((int) OutX[1])*n;
     aux[indice2] = green;
     //red
     indice2--;
     red=((int) OutX[0])*n;
     aux[indice2] = red;
     
     /* right border */
     for ( ; indice2 >=(indice1+BytesPerRow-3*n); ) {
       //blue
       indice2--;
       blue -= OutX[2];
       blue += pic24[indice2];
       aux[indice2] = blue;
       //green
       indice2--;
       green -= OutX[1];
       green += pic24[indice2];
       aux[indice2] = green;
       //red
       indice2--;
       red -= OutX[0];
       red += pic24[indice2];
       aux[indice2] = red;
     }
     
     /* middle */
     OutX=pic24+3*n;
     for ( ; indice2>indice1 ;) {
       //blue
       indice2--;
       blue -= OutX[indice2];
       blue += pic24[indice2];
       aux[indice2] = blue;
       //green
       indice2--;
       green -= OutX[indice2];
       green += pic24[indice2];
       aux[indice2] = green;
       //red
       indice2--;
       red -= OutX[indice2];
       red += pic24[indice2];
       aux[indice2] = red;
     }
  }

     
  /* Allocate memory */
  aux2 = (int*) malloc( (n+1)*sizeof(int) );
  if (!aux2) {
    free(aux);
    Error("Fallo al requerir memoria");
    return;
  }

  /* BLUR Y (calculated with a "mirror image" beyond the border )*/
  /* Don't worry about the component, they don't mix */
  for ( indice1=BytesPerRow ; indice1 ; ) {
     
     indice1--;
          
     /* bottom border. fil with buffer */
     /* store in buffer blured bottom*/
     OutY=aux+BytesPerRow*(filas-1)+indice1;
     blue=OutY[0]*n;
     for(indice2=0;indice2<n;indice2++) {
       blue -= OutY[0];
       blue += OutY[-indice2*BytesPerRow];; 
       aux2[indice2]=blue;
     }
        
     /* middle */
     indice2=BytesPerRow*(filas-n)+indice1; 
     OutY=aux+n*BytesPerRow;
     for ( ; indice2>=BytesPerRow ;) {
       indice2 -= BytesPerRow;
       blue -= OutY[indice2];
       blue += aux[indice2];
       OutY[indice2]=blue;
     }

     /* top border */
     OutY=aux+indice1;
     for(indice2=n;indice2>0;) { 
       indice2--;
       blue -= OutY[indice2*BytesPerRow];
       blue += OutY[0];
       OutY[indice2*BytesPerRow]=blue;
     }

     /* Restore correct bottom 
     OutY=aux+BytesPerRow*(filas-1)+indice1;
     blue=OutY[0]*n;
     for(indice2=0;indice2<n;indice2++) 
       OutY[-indice2*BytesPerRow] = aux2[indice2];     
     */
  }

  /* Scale factor */
  blue = n*n; //factor => number of pixels in window (reuse n as n*n)
  /* convert float to byte*/
  for(indice2=filas*BytesPerRow ; indice2;){
    indice2--;
    /* sum of window pixels / num of pixels */
    //results[indice2]=  (aux[indice2]>0) ? 
    //                   ((unsigned char)(aux[indice2]/blue)): 0;
    results[indice2]= (unsigned char)(aux[indice2]/blue);
  }

  /* FREE MEMORY */
  free(aux);
  free(aux2);
}


/******************************************************************************/
void ImgBasic::doBlurConvolvGrey( byte *picGrey, int columnas, int filas,
                                   byte *results,int n)
{
  /*       U L T R A     F A S T !!!  (but requires extra memory)     */
  /* convolves with an n*n array, consisting of only 1's.  
     Operates on rectangular region 'selx,sely,selw,selh' (in pic coords)
     Region is guaranteed to be completely within pic boundaries
     'n' can be odd or even.
     2*columnas*filas sums, filas+columnas multiplications
  */

  int *aux,*aux2,*OutY;      /* Auxiliar image for separating components */
  int indice1;
  register int valor1,indice2;
  byte *OutX;
  
  /* Allocate memory */
  aux = (int*) malloc( (long)filas*(long)columnas*sizeof(int) );
  if (!aux) {
    Error("Fallo al requerir memoria");
    return;
  }

  /* BLUR X (calculated with a "mirror image" beyond the border )*/
  for ( indice1=filas*columnas ; indice1 ;) {
     
     indice2=indice1-1;
     indice1-=columnas;

     /* starting value */
     OutX=picGrey+indice2;
     valor1=((int) *OutX)*n;
     aux[indice2] = valor1;
     
     /* right border */
     for ( ; indice2 >=(indice1+columnas-n); ) {
       indice2--;
       valor1 -= *OutX;
       valor1 += picGrey[indice2];
       aux[indice2]=valor1;
     }
     
     /* middle */
     OutX=picGrey+n;
     for ( ; indice2>indice1 ;) {
       indice2--;
       valor1 -= OutX[indice2];
       valor1 += picGrey[indice2];
       aux[indice2]=valor1;
     }
  }

  /* Allocate memory */
  aux2 = (int*) malloc( (n+1)*sizeof(int) );
  if (!aux2) {
    free(aux);
    Error("Fallo al requerir memoria");
    return;
  }

  /* BLUR Y (calculated with a "mirror image" beyond the border )*/
  for ( indice1=columnas ; indice1 ; ) {
     
     indice1--;
     
     /* bottom border. fil with buffer */
     /* store in buffer blured bottom*/
     OutY=aux+columnas*(filas-1)+indice1;
     valor1=OutY[0]*n;
     for(indice2=1;indice2<n;indice2++) {
       valor1 += OutY[-indice2*columnas]-OutY[0];
       aux2[indice2]=valor1;
     }
     
     /* middle */
     indice2=columnas*(filas-n)+indice1; 
     OutY=aux+n*columnas;
     for ( ; indice2 >= columnas ;) {
       indice2 -= columnas;
       valor1 += aux[indice2]-OutY[indice2];
       OutY[indice2]=valor1;
     }

     /* top border */
     OutY=aux+indice1;
     for(indice2=n;indice2;) { 
       indice2--;
       valor1 += OutY[0] - OutY[indice2*columnas];
       OutY[indice2*columnas]=valor1;
     }

     /* Restore correct bottom
     OutY=aux+columnas*(filas-1)+indice1;
     valor1=OutY[0]*n;
     for(indice2=0;indice2<n;indice2++) {
       OutY[-indice2*columnas] = aux2[indice2];
     }
     */
  }

 
  /* Scale factor */
  valor1 = n*n; //factor => number of pixels in window (reuse n as n*n)
  /* convert float to byte*/
  for(indice2=filas*columnas ; indice2;){
    indice2--;
    /* sum of window pixels / num of pixels */
    //results[indice2]= (aux[indice2]>0) ? 
    //                  ((unsigned char)(aux[indice2]/valor1)): 0;
    results[indice2]= (unsigned char)(aux[indice2]/valor1);
  }

  /* FREE MEMORY */
  free(aux2);
}


/******************************************************************************/
void ImgBasic::doSharpConvolv(byte *pic24, int w, int h, byte *results, int selx,
                               int sely,int selw,int selh,int  n)
{
  byte  *p24;
  int    rv, gv, bv;
  byte  *rp;
  int    i,x,y;
  double fact, ifact, hue,sat,val, vsum;
  double *linem1, *line0, *linep1, *tmpptr;

  fact  = n / 100.0;
  ifact = 1.0 - fact;

  if (selw<3 || selh<3) return;  /* too small */


  linem1 = (double *) malloc(selw * sizeof(double));
  line0  = (double *) malloc(selw * sizeof(double));
  linep1 = (double *) malloc(selw * sizeof(double));

  if (!linem1 || !line0 || !linep1) {
    Error("Malloc() error en doSharpConvov().");
    if (linem1) free(linem1);
    if (line0)  free(line0);
    if (linep1) free(linep1);
    return;
  }


  /* load up line arrays */
  p24 = pic24 + (((sely+1)-1)*w + selx) * 3;
  for (x=selx, tmpptr=line0; x<selx+selw; x++, p24+=3, tmpptr++) {
    rgb2hsv((int) p24[0], (int) p24[1], (int) p24[2], &hue,&sat,&val);
    *tmpptr = val;
  }

  p24 = pic24 + (((sely+1)+0)*w + selx) * 3;
  for (x=selx, tmpptr=linep1; x<selx+selw; x++, p24+=3, tmpptr++) {
    rgb2hsv((int) p24[0], (int) p24[1], (int) p24[2], &hue,&sat,&val);
    *tmpptr = val;
  }


  for (y=sely+1; y<(sely+selh)-1; y++) {
    ProgressMeter(sely+1, (sely+selh)-2, y, "Sharpen");
    if ((y & 15) == 0) WaitCursor();
    
    tmpptr = linem1;   linem1 = line0;   line0 = linep1;   linep1 = tmpptr;

    /* get next line */
    p24 = pic24 + ((y+1)*w + selx) * 3;
    for (x=selx, tmpptr=linep1; x<selx+selw; x++, p24+=3, tmpptr++) {
      rgb2hsv((int) p24[0], (int) p24[1], (int) p24[2], &hue,&sat,&val);
      *tmpptr = val;
    }

    p24 = pic24   + (y*w + selx+1) * 3;
    rp  = results + (y*w + selx+1) * 3;

    for (x=selx+1; x<selx+selw-1; x++, p24+=3) {
      vsum = 0.0;  i = x-selx;
      vsum = linem1[i-1] + linem1[i] + linem1[i+1] +
	     line0 [i-1] + line0 [i] + line0 [i+1] +
	     linep1[i-1] + linep1[i] + linep1[i+1];
      
      rgb2hsv((int) p24[0], (int) p24[1], (int) p24[2], &hue, &sat, &val);

      val = ((val - (fact * vsum) / 9) / ifact);
      RANGE(val, 0.0, 1.0);
      hsv2rgb(hue,sat,val, &rv, &gv, &bv);

      RANGE(rv,0,255);
      RANGE(gv,0,255);
      RANGE(bv,0,255);

      *rp++ = (byte) rv;
      *rp++ = (byte) gv;
      *rp++ = (byte) bv;
    }
  }
  free(linem1);  free(line0);  free(linep1);
}



/************************/
void ImgBasic::doAngleConvolv(byte *pic24, int w, int h, byte *results, int selx,
                           int sely, int selw, int selh)
{

  /* convolves with edge detection mask, at 45 degrees to horizontal.
     
     The mask is (hard coded):

             -2 -1 0
             -1  0 1
              0  1 2
          
     Also, only does pixels in which the masks fit fully onto the picture
     (no pesky boundary conditionals)

     Adds value of rsum,gsum,bsum to results pic */

  register byte *p24;
  register int   bperlin,rsum,gsum,bsum;
  byte          *rp;
  int            x,y;


  bperlin = w * 3;

  for (y=sely+1; y<sely+selh-1; y++) {
    ProgressMeter(sely+1, (sely+selh-1)-1, y, "Convolve");
    if ((y & 63) == 0) WaitCursor();

    rp  = results + (y*w + selx+1)*3;
    p24 = pic24   + (y*w + selx+1)*3;

    for (x=selx+1; x<selx+selw-1; x++, p24+=3) {

      /* compute weighted average for *p24 (pic24[x,y]) */
      rsum = gsum = bsum = 0;

      rsum -= (p24[-bperlin-3] * 2);   /* top left */
      gsum -= (p24[-bperlin-2] * 2);
      bsum -= (p24[-bperlin-1] * 2);

      rsum -= p24[-bperlin];           /* top mid */
      gsum -= p24[-bperlin+1];
      bsum -= p24[-bperlin+2];

      rsum -= p24[-3];                 /* mid left */
      gsum -= p24[-2];
      bsum -= p24[-1];

      rsum += p24[3];                  /* mid right */
      gsum += p24[4];
      bsum += p24[5];

      rsum += p24[bperlin];            /* bottom mid */
      gsum += p24[bperlin+1];
      bsum += p24[bperlin+2];

      rsum += (p24[bperlin+3] * 2);    /* bottom right */
      gsum += (p24[bperlin+4] * 2);
      bsum += (p24[bperlin+5] * 2);

      rsum = rsum / 8;
      gsum = gsum / 8;
      bsum = bsum / 8;

      rsum += rp[0];  RANGE(rsum,0,255);
      gsum += rp[1];  RANGE(gsum,0,255);
      bsum += rp[2];  RANGE(bsum,0,255);

      *rp++ = (byte) rsum;
      *rp++ = (byte) gsum;
      *rp++ = (byte) bsum;
    }
  }
}




/************************/
void ImgBasic::doOilPaint(byte *pic24, int w, int h, byte *results, int selx,int sely,
                       int selw,int selh, int n)                                 
{

  /* does an 'oil transfer', as described in the book "Beyond Photography",
     by Holzmann, chapter 4, photo 7.  It is a sort of localized smearing.

     The following algorithm (but no actual code) was borrowed from
     'pgmoil.c', written by Wilson Bent (whb@hoh-2.att.com),
     and distributed as part of the PBMPLUS package.

     for each pixel in the image (assume, for a second, a grayscale image),
     compute a histogram of the n*n rectangle centered on the pixel.
     replace the pixel with the color that had the greatest # of hits in
     the histogram.  Note that 'n' should be odd. 

     I've modified the algorithm to do the *right* thing for RGB images.
     (jhb, 6/94)  */


  register byte *pp;
  register int   bperlin;
  byte          *rp, *p24, *plin;
  int            i,j,k,x,y,n2,col,cnt,maxcnt;
  int           *nnrect;

  if (n & 1) n++;   /* n must be odd */

  bperlin = w * 3;
  n2 = n/2;

  /* nnrect[] is an n*n array of ints, with '-1' meaning 'outside the region'
     otherwise they'll have 24-bit RGB values */
  
  nnrect = (int *) malloc(n * n * sizeof(int));
  if (!nnrect) FatalError("can't malloc nnrect[] in doOilPaint()\n");

  for (y=sely; y<sely+selh; y++) {
    if ((y & 15) == 0) WaitCursor();
    ProgressMeter(sely, (sely+selh)-1, y, "Oil Paint");

    p24 = pic24 + ((y-n2)*w + selx-n2)*3;   /* pts to top-left of mask */
    rp  = results + (y*w + selx)*3;
    
    for (x=selx; x<selx+selw; x++) {
      /* fill 'nnrect' with valid pixels from n*n region centered round x,y */
      pp = plin = p24;
      for (i=y-n2, k=0; i<y+n2; i++) {
	for (j=x-n2; j<x+n2; j++, k++, pp+=3) {
	  if (i>=sely && i<sely+selh && j>=selx && j<selx+selw) { 
	    nnrect[k] = (((int) pp[0])<<16) | (((int) pp[1])<<8) | pp[2];
	  }
	  else nnrect[k] = -1;
	}
	plin += bperlin;  pp = plin;
      }

      
      /* find 'most popular color' in nnrect, not counting '-1' */
      maxcnt = cnt = col = 0;
      for (i=0; i<n*n; i++) {
	if (nnrect[i] != -1) {
	  cnt = 1;
	  for (j=i+1; j<n*n; j++) { /* count it */
	    if (nnrect[i] == nnrect[j]) cnt++;
	  }
	  if (cnt>maxcnt) { col = nnrect[i];  maxcnt = cnt; }
	}
      }

      *rp++ = (byte) ((col>>16) & 0xff);
      *rp++ = (byte) ((col>>8)  & 0xff);
      *rp++ = (byte) ((col)     & 0xff);

      p24 += 3;
    }
  }
  free(nnrect);
}


/************************/
void ImgBasic::doBlend(byte *pic24, int w, int h, byte *results,
                    int selx,int sely,int selw,int selh)
{
  /* 'blends' a rectangular region out of existence.  computes the average
     color of all the pixels on the edge of the rect, stores this in the
     center, and for each pixel in the rect, replaces it with a weighted
     average of the center color, and the color on the edge that intersects
     a line drawn from the center to the point */

  byte  *p24;
  int    i,x,y;
  int    cx,cy,cR,cG,cB;
  int    ex,ey,eR,eG,eB;
  int    dx,dy,r,g,b;
  double rf,gf,bf, slope,dslope, d,d1, ratio;

  if (selw<3 || selh<3) return;        /* too small to blend */

  /*** COMPUTE COLOR OF CENTER POINT ***/

  i = 0;  rf = gf = bf = 0.0;
  for (x=selx; x<selx+selw; x++) {
    p24 = pic24 + (sely*w + x) * 3;
    rf += (double) p24[0];  gf += (double) p24[1];  bf += (double) p24[2];
    i++;

    p24 = pic24 + ((sely+selh-1)*w + x) * 3;
    rf += (double) p24[0];  gf += (double) p24[1];  bf += (double) p24[2];
    i++;
  }
  for (y=sely; y<sely+selh; y++) {
    p24 = pic24 + (y*w + selx) * 3;
    rf += (double) p24[0];  gf += (double) p24[1];  bf += (double) p24[2];
    i++;
    
    p24 = pic24 + (y*w + (selx+selw-1)) * 3;
    rf += (double) p24[0];  gf += (double) p24[1];  bf += (double) p24[2];
    i++;
  }

  cR = (int) (rf / i);
  cG = (int) (gf / i);
  cB = (int) (bf / i);

  cx = selx + selw/2;  cy = sely + selh/2;

  dslope = ((double) selh) / selw;

  /* for each point on INTERIOR of region, do the thing */
  for (y=sely+1; y<sely+selh-1; y++) {
    if ((y & 15) == 0) WaitCursor();
    ProgressMeter(sely+1, (sely+selh-1)-1, y, "Blend");

    for (x=selx+1; x<selx+selw-1; x++) {

      /* compute edge intercept point ex,ey */
      dx = x - cx;  dy = y - cy;
      if (dx==0 && dy==0) { ex = selx;  ey = sely; }  /* don't care */
      else if (dx==0) {	ex = cx;  ey = (dy<0) ? sely : sely+selh-1; }
      else if (dy==0) {	ey = cy;  ex = (dx<0) ? selx : selx+selw-1; }
      else { 
	slope = ((double) dy) / dx;
	if (fabs(slope) > fabs(dslope)) {   /* y axis is major */
	  ey = (dy<0) ? sely : sely+selh-1;
	  ex = cx + ((ey-cy) * dx) / dy;
	}
	else {                              /* x axis is major */
	  ex = (dx<0) ? selx : selx+selw-1;
	  ey = cy + ((ex-cx) * dy) / dx;
	}
      }

      /* let's play it safe... */
      RANGE(ex, selx, selx+selw-1);
      RANGE(ey, sely, sely+selh-1);

      /* given (cx,cy), (x,y), and (ex,ey), compute d, d1 */
      d  = sqrt((double) (dx * dx) + (double) (dy * dy));
      d1 = sqrt((double) ((ex-cx) * (ex-cx)) + (double) ((ey-cy) * (ey-cy)));
      ratio = pow(d/d1, 2.0);

      /* fetch color of ex,ey */
      p24 = pic24 + (ey*w + ex) * 3;
      eR = p24[0];  eG=p24[1];  eB=p24[2];

      /* compute new color for x,y */
      if (dx==0 && dy==0) { r=cR;  g=cG;  b=cB; }
      else {
	r = cR + (int) ((eR-cR) * ratio);
	g = cG + (int) ((eG-cG) * ratio);
	b = cB + (int) ((eB-cB) * ratio);
	RANGE(r,0,255);
	RANGE(g,0,255);
	RANGE(b,0,255);
      }

      /* and stuff it... */
      p24 = results + (y*w + x) * 3;
      p24[0] = (byte) r;  p24[1] = (byte) g;  p24[2] = (byte) b;
    }
  }
}

  

/************************/
void ImgBasic::doRotate(byte *pic24, int w, int h, byte *results, int selx,int sely,
                     int selw,int selh, double rotval, int clear)
{
  /* rotates a rectangular region (selx,sely,selw,selh) of an image (pic24,w,h)
     by the amount specified in degrees (rotval), and stores the result in
     'results', which is also a w*h 24-bit image.  The rotated bits are
     clipped to fit in 'results'.  If 'clear', the (unrotated) rectangular
     region is cleared (in results) first.  
     sel[x,y,w,h] is guaranteed to be within image bounds */

  byte  *pp, *dp;
  int    i, j, x, y, ox,oy, ox1,oy1;
  int    rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4;
  int    rbx, rby, rbx1, rby1, rbw, rbh;
  double rotrad, xf,yf, xfrac,yfrac, px,py, apx, apy, cfx,cfy;

  if (selw<1 || selh<1) return;

  /*
   * cfx,cfy  -  center point of sel rectangle (double) 
   * rx1,ry1  -  top-left  of sel, rotated
   * rx2,ry2  -  bot-left  of sel, rotated
   * rx3,ry3  -  top-right of sel, rotated
   * rx4,ry4  -  bot-right of sel, rotated
   * rbx, rby, rbw, rbh  -  bounding box for rotated sel, clipped to pic
   */

  rotrad = rotval * M_PI / 180.0;

  /* compute corner points of 'sel' after rotation */
  cfx = selx + (selw-1)/2.0;  cfy = sely + (selh-1)/2.0;

  rotXfer(selx,        sely,      &xf, &yf, cfx, cfy, rotrad);
  rx1 = (int) (xf + ((xf<0) ? -0.5 : 0.5));
  ry1 = (int) (yf + ((yf<0) ? -0.5 : 0.5));

  rotXfer(selx,        sely+selh, &xf, &yf, cfx, cfy, rotrad);
  rx2 = (int) (xf + ((xf<0) ? -0.5 : 0.5));
  ry2 = (int) (yf + ((yf<0) ? -0.5 : 0.5));

  rotXfer(selx+selw,   sely,      &xf, &yf, cfx, cfy, rotrad);
  rx3 = (int) (xf + ((xf<0) ? -0.5 : 0.5));
  ry3 = (int) (yf + ((yf<0) ? -0.5 : 0.5));

  rotXfer(selx+selw,   sely+selh, &xf, &yf, cfx, cfy, rotrad);
  rx4 = (int) (xf + ((xf<0) ? -0.5 : 0.5));
  ry4 = (int) (yf + ((yf<0) ? -0.5 : 0.5));

  /* compute bounding box for rotated rect */

  rbx = rbx1 = rx1;  rby = rby1 = ry1;
  add2bb(&rbx, &rby, &rbx1, &rby1, rx2,ry2);
  add2bb(&rbx, &rby, &rbx1, &rby1, rx3,ry3);
  add2bb(&rbx, &rby, &rbx1, &rby1, rx4,ry4);
  rbw = rbx1 - rbx;  rbh = rby1 - rby;

  /* make it a *little* bigger, just to be safe... */
  rbx--;  rby--;  rbw+=2;  rbh+=2;
  CropRect2Rect(&rbx, &rby, &rbw, &rbh, 0,0,w,h);

  if (clear) {           /* clear original selection in results pic */
    for (i=sely; i<sely+selh; i++) {
      dp = results + (i*w + selx)*3;
      for (j=selx; j<selx+selw; j++) {
	*dp++ = 0;
	*dp++ = 0;
	*dp++ = 0;
      }
    }
  }


  /* now, for each pixel in rb[x,y,w,h], do the inverse rotation to see if
     it would be in the original unrotated selection rectangle.  if it *is*,
     compute and store an appropriate color, otherwise skip it */
 
  for (y=rby; y<rby+rbh; y++) {
    dp = results + (y * w + rbx) * 3;

    if ((y & 15) == 0) WaitCursor();
    ProgressMeter(rby, rby+rbh-1, y, "Rotate");

    for (x=rbx; x<rbx+rbw; x++, dp+=3) {
      rotXfer(x,y, &xf, &yf, cfx, cfy, -rotrad);

      /* cheat a little... */
      if (xf < 0.0  &&  xf > -0.5) xf = 0.0;
      if (yf < 0.0  &&  yf > -0.5) yf = 0.0;

      ox = (int) floor(xf);   oy = (int) floor(yf);

      if (PTINRECT(ox,oy, selx,sely,selw,selh)) {
	int    p0r,p0g,p0b, p1r,p1g,p1b, p2r,p2g,p2b, p3r,p3g,p3b;
	int    rv,gv,bv;
	double rd,gd,bd, p0wgt, p1wgt, p2wgt, p3wgt;
	
	/* compute the color, same idea as in Smooth**().  The color
	   will be a linear combination of the colors of the center pixel,
	   its left-or-right neighbor, its top-or-bottom neighbor, and
	   its corner neighbor.  *which* neighbors are used are determined by
	   the position of the fractional part of xf,yf within the 1-unit
	   square of the pixel.  */

	/* compute px,py: fractional offset from center of pixel (x.5,y.5) */
	xfrac = xf - ox;        /* 0 - .9999 */
	yfrac = yf - oy;
	px = ((xfrac >= .5) ? (xfrac - .5) : (-.5 + xfrac));
	py = ((yfrac >= .5) ? (yfrac - .5) : (-.5 + yfrac));
	apx = fabs(px);  apy = fabs(py);

	/* get neighbor colors:  p0col, p1col, p2col, p3col */
	ox1 = ox + ((px < 0.0) ? -1 : 1);
	oy1 = oy + ((py < 0.0) ? -1 : 1);

	pp = pic24 + (oy * w + ox) * 3;
	p0r = pp[0];  p0g = pp[1];  p0b = pp[2];                   /* ctr */

	if (ox1 >= selx && ox1 < selx+selw) {
	  pp = pic24 + (oy * w + ox1) * 3;
	  p1r = pp[0];  p1g = pp[1];  p1b = pp[2];                 /* l/r */
	  p1wgt = apx * (1.0 - apy);
	}
	else { p1r=p1g=p1b=0;  p1wgt = 0.0; }

	if (oy1 >= sely && oy1 < sely+selh) {
	  pp = pic24 + (oy1 * w + ox) * 3;
	  p2r = pp[0];  p2g = pp[1];  p2b = pp[2];                 /* t/b */
	  p2wgt = apy * (1.0 - apx);
	}
	else { p2r=p2g=p2b=0;  p2wgt = 0.0; }

	if (ox1>=selx && ox1<selx+selw && oy1>=sely && oy1<sely+selh){
	  pp = pic24 + (oy1 * w + ox1) * 3;
	  p3r = pp[0];  p3g = pp[1];  p3b = pp[2];                 /* diag */
	  p3wgt = apx * apy;
	}
	else { p3r=p3g=p3b=0;  p3wgt = 0.0; }

	p1wgt = p1wgt * .7;        /* black art */
	p2wgt = p2wgt * .7;
	p3wgt = p3wgt * .7;

	p0wgt = 1.0 - (p1wgt + p2wgt + p3wgt);

	/* okay, compute and store resulting color */
	rd = p0r * p0wgt + p1r * p1wgt + p2r * p2wgt + p3r * p3wgt;
	gd = p0g * p0wgt + p1g * p1wgt + p2g * p2wgt + p3g * p3wgt;
	bd = p0b * p0wgt + p1b * p1wgt + p2b * p2wgt + p3b * p3wgt;

	rv = (int) (rd + 0.5);
	gv = (int) (gd + 0.5);
	bv = (int) (bd + 0.5);

	RANGE(rv,0,255);  RANGE(gv,0,255);  RANGE(bv,0,255);

#ifdef ROTATE_FOO
if (0)	{
  debug_printf("--------\n%3d,%3d in results maps to %6.2f,%6.2f in pic24\n",
	  x,y,xf,yf);
  debug_printf("ox,oy = %3d,%3d (%3d,%3d)  frac=%4.2f,%4.2f  px,py=%4.2f,%4.2f\n", ox,oy, ox1,oy1, xfrac,yfrac, px,py);
  debug_printf("Colors: 0:%02x%02x%02x 1:%02x%02x%02x 2:%02x%02x%02x 3:%02x%02x%02x\n", p0r,p0g,p0b, p1r,p1g,p1b, p2r,p2g,p2b, p3r,p3g,p3b);
  debug_printf("Weights: 0[%f]  1[%f]  2[%f]  3[%f] -> %3d,%3d,%3d\n",
	  p0wgt, p1wgt, p2wgt, p3wgt, rv, gv, bv);
}
#endif /* ROTATE_FOO */

	dp[0] = (byte) (rv&0xff);  
	dp[1] = (byte) (gv&0xff);  
	dp[2] = (byte) (bv&0xff);  
      }
    }
  }
}


/***********************************************/
static void add2bb(int *x1, int *y1, int *x2, int *y2, int x,int y)
{
  if (x < *x1) *x1 = x;
  if (x > *x2) *x2 = x;
  if (y < *y1) *y1 = y;
  if (y > *y2) *y2 = y;
}


/***********************************************/
static void rotXfer(int x,int y, double *rx,double *ry, double cx,double cy, double rad)
{
  /* take point x,y, rotate it 'rad' radians around cx,cy, return rx,ry */
  double d, xf, yf, ang;

  xf = x;  yf = y;

  d  = sqrt((xf-cx) * (xf-cx) + (yf-cy) * (yf-cy));
  if      ((xf-cx) != 0.0) {
    ang = atan((cy-yf) / (xf-cx));                 /* y-axis flip */
    if ((xf-cx) < 0) ang += M_PI;
  }

  else if ((yf-cy) > 0.0) ang = M_PI * 3.0 / 2.0;
  else                    ang = M_PI * 1.0 / 2.0;

  *rx = (double) cx + (d * cos(ang + rad));
  *ry = (double) cy - (d * sin(ang + rad));

#ifdef FOO
  fprintf(stderr,"rotXfer:  rotating (%4d,%4d) %7.2f degrees around",
	  x,y, rad*180.0 / M_PI);
  fprintf(stderr,"(%4d,%4d) -> %7.2f %7.2f  (d=%f ang=%f)\n", 
	  cx,cy, *rx,*ry, d, ang);
#endif
}
  


/************************/
void ImgBasic::doPixel(byte *pic24, int w, int h, byte *results,  int selx, int sely, int selw, int selh, int pixX, int pixY)
{
  /* runs 'pixelization' algorithm.  replaces each pixX-by-pixY region 
     (smaller on edges) with the average color within that region */
  
  byte  *pp;
  int    nwide, nhigh, i,j, x,y, x1,y1, stx,sty;
  int    nsum, rsum, gsum, bsum;
  
  /* center grid on selection */
  nwide = (selw + pixX-1) / pixX;
  nhigh = (selh + pixY-1) / pixY;
  
  stx = selx - (nwide*pixX - selw)/2;
  sty = sely - (nhigh*pixY - selh)/2;
  
  y = sty;
  for (i=0; i<nhigh; i++, y+=pixY) {
    ProgressMeter(0, nhigh-1, i, "Pixelize");
    
    x = stx;
    for (j=0; j<nwide; j++, x+=pixX) {
      
      /* COMPUTE AVERAGE COLOR FOR RECT:[x,y,pixX,pixY] */
      nsum = rsum = gsum = bsum = 0;
      for (y1=y; y1<y+pixY; y1++) {
	pp = pic24 + (y1 * w + x) * 3;
	for (x1=x; x1<x+pixX; x1++) {
	  if (PTINRECT(x1,y1, selx,sely,selw,selh)) {
	    nsum++;
	    rsum += *pp++;  gsum += *pp++;  bsum += *pp++;
	  }
	}
      }
      
      if (nsum>0) {   /* just to be safe... */
	rsum /= nsum;  gsum /= nsum;  bsum /= nsum;
	RANGE(rsum,0,255);  RANGE(gsum,0,255);  RANGE(bsum,0,255);
      }
      
      
      /* STORE color in rect:[x,y,pixX,pixY] */
      for (y1=y; y1<y+pixY; y1++) {
	//if (!j && (y1 & 255)==0) WaitCursor();
	
	pp = results + (y1 * w + x) * 3;
	for (x1=x; x1<x+pixX; x1++, pp+=3) {
	  if (PTINRECT(x1,y1, selx,sely,selw,selh)) {
	    pp[0] = (byte) rsum;  pp[1] = (byte) gsum;  pp[2] = (byte) bsum;
	  }
	}
      }
    }
  }

}

  

/************************/
void ImgBasic::doSpread( byte *pic24, int w, int h, byte *results, int selx,
                     int sely, int selw, int selh, int pixX, int pixY)
{
  /* runs 'spread' algorithm.  For each pixel in the image, swaps it with
     a random pixel near it.  Distances of random pixels are controlled
     by pixX,pixY.  If pixX<0, it is treated as a single 'distance' value
     (after being abs()'d). */

  /* assumes that initially 'results' is a copy of pic24.  Doesn't 
     even look at pic24 */
  
  byte  *pp, *dp, r,g,b;
  int    x,y, x1,y1,d;
  int    minx, maxx, miny, maxy, rdist;

  clock_t start = clock();
  srandom((unsigned int) start);


  for (y=sely; y<sely+selh; y++) {
    ProgressMeter(sely, sely+selh-1, y, "Spread");
    pp = results + (y * w + selx) * 3;
    for (x=selx; x<selx+selw; x++, pp+=3) {

      if (pixX < 0) {
	/* compute a random neighbor within radius 'd', cropped to 'sel' */

	d = abs(pixX);

	minx = x - d;	if (minx < selx)       minx = selx;
	maxx = x + d;	if (maxx >= selx+selw) maxx = selx+selw-1;
	x1 = minx + abs(random()) % ((maxx - minx) + 1);

	miny = y - d;	if (miny < sely)       miny = sely;
	maxy = y + d;	if (maxy >= sely+selh) maxy = sely+selh-1;

	rdist = d - abs(x1 - x);
	if (y - miny > rdist) miny = (y-rdist);
	if (maxy - y > rdist) maxy = (y+rdist);

	y1 = miny + abs(random()) % ((maxy - miny) + 1);
      }

      else {
	/* compute a neighbor within +/-pixX by +/-pixY, cropped to sel */

	minx = x - pixX;  if (minx < selx)       minx = selx;
	maxx = x + pixX;  if (maxx >= selx+selw) maxx = selx+selw-1;
	x1 = minx + abs(random()) % ((maxx - minx) + 1);

	miny = y - pixY;  if (miny < sely)       miny = sely;
	maxy = y + pixY;  if (maxy >= sely+selh) maxy = sely+selh-1;
	y1 = miny + abs(random()) % ((maxy - miny) + 1);
      }

      if (PTINRECT(x1,y1, selx,sely,selw,selh)) {  /* should always be true */
	dp = results + (y1 * w + x1) * 3;
	r     = pp[0];  g     = pp[1];  b     = pp[2];
	pp[0] = dp[0];  pp[1] = dp[1];  pp[2] = dp[2];
	dp[0] = r;      dp[1] = g;      dp[2] = b;
      }
    }
  }
}

  

/************************/
void ImgBasic::doMedianFilter(byte *pic24, int w, int h, byte *results, int selx,
                           int sely,int selw,int selh,int n)
{
  /* runs the median filter algorithm
     Operates on rectangular region 'selx,sely,selw,selh' (in pic coords)
     Region is guaranteed to be completely within pic boundaries
     'n' must be odd */

  register byte *p24;
  register int   rsum,gsum,bsum;
  byte          *rp;
  int            x,y,x1,y1,count,n2,nsq,c2;
  int           *rtab, *gtab, *btab;

  n2 = n/2;  nsq = n * n;

  rtab = (int *) malloc(nsq * sizeof(int));
  gtab = (int *) malloc(nsq * sizeof(int));
  btab = (int *) malloc(nsq * sizeof(int));
  if (!rtab || !gtab || !btab) FatalError("can't malloc in doMedianFilter!");

  for (y=sely; y<sely+selh; y++) {
    ProgressMeter(sely, (sely+selh)-1, y, "DeSpeckle");
    if ((y & 15) == 0) WaitCursor();

    p24 = pic24   + y*w*3 + selx*3;
    rp  = results + y*w*3 + selx*3;

    for (x=selx; x<selx+selw; x++) {

      rsum = gsum = bsum = 0;  count = 0;

      for (y1=y-n2; y1<=y+n2; y1++) {

	if (y1>=sely && y1<sely+selh) {
	  p24 = pic24 + y1*w*3 +(x-n2)*3; 

	  for (x1=x-n2; x1<=x+n2; x1++) {
	    if (x1>=selx && x1<selx+selw) {
	      rtab[count] = *p24++;
	      gtab[count] = *p24++;
	      btab[count] = *p24++;
	      count++;
	    }
	    else p24 += 3;
	  }
	}
      }


      /* now sort the rtab,gtab,btab arrays, (using shell sort) 
	 and pick the middle value.  doing it in-line, rather than 
	 as a function call (ie, 'qsort()') , for speed */
      {  
	int i,j,t,d;
	
	for (d=count/2;  d>0;  d=d/2) {
	  for (i=d; i<count; i++) {
	    for (j=i-d;  j>=0 && rtab[j]>rtab[j+d];  j-=d) {
	      t = rtab[j];  rtab[j] = rtab[j+d];  rtab[j+d] = t;
	    }

	    for (j=i-d;  j>=0 && gtab[j]>gtab[j+d];  j-=d) {
	      t = gtab[j];  gtab[j] = gtab[j+d];  gtab[j+d] = t;
	    }

	    for (j=i-d;  j>=0 && btab[j]>btab[j+d];  j-=d) {
	      t = btab[j];  btab[j] = btab[j+d];  btab[j+d] = t;
	    }
	  }
	}
      }
      
      c2 = count/2;
      *rp++ = (byte) ( (count&1) ? rtab[c2] : (rtab[c2] + rtab[c2-1])/2);
      *rp++ = (byte) ( (count&1) ? gtab[c2] : (gtab[c2] + gtab[c2-1])/2);
      *rp++ = (byte) ( (count&1) ? btab[c2] : (btab[c2] + btab[c2-1])/2);
    }
  }
  
  free(rtab);  free(gtab);  free(btab);

}


/* *************** Gamma factor stuff in GAM.CPP *************** */
//byte *GammifyPic24(byte *pic24, int wide, int high,
//                   byte RedRemap[/*256*/], byte GreenMap[/*256*/],
//                   byte BlueMap[/*256*/], int Saturation = 0,
//                   int ValueRemap[/*256*/]=NULL, int HueRemap[/*360*/] = NULL);
//void GenerateFSGamma(int x[],int y[],int n, byte fsgamcr[]);

void ImgBasic::Gamma() 
{
	Warning("Esta funcin est implementada pero hace falta"
          "realizar el interfaz grfico para introducir"
          "los parmetros de la funci GammifyPic24.\n");
}

void ImgBasic::Saturate() 
{
	//Warning("Esta funcin est sin implementar");
  printf("Algoritmo de saturacin de una imagen\n"
         "=====================================\n\n"
         "Para cada punto, su luminosidad se multiplica por un\n"
         "factor. Si la luminosidad supera el mximo correspondiente\n"
         "al color blanco, se satura.\n");
  printf("El valor 1.0 corresponde al factor que hace que justo\n"
         "ningn pixel de la imagen se sature. Un valor mayor da\n"
         "mayor luminosidad a la imagen, a costa de saturar las\n");
  printf("zonas ms claras. Un valor menor que 1.0 oscurece la\n"
         "imagen. Un valor cero o negativo aborta el algoritmo.\n"
         "Introduce el valor\n");
  float factor;
  if (scanf("%f",&factor) != 1) {
    Warning("No se ha podido leer el dato");
    return;
  }

  if (factor<=0.01 || factor >= 100) {
    Warning("El factor de escalado es muy grande o pequeo");
    return;
  }
  
  SetISTR(ISTR_INFO, "Saturando la imagen...");

  byte *thepic, *tmpPic;
  if (mono) { // GRAY IMAGE
      if (startGreyAlg(&thepic, &tmpPic)) return;
      xvbcopy((char *) thepic, (char *) tmpPic, (size_t) (pWIDE*pHIGH));
      saturePicGrey(tmpPic, pWIDE,0,0,pWIDE,pHIGH, (float) factor );
      endGreyAlg(thepic, tmpPic);
    
  } else {  // COLOR IMAGE
      if (start24bitAlg(&thepic, &tmpPic)) return;
      xvbcopy((char *) thepic, (char *) tmpPic, (size_t) (pWIDE*pHIGH*3));
      saturePic24(tmpPic,pWIDE,0,0,pWIDE,pHIGH, (float) factor );
      end24bitAlg(thepic, tmpPic);
  }
}


