/*********************************************************************************\
*     File:               B A S I C . C P P                                       *
*                                                                                 *
*     Standard   C++                                                              *
\*********************************************************************************/


// Basic manipulations for class ImgBasic

#include "ImgBasic.h"

#define LSBFirst 0
#define MSBFirst 1

void ImgBasic::InvertPic( byte *pic, int picType, u_int  wp, u_int  hp,
                          int mono, byte *rMap, byte *gMap, byte *bMap,
                          int numcols )
{
  int i;
  byte *ptr; 
  if (picType==PIC24) /* FULLCOLOR image*/
    for (i=wp*hp*3, ptr=pic; i; i--, ptr++)
      *ptr =~*ptr;
  else if (/*picType==PIC8 &&*/ mono && numcols>2) /* Direct Grayscale image */
    for (i=wp*hp, ptr=pic; i; i--, ptr++)
      *ptr = ~*ptr;
  else if (/*picType==PIC8 &&*/ mono && numcols<=2) /* Monocrome image */
    for (i=wp*hp, ptr=pic; i; i--, ptr++) *ptr ^= 1;
  else /* Mapped Grayscale or Mapped Color */ {
    if (bMap) 
      for (i=numcols, ptr=bMap; i ;i--, ptr++)  *ptr = ~*ptr;
    if (gMap) 
      for (i=numcols, ptr=gMap; i ;i--, ptr++)  *ptr = ~*ptr;
    if (rMap) 
      for (i=numcols, ptr=rMap; i ;i--, ptr++)  *ptr = ~*ptr;
  }
}

/************************/
void ImgBasic::RotatePic(int dir, byte *pic, int ptype, u_int  *wp, u_int  *hp )
{
  /* rotates a w*h array of bytes 90 deg clockwise (dir=0) 
     or counter-clockwise (dir != 0).  swaps w and h */

  byte        *pic1, *pix1, *pix;
  int          i,j,bperpix;
  unsigned int w,h;

  bperpix = (ptype == PIC8) ? 1 : 3;

  w = *wp;  h = *hp;  
  pix1 = pic1 = (byte *) malloc((size_t) (w*h*bperpix));
  if (!pic1) FatalError("Not enough memory to rotate!");

  /* do the rotation */
  if (dir==0) {
    for (i=0; i<w; i++) {       /* CW */
      if (bperpix == 1) {
	for (j=h-1, pix=pic+(h-1)*w + i;  j>=0;  j--, pix1++, pix-=w) 
	  *pix1 = *pix;
      }
      else {
	int bperlin = w*bperpix;
	int k;
	
	for (j=h-1, pix=pic+(h-1)*w*bperpix + i*bperpix;  
	     j>=0;  j--, pix -= bperlin) 
	  for (k=0; k<bperpix; k++) *pix1++ = pix[k];
      }
    }
  }
  else {
    for (i=w-1; i>=0; i--) {    /* CCW */
      if (bperpix == 1) {
	for (j=0, pix=pic+i; j<h; j++, pix1++, pix+=w) 
	  *pix1 = *pix;
      }
      else {
	int k;
	int bperlin = w*bperpix;
	
	for (j=0, pix=pic+i*bperpix; j<h; j++, pix+=bperlin) 
	  for (k=0; k<bperpix; k++) *pix1++ = pix[k];
      }
    }
  }
  
  
  /* copy the rotated buffer into the original buffer */
  xvbcopy((char *) pic1, (char *) pic, (size_t) (w*h*bperpix));
  
  free(pic1);
  
  /* swap w and h */
  *wp = h;  *hp = w;
}


/************************/
void ImgBasic::FlipPic( int dir, byte *pic, int picType, u_int w, u_int h)
{
  /* flips a w*h array of bytes horizontally (dir=0) or vertically (dir!=0) */
  
  byte *plin;
  int   i,j,k,l,bperpix,bperlin;
  
  bperpix = (picType == PIC8) ? 1 : 3;
  bperlin = w * bperpix;
  
  if (dir==0) {                /* horizontal flip */
    byte *leftp, *rightp;
    
    for (i=0; i<h; i++) {
      plin   = pic + i*bperlin;
      leftp  = plin;
      rightp = plin + (w-1)*bperpix;
      
      for (j=0; j<w/2; j++, rightp -= (2*bperpix)) {
	for (l=0; l<bperpix; l++, leftp++, rightp++) {
	  k = *leftp;  *leftp = *rightp;  *rightp = k;
	}
      }
    }
  }
  
  else {                      /* vertical flip */
    byte *topp, *botp;
    
    for (i=0; i<w; i++) {
      topp = pic + i*bperpix;
      botp = pic + (h-1)*bperlin + i*bperpix;
      
      for (j=0; j<h/2; j++, topp+=(w-1)*bperpix, botp-=(w+1)*bperpix) {
	for (l=0; l<bperpix; l++, topp++, botp++) {
	  k = *topp;  *topp = *botp;  *botp = k;
	}
      }
    }
  }
}


/***********************************************************/
byte *ImgBasic::XVGetSubImage(byte *pic, int ptype, int w,int h,
                                        int sx,int sy,int sw,int sh)
{
  /* mallocs and returns the selected subimage (sx,sy,sw,sh) of pic.
     selection is guaranteed to be within pic boundaries.  
     NEVER RETURNS NULL */

  byte *rpic, *sp, *dp;
  int   bperpix,y;

  /* sanity check: */
  if (sx<0 || sy<0 || sw<1 || sh<1 || sw>w || sh>h || sx>w || sy>h) {
    char buf[256];
    sprintf(buf,"XVGetSubImage: Valores fuera de rango\n"
                "w,h=%d,%d  sx=%d, sy=%d, sw=%d, sh=%d\n",
	               w, h, sx, sy, sw, sh);
    Error(buf);
    return NULL;
  }


  bperpix = (ptype==PIC8) ? 1 : 3;
  rpic = (byte *) malloc((size_t) bperpix * sw * sh);
  if (!rpic) {
    FatalError("out of memory in XVGetSubImage");
    return NULL;
  }

  for (y=0; y<sh; y++) {
    sp = pic  + ((y+sy)*w + sx) * bperpix;
    dp = rpic + (y * sw) * bperpix;
    xvbcopy((char*)sp,(char*)dp,sw*bperpix);
  }

  return rpic;
}

