/*********************************************************************************\
*     File:                  C A N N Y . C P P                                    *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * canny.cpp: contour extraction with canny algorithm
 *
 *  Contains:
 *         void AlgCanny();
 */

#include "ImgBasic.h"
#include "..\miguelan\miguelan.h"
#include <math.h>

#undef USE_ANGLE_UMBRALIZA_HISTERESIS

/*********************************************************************************/
void ImgBasic::AlgCanny(void)
{
  byte *thepic=NULL, *results=NULL;   /* Greyscaled final image */
  static char buf[20] = "0.5";        /* Buffer for user input  */
  static char buf2[20] = "50 100";
  static char buf3[20] = "25";
  int  i;               /* Counter                              */  
  float sigma;          /* Gaussian parameter                   */
  int min_threshold = UMBRAL_DETECCION_MINIMO;
  int max_threshold = 25, min_len=25;
  int pfree, typePic;
  u_int NewNumCols;
  byte *rp, *gp, *bp;

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Filtro de suavizado gausiano:"
                  "\nIntroduce la caracterstica sigma",
                   buf, 20, "0123456789.", 1);
  
  /* Check for right values                                      */
  if (strlen(buf)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  sigma = (float) atof(buf);

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Umbral mnimo y mximo para la histeresis de contornos:",
                  buf2, 20, "0123456789 ", 1);
  
  /* Check for right values                                      */
  if (strlen(buf2)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  
  sscanf(buf2,"%d %d",&min_threshold, &max_threshold);
  
  i = InputDialog("Longitud mnima de los contornos:",
                  buf3, 20, "0123456789 ", 1);
  
  /* Check for right values                                      */
  if (strlen(buf3)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  
  min_len = atoi(buf3);
  
  if (min_threshold>255) min_threshold = 255;
  else if (min_threshold<0) min_threshold=0;
  
  if (max_threshold>255) min_threshold = 255;
  else if (max_threshold<0) max_threshold=0;
  
  if (max_threshold<min_threshold) {
    i = min_threshold;
    min_threshold = max_threshold;
    max_threshold = i;
  }

  /* Call a function depending of the color information */

  if (mono) { // MONOCHROME IMAGE

    thepic = GetImage(&typePic, F_GREYSCALE, &pfree, &NewNumCols, &rp, &gp, &bp,256);
    if (!thepic) return;  //NULL PIC

    results = doCannyGrey( thepic, pWIDE, pHIGH, sigma,
                min_threshold,max_threshold,min_len);

  } else {  // COLOR IMAGE

    thepic = GetImage(&typePic, F_FULLCOLOR, &pfree, &NewNumCols, &rp, &gp, &bp,256);
    if (!thepic) return;  //NULL PIC

    results = doCanny24  ( thepic, pWIDE, pHIGH, sigma,
                min_threshold,max_threshold,min_len);
  }
 
  /* free thepic if needed */
  if (pfree) free(thepic);
  
  if (results) {
    // INSTALL NEW PIC   
    /* Store new image, allow 'undo', uptade display, etc.   */
    StoreNewPic(results,PIC8,TRUE);
  }
  RedrawPic(); // Display new image

}   /* End of GaussianBlur image                             */


/*****************************************************************************/

byte *ImgBasic::doCanny24   ( byte *pic24, int w, int h,float sigma,
                              int min_threshold, int max_threshold, int min_len ,
                              int seguimiento, int pasoApaso)
{
  byte *aux_src, *aux_dst=NULL;
  float *gradient,factor;
  int suavizado = (sigma>0.3);
  
  /* ALLOCATE MEMORY */
  aux_src = (byte*) malloc( (long)w*(long)h );
  if (!aux_src) {
    Error("Fallo al requerir memoria");
    return NULL;
  }
  gradient = (float*) calloc ( (long)w*(long)h,sizeof(float));
  if (!gradient) {
    free(aux_src);
    free(aux_dst);
    Error("Fallo al requerir memoria");
    return NULL;
  }
  aux_dst = (byte*) malloc( (long)w*(long)h );
  if (!aux_dst) {
    free(aux_src);
    free(gradient);
    Error("Fallo al requerir memoria");
    return NULL;
  }

  factor = (float) sqrt(2.0*M_PI)*sigma;
  if (factor<1) factor = (float) 1;

  /* BLUR RED COMPONENT AND COMPUTE GRADIENT! */
  RedComponent(aux_src, pic24);
  /* Si el valor de sigma es muy pequeno no se produce suavizado */  
  if (suavizado)
    doGaussianConvGrey(aux_src, aux_dst, w, h, sigma);
  doGradientComponent ( (suavizado) ? aux_dst : aux_src, gradient,
                        (suavizado) ? aux_src : aux_dst, w, h,factor);

  /* BLUR GREEN COMPONENT AND COMPUTE GRADIENT! */
  GreenComponent(aux_src, pic24);
  /* Si el valor de sigma es muy pequeno no se produce suavizado */  
  if (suavizado)
    doGaussianConvGrey(aux_src, aux_dst, w, h, sigma);
  doGradientComponent ( (suavizado) ? aux_dst : aux_src, gradient,
                        (suavizado) ? aux_src : aux_dst, w, h,factor);
  
  /* BLUR BLUE COMPONENT AND COMPUTE GRADIENT! */
  BlueComponent(aux_src, pic24);
  /* Si el valor de sigma es muy pequeno no se produce suavizado */  
  if (suavizado)
    doGaussianConvGrey(aux_src, aux_dst, w, h, sigma);
  doGradientComponent ( (suavizado) ? aux_dst : aux_src, gradient,
                        (suavizado) ? aux_src : aux_dst, w, h,factor);

  if (pasoApaso) {
    /* SHOW GRADIENT */
    //ShowImage((suavizado) ? aux_src : aux_dst,F_GREYSCALE,"Angulo del gradiente");  
    FitFltInByte(gradient,(suavizado) ? aux_dst : aux_src,w*h,(float) 4);
    ShowImage((suavizado) ? aux_dst : aux_src,F_GREYSCALE,"Mdulo del gradiente");
  }

  /* REALIZO UN SUPRESION DE PUNTOS NO MAXIMOS EN EL ANALISIS*/
  supresion_puntos_32(gradient,
                      (suavizado) ? aux_src : aux_dst,
                      (suavizado) ? aux_dst : aux_src,
                      w,h,(float)min_threshold,
                      MaximumFlt(gradient,w*h)/(float)255.49);
    
  if (pasoApaso && seguimiento) {
    /* SHOW IMAGE WITH ONLY COTOUR POINTS*/
    ShowImage((suavizado) ? aux_dst : aux_src, F_GREYSCALE,
              "Resultado de la supresin de puntos no maximos");
  }

if (seguimiento==1) {
    /* HISTERESIS THRESHOLD */
    FAST_UMBRALIZA_HISTERESIS(
                 /*src*/  (suavizado) ? aux_dst : aux_src,
                 /*dst*/  (suavizado) ? aux_src : aux_dst,
                 /*almac*/(byte**) gradient, min_threshold,
                          max_threshold,  min_len, w,h);
    /* free memory and return */
    free((suavizado) ? aux_dst : aux_src); /* src */
    free(gradient);
    return (suavizado) ? aux_src : aux_dst;/* dst */
}

if (seguimiento==2) {
    /* Seguimiento utilizando la direccin del gradiente para rechazar puntos
     * espreos (ngulo codificado de 0..256 -> 0..360 */
    ANGLE_UMBRALIZA_HISTERESIS ( 
                 /*src*/  (suavizado) ? aux_dst : aux_src,
                 /*buff_intermedio*/  (byte*) gradient,
                 /*tita*/ (suavizado) ? aux_src : aux_dst,
                 /*almac*/(byte**) (((byte*)gradient)+w*h),
                 /*final_dst */ (suavizado) ? aux_src : aux_dst,
                 max_threshold,
                 (min_threshold+max_threshold)*min_len/2,
                 min_len ,w,h);        
    /* free memory and return */
    free((suavizado) ? aux_dst : aux_src);/*src*/
    free(gradient);
    return (suavizado) ? aux_src : aux_dst;
}
    
  /* FREE TEMPORAL MEMORY */
  free((suavizado) ? aux_src : aux_dst); /* tita */
  free(gradient);

  /* RETURN CONTOUR */
  return (suavizado) ? aux_dst : aux_src;
}

/*********************************************************************************/
byte *ImgBasic::doCannyGrey ( byte *thepic, int w, int h,float sigma,
                              int min_threshold, int max_threshold, int min_len,
                              int seguimiento, int pasoApaso )
{
  byte *aux_dst=NULL, *grad_tita=NULL;
  float *gradient, factor;
  int suavizado = (sigma>0.3);


  /* ALLOCATE MEMORY */
  gradient = (float*) calloc ( (long)w*(long)h, sizeof(float));
  if (!gradient) {
    Error("Fallo al requerir memoria");
    return NULL;
  }
  grad_tita = (byte*) calloc( (long)w*(long)h, 1 );
  if (!grad_tita) {
    free(gradient);
    Error("Fallo al requerir memoria");
    return NULL;
  }

  aux_dst = (byte*) malloc( (long)w*(long)h );
  if (!aux_dst) {
    free(gradient);
    free(grad_tita);
    Error("Fallo al requerir memoria");
    return NULL;
  }

  factor = (float) sqrt(2.0*M_PI)*sigma;
  if (factor<1) factor = (float) 1;

  /* BLUR GREY COMPONENT AND COMPUTE GRADIENT! */
  if (suavizado) 
    doGaussianConvGrey(thepic, aux_dst, w, h, sigma);

  /* Si el valor de sigma es muy pequeno no se produce suavizado */  
  doGradientComponent ( (suavizado) ? aux_dst : thepic,
                        gradient, grad_tita, w, h,factor);

  if (pasoApaso) {
    /* SHOW GRADIENT */
    //ShowImage(grad_tita,F_GREYSCALE,"Angulo del gradiente");  
    FitFltInByte(gradient,aux_dst,w*h,(float) 4);
    ShowImage(aux_dst,F_GREYSCALE,"Mdulo del gradiente");  
  }

  /* REALIZO UN SUPRESION DE PUNTOS NO MAXIMOS EN EL ANALISIS*/
  supresion_puntos_32(gradient,grad_tita,aux_dst,
                      w,h,(float)min_threshold,
                      MaximumFlt(gradient,w*h)/(float)255.49);
    
  if (pasoApaso && seguimiento ) {
    /* SHOW IMAGE (before seguimiento) */
    ShowImage(aux_dst,F_GREYSCALE,
              "Resultado de la supresin de puntos no maximos");
  }

  if (seguimiento==1) {
    /* HISTERESIS THRESHOLD */
    FAST_UMBRALIZA_HISTERESIS(/*src*/ aux_dst, /*dst*/grad_tita,
                           /*almac*/ (byte**) gradient, min_threshold,
                           max_threshold,  min_len, w,h);
    free(aux_dst);
    free(gradient);
    /* return histeresis solution */
    return grad_tita;  
  }
  
  if  (seguimiento==2) {
    /* Seguimiento utilizando la direccin del gradiente para rechazar puntos
     * espreos (ngulo codificado de 0..256 -> 0..360 */
    ANGLE_UMBRALIZA_HISTERESIS ( 
                 /*src*/  aux_dst,  /*buff_aux w*h */  (byte*) gradient,
                 /*tita*/ grad_tita,
                 /*almac*/(byte**) (((byte*)gradient)+w*h),
                 /*final_dst */ grad_tita,
                 max_threshold,
                 (min_threshold+max_threshold)*min_len/2,
                 min_len ,w,h);
     /*free memory and return */
    free(aux_dst);
    free(gradient);
    /* return histeresis solution */
    return grad_tita;  
  }
  
  /* no hay seguimiento */
  /* FREE TEMPORAL MEMORY */
  free(grad_tita);
  free(gradient);

  /* RETURN SOLUTION  (without histeresis) */
  return  aux_dst;
}
/*********************************************************************************/
