/*********************************************************************************\
*     File:                  G R A D _ R O B . C P P                              *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * grad_rob.cpp: gradient computed with Roberts' window.
 *
 *  Contains:
 *         void GradientRoberts(void); // Compute the gradient!
 *         void doRobGradient24();    // Color implementation
 *         void doRobGradientGrey();  // Greyscale implementation
 */

    
#include "ImgBasic.h"
#include "..\miguelan\miguelan.h"



/*********************************************************************************/
void ImgBasic::GradientRoberts(void)
{
  /* This funcition calls doRobGreadient24 if the image is coloured  *
   * or doRobGreadientGrey if the image is monocrome                 */

  byte *thepic;         /* Copy of the formatted original image */
  byte *tmpPic;         /* Buffer where store transformed image */
  static char buf[10] = "B";   /* Buffer for user input                */
  int  i;               /* Counter                              */  
  char tipo;          /* Gaussian parameter                   */

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Gradiente segn ventanas de Roberts [-1 0; 0 1] y [0 1; -1 0]:"
                  "\nIntroduce el tipo de mdulo, 'A', 'B', 'C' o 'D'",
                  buf, 10, "ABCDabcd", 1);
  
  /* Check for right values                                      */
  if (strlen(buf)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  tipo = toupper(buf[0]);
  if (tipo<'A' || tipo>'D' ) {
    Error("Slo son vlidas letras entre 'A' y 'D'");
    return;
  }

  /* Call a function depending of the color information */

  if ( mono ) { // MONOCHROME IMAGE
    
    /* Get a standard greyscale version of the real image    */
    if (startGreyAlg(&thepic, &tmpPic)) 
	    return;      /* Failed to get the greyscale image          */

    doRobGradientGrey( thepic, tmpPic, tipo, pWIDE, pHIGH );
    
    /* Store new image, allow 'undo', uptade display, etc.   */
    endGreyAlg(thepic, tmpPic);
    
  } else {  // COLOR IMAGE

    /* Get a standard 24-bit RGB version of the real image   */
    if (start24bitAlg(&thepic, &tmpPic)) 
	    return; /* Failed to get the RGB-24bit image         */

    doRobGradient24( thepic, tmpPic, tipo, pWIDE, pHIGH );

    /* Store new image, allow 'undo', uptade display, etc.   */
    end24bitAlg(thepic, tmpPic);    
  }

}   /* End of GaussianBlur image                             */


/*********************************************************************************/
void ImgBasic::doRobGradient24  ( byte *pic24, byte *results,
                                   char tipo, int w, int h)
{
  float * gradiente;

  register int i;
  register byte *byte_ptr;
  register float *flt_ptr;
  byte *aux_dst;

  /* ALLOCATE MEMORY */
  
  gradiente = (float *) malloc(w * h * sizeof(float));
  if (!gradiente) {
     FatalError("Fallo al requerir memoria");
  }
  
  aux_dst   = (byte *)  malloc(w * h );
  if (!aux_dst) {
     FatalError("Fallo al requerir memoria");
  }
   
  /* GRADIENT OF RED COMPONENT! */
  byte_ptr=pic24; flt_ptr=gradiente;
  for(i=w * h ; i; i--, byte_ptr+=3)
   *flt_ptr++ = *byte_ptr;
  
  gradiente_32_exp(gradiente, gradiente, aux_dst,
		        (float)1.0,tipo, w, h, NO_UMBRALIZA);
  
  flt_ptr=gradiente; byte_ptr=results;
  for(i=w * h ; i; i--, byte_ptr+=3, flt_ptr++) 
   *byte_ptr = (*flt_ptr>255) ? 255 : /*(*flt_ptr<0) ? 0 :*/ (byte) *flt_ptr ;
  
  /* GRADIENT OF GREEN COMPONENT! */
  byte_ptr=pic24+1; flt_ptr=gradiente;
  for(i=w * h ; i; i--, byte_ptr+=3) *flt_ptr++ = *byte_ptr;
  
  gradiente_32_exp(gradiente, gradiente, aux_dst,
		        (float)1.0,tipo, w, h, NO_UMBRALIZA);
  
  flt_ptr=gradiente; byte_ptr=results+1;
  for(i=w * h ; i; i--, byte_ptr+=3, flt_ptr++) 
   *byte_ptr = (*flt_ptr>255) ? 255 : /*(*flt_ptr<0) ? 0 :*/ (byte) *flt_ptr ;
  
  /* GRADIENT OF BLUE COMPONENT! */
  byte_ptr=pic24+2; flt_ptr=gradiente;
  for(i=w * h ; i; i--, byte_ptr+=3) *flt_ptr++ = *byte_ptr;
  
  gradiente_32_exp(gradiente, gradiente, aux_dst,
		        (float)1.0,tipo, w, h, NO_UMBRALIZA);
  
  flt_ptr=gradiente; byte_ptr=results+2;
  for(i=w * h ; i; i--, byte_ptr+=3, flt_ptr++) 
   *byte_ptr = (*flt_ptr>255) ? 255 : /*(*flt_ptr<0) ? 0 :*/ (byte) *flt_ptr ;
  
  /* FREE MEMORY */
  free(gradiente);
  free(aux_dst);

  saturePic24(results,0,0,w,h,4);
}

/*********************************************************************************/
void ImgBasic::doRobGradientGrey( byte *picGrey, byte *results,
                                   char tipo, int w, int h)
{
  float *gradiente;
  byte  *gradientetita = results;

  /* Call the actual algorithm for greyscale images        */
  gradiente = (float *)malloc(w * h * sizeof(float));
  if (!gradiente) {
     FatalError("Fallo al requerir memoria");
  }
  
  /* convert byte to float */
  ConvByteToFloat(picGrey,gradiente,w,h);
  
  gradiente_32_exp(gradiente, gradiente, gradientetita,
		        (float)1.0,tipo, w, h, NO_UMBRALIZA);
  
  /* convert float to byte */
  FitFltInByte(gradiente,results, w * h,(float) 4);

  free(gradiente);      
}
/*********************************************************************************/
