/*********************************************************************************\
*     File:                  G S N O I S E . C P P                                *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * gsnoise.cpp: adds gaussian noise to image
 *
 *  Contains:
 *         void GaussianNoise();
 *         void doGaussianNoise24();
 *         void doGaussianNoiseGrey();
 */

#include "ImgBasic.h"
#include "..\miguelan\miguelan.h"

/* Algoritmo para perturbar con ruido gausiano una imagen.             */
/* Parmetros: imagen original (tamao filas*columnas), imagen final   */
/*            (con ruido aadido), magnitud del ruido ( 1 = 20dB)      */
/*            nmero de filas, nmero de columnas.                     */
void ensucia(unsigned char *picture,unsigned char *solucion,int snr,
	     int columnas,int filas);

/* Declaracin de funciones privadas de este mdulo                    */

/* Ruido gausiano */
double gaussian();

/* Devuelve un nmero aleatorio normalizado en el rango 0 a 1 */
double uniform();

/* Calcula la varianza de una matriz de puntos (tipo byte) de dimensin */
/* filas * columnas                                                     */
float variance (unsigned char *data,int columnas,int filas);

/*********************************************************************************/
void ImgBasic::GaussianNoise(void)
{

  byte *thepic;         /* Copy of the formatted original image */
  byte *tmpPic;         /* Buffer where store transformed image */
  char buf[64] = "20";  /* Buffer for user input                */
  int  i;               /* Counter                              */  
  float snr;            /* Gaussian parameter                   */
  static char *labels[] = { "Ok" }; /* Button for user input    */

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Ruido gausiano:"
                  "\nIntroduce la relacin seal/ruido en dB",
                  buf, 64, "0123456789.", 1);
  
  /* Check for right values                                      */
  if (strlen(buf)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  snr = (float) atof(buf);
  if (snr<0.1) {
    Error("El valor del ruido es muy pequeno");
    return;
  }

  /* Call a function depending of the color information */

  if (mono) { // MONOCHROME IMAGE

    /* Get a standard greyscale version of the real image    */
    if (startGreyAlg(&thepic, &tmpPic)) 
	     return;      /* Failed to get the greyscale image    */

    /* Call the actual algorithm for greyscale images        */ 
    ensucia(thepic, tmpPic,snr,pWIDE,pHIGH);

    /* Store new image, allow 'undo', uptade display, etc.   */
    endGreyAlg(thepic, tmpPic);
    
  } else {  // COLOR IMAGE

    /* Get a standard 24-bit RGB version of the real image   */
    if (start24bitAlg(&thepic, &tmpPic)) 
	return; /* Failed to get the RGB-24bit image         */
    
    /* Call the actual algorithm for color images            */ 
    doGaussianNoise24(thepic, tmpPic, pWIDE, pHIGH, snr);

    /* Store new image, allow 'undo', uptade display, etc.   */
    end24bitAlg(thepic, tmpPic);    
  }

}   /* End of GaussianNoise function                             */


/*********************************************************************************/
void ImgBasic::doGaussianNoise24   ( byte *pic24, byte *results,
                                     int w, int h, float snr )
{
  register int i;
  register byte *src, *dst;
  byte *aux_src, *aux_dst;
  
  /* ALLOCATE MEMORY */
  aux_src = (byte*) malloc( (long)w*(long)h );
  if (!aux_src) {
    Error("Fallo al requerir memoria");
    return;
  }
  aux_dst = (byte*) malloc( (long)w*(long)h );
  if (!aux_dst) {
    free(aux_src);
    Error("Fallo al requerir memoria");
    return;
  }

  /* BLUR RED COMPONENT! */
  src=pic24; dst=aux_src;
  for(i=w*h ; i; i--, src+=3) *dst++ = *src;
  ensucia(aux_src, aux_dst, snr, w, h);
  src=aux_dst; dst=results;
  for(i=w*h ; i; i--, dst+=3, src++) *dst = *src;
  
  /* BLUR GREEN COMPONENT! */
  src=pic24+1; dst=aux_src;
  for(i=w*h ; i; i--, src+=3) *dst++ = *src;
  ensucia(aux_src, aux_dst, snr, w, h);
  src=aux_dst; dst=results+1;
  for(i=w*h ; i; i--, dst+=3, src++) *dst = *src;

  /* BLUR BLUE COMPONENT! */
  src=pic24+2; dst=aux_src;
  for(i=w*h ; i; i--, src+=3) *dst++ = *src;
  ensucia(aux_src, aux_dst, snr, w, h);
  src=aux_dst; dst=results+2;
  for(i=w*h ; i; i--, dst+=3, src++) *dst = *src;

  /* FREE MEMORY */
  free(aux_src);
  free(aux_dst);
}
/*********************************************************************************/
void ImgBasic::doGaussianNoiseGrey   ( byte *picture, byte *results,
                                        int w, int h, float snr )
{
    /* Call the actual algorithm for greyscale images        */ 
    ensucia(picture,results,snr,w,h);
}
/*********************************************************************************/
