/*********************************************************************************\
*     File:               I M A G E . H                                           *
*                                                                                 *
*     Defines the ImgBasic class for data manipulation of an image               *
*     Standar    C++                                                              *
\*********************************************************************************/

#ifndef _IMAGE_H_
#define _IMAGE_H_

#include "..\\IOfile\\IOfile.h"
#include <math.h>

/* indicies to doAlg (every number reffer a algorithm) */
#define ALG_NONE      0
#define ALG_BLUR      1
#define ALG_SHARPEN   2
#define ALG_EDGE      3
#define ALG_TINF      4
#define ALG_OIL       5
#define ALG_BLEND     6
#define ALG_ROTATE    7
#define ALG_ROTATECLR 8
#define ALG_PIXEL     9
#define ALG_SPREAD    10
#define ALG_MEDIAN    11
#define ALG_RESIZE    12
#define ALG_CROP      13
#define ALG_GSBLUR    14
#define ALG_GSNOISE   15
#define ALG_GRAD_ROB  16
#define ALG_CANNY     17
#define ALG_MULTIRES  18
#define ALG_FASTMRES  19
#define ALG_CANYMRES  20
#define ALG_SATURATE  21
#define ALG_GAMMA     22
#define ALG_CANYMRES4 23
/*
#define ALG_PROY_A1   21
#define ALG_PROY_A2   22
#define ALG_PROY_B1   23
#define ALG_PROY_B2   24
*/

/* *************** Spline Line functions in GAM.CPP *************** */
double EvalSpline(int x[], int y[], double y2[], int n, double xPoint);
void   InitSpline(int x[], int y[], int n, double y2[]);
void   SortSpline(int x[], int y[], int n);
void   intsort   (int x[], int n);


/* *************** Gamma factor stuff in GAM.CPP *************** */
byte *GammifyPic24(byte *pic24, int wide, int high,
                   byte RedRemap[/*256*/], byte GreenMap[/*256*/],
                   byte BlueMap[/*256*/], int Saturation = 0,
                   int ValueRemap[/*256*/]=NULL, int HueRemap[/*360*/] = NULL);
void GenerateFSGamma(int x[],int y[],int n, byte fsgamcr[]);


class ImgBasic : virtual public IOfile
{

/* Constructor/Destructor of the class.*/
public:
    // ImgBasic Constructor
    ImgBasic( ) {};
    // ImgBasic Destructor
    ~ImgBasic() {};
    

// functions about gamma correction factor
/**************************** GAM.CPP **************************/
public:
    byte *FSDither              (byte *inpic, int intype, int w, int h,
                                 byte *fsgamcr, byte *rmap = NULL,
                                 byte *gmap = NULL, byte *bmap = NULL);
    virtual void Gamma          (void); 

// Basic display functions
/*************************** BASIC.CPP ***************************/
public:
    void InvertPic              ( byte *pic, int picType, u_int  pWIDE, u_int  pHIGH,
                                  int mono = FALSE, byte *rMap=NULL, byte *gMap=NULL,
                                  byte *bMap=NULL, int ncols=256 );
    void InvertPic              (void);
    void floydDitherize1        (byte *,int,int,byte *,int,int,int,
                                 byte *, byte *, byte *);
    void FlipPic                (int , byte *, int, u_int, u_int);
    void FlipPic                (int dir);
    byte *XVGetSubImage         (byte *, int, int,int, int,int,int,int);
    void RotatePic              (int dir, byte *pic, int ptype,
                                 u_int  *wp, u_int  *hp);
    void RotatePic              (int dir);    

/*************************** ALG.CPP ***************************/
public:
    virtual void DoAlg          (int);
    virtual void Blur           (void);
    virtual void Sharpen        (void);
    virtual void TinFoil        (void);
    virtual void OilPaint       (void);
    virtual void Blend          (void);
    virtual void FineRotate     (int);
    virtual void Pixelize       (void);
    virtual void Spread         (void);
    virtual void MedianFilter   (void);
    virtual void Saturate       (void); 

protected:
    /* Image processing algoritms */
    void doBlurConvolv24(byte *,int,int,byte *, int);
    void doBlurConvolvGrey(byte *,int,int,byte *, int);
    void doSharpConvolv (byte *,int,int,byte *, int,int,int,int, int);
    void doAngleConvolv (byte *,int,int,byte *, int,int,int,int);
    void doOilPaint     (byte *,int,int,byte *, int,int,int,int, int);
    void doBlend        (byte *,int,int,byte *, int,int,int,int);
    void doRotate       (byte *,int,int,byte *, int,int,int,int,double, int);
    void doPixel        (byte *,int,int,byte *, int,int,int,int,int, int);
    void doSpread       (byte *,int,int,byte *, int,int,int,int,int, int);
    void doMedianFilter (byte *,int,int,byte *, int,int,int,int, int);

 /*************************** SOBEL.CPP ***************************/
public:
   virtual  void SobelGradient    (void);
    void saturePicGrey    (byte* tmpPic, int pWIDE, int sx, int sy,
                           int sw, int sh, float factor=1.0);
    void saturePic24      (byte* tmpPic, int pWIDE, int sx, int sy,
                           int sw, int sh, float factor=1.0);
protected:
    void doEdgeConvolv24  (byte *,int,int,byte *, int,int,int,int);
    void doEdgeConvolvGrey( byte *pic8, int w, int h,  byte *results,
                           int selx, int sely, int selw, int selh);
 /*************************** RESIZE.CPP ***************************/
public:
    virtual void ResizePic(void);
 /*************************** CROP.CPP ***************************/
public:
    virtual void CropPic(void);      
 /*************************** GSBLUR.CPP ***************************/
public:
    virtual void GaussianBlur       (void);
protected:
    void doGaussianConv24   ( byte *pic24, byte *results, int w, int h,
                              float sigma );
    void doGaussianConvGrey ( byte *picGrey, byte *results, int w, int h,
                              float sigma );
   
 /*************************** GSNOISE.CPP ***************************/
public:
     virtual void GaussianNoise      (void);
protected:
    void doGaussianNoise24  ( byte *pic24, byte *results, int w, int h,
                              float snr );
    void doGaussianNoiseGrey( byte *picGrey, byte *results, int w, int h,
                              float snr );

 /*************************** GRAD_ROB.CPP ***************************/
public:
     virtual void GradientRoberts    (void);
protected:
    void doRobGradient24    ( byte *pic24, byte *results,
                              char tipo, int w, int h);
    void doRobGradientGrey  ( byte *picGrey, byte *results,
                              char tipo, int w, int h);
 /*************************** CANNY.CPP ***************************/
public:
   virtual void AlgCanny             (void);
protected:
  byte* doCanny24           ( byte *pic24, int w, int h,float sigma,
                              int min_threshold, int max_threshold, int lenght,
                              int seguimiento=0, int debug=FALSE);
  byte* doCannyGrey         ( byte *picGrey, int w, int h,float sigma,
                              int min_threshold, int max_threshold, int lenght,
                              int seguimiento=0, int debug=FALSE);


 /*************************** GSMULTI.CPP ***************************/
public:
     virtual void Multires            (int type);
     virtual void CannyMRes           (int type = 0);

protected:
    byte* doGaussMultires24  ( byte *pic24, int w, int h, int minlevel,
                              int maxlevel, int threshold, float reclass, int minlen,
                              int seguimiento=0, int debug=FALSE);
    byte* doGaussMultiresGrey( byte *picGrey, int w, int h, int minlevel,
                              int maxlevel, int threshold, float reclass, int minlen,
                              int seguimiento=0, int debug=FALSE);
    byte* doFastMultires24   ( byte *pic24, int w, int h, int minlevel,
                              int maxlevel, int threshold, float reclass, int minlen,
                              int seguimiento=0, int debug=FALSE);
    byte* doFastMultiresGrey ( byte *picGrey, int w, int h, int minlevel,
                              int maxlevel, int threshold, float reclass, int minlen,
                              int seguimiento=0, int debug=FALSE);
    byte* doCannyMultires24 ( byte *pic24, int w, int h, int minlevel,
                              int maxlevel, int threshold, int noise_level,
                              float reclass, int minlen,
                              int OrdenCasamiento=0, int debug=FALSE,
                              int HolguraPequena = TRUE, int noise_len = 3,
                              int holg_ang = 10, int noise_comp = -1);
    byte* doCannyMultiresGrey( byte *picGrey, int w, int h, int minlevel,
                              int maxlevel, int threshold, int noise_level,
                              float reclass, int minlen,
                              int OrdenCasamiento=0, int debug=FALSE,
                              int HolguraPequena = TRUE, int noise_len = 3,
                              int holg_ang = 10, int noise_comp = -1);

private:
    byte *AnalisisDescendente(int minlevel, int maxlevel, float reclass, 
                             int minlen, int debug=0);


/*************************** SMOOTH.CPP ***************************/
protected:
    byte *SmoothResize (byte *srcpic8, int swide, int shigh, int dwide,
                        int dhigh, byte *rmap, byte *gmap, byte *bmap,
                        byte *rdmap, byte *gdmap, byte *bdmap, int maplen);
    byte *Smooth24 (byte *pic824, int is24, int swide, int shigh, int dwide,
                          int dhigh, byte *rmap, byte *gmap, byte *bmap);
    byte *DoColorDither (byte *pic24, byte *pic8, int w, int h, byte *rmap,
                               byte *gmap, byte *bmap, byte *rdisp, byte *gdisp,
                               byte *bdisp, int maplen);
    byte *Do332ColorDither(byte *pic24, byte *pic8, int w, int h, byte *rmap, byte *gmap,
                           byte *bmap, byte *rdisp, byte *gdisp, byte *bdisp, int maplen);
private:
    int smoothX(byte *pic24, byte *pic824, int is24, int swide, int shigh,
                   int dwide, int dhigh, byte *rmap, byte *gmap, byte *bmap);
    int smoothY(byte *pic24, byte *pic824, int is24, int swide, int shigh,
                   int dwide, int dhigh, byte *rmap, byte *gmap, byte *bmap);
    int smoothXY(byte *pic24, byte *pic824, int is24, int swide, int shigh,
                    int dwide, int dhigh, byte *rmap, byte *gmap, byte *bmap);

};   // End of class ImgBasic

// inline member functions in ImgBasic

inline void ImgBasic::InvertPic (void) 
{  
  InvertPic(pic, picType, pWIDE, pHIGH,
            mono, rMap, gMap, bMap,
            numcols);
  RedrawPic();
};

inline void ImgBasic::FlipPic (int dir)
{
  FlipPic (dir, pic, picType, pWIDE, pHIGH ); 
  RedrawPic();
}

inline void ImgBasic::RotatePic (int dir)
{ 
  RotatePic (dir, pic, picType, &(pWIDE), &(pHIGH) );
  RedrawPic();
}    



#endif // #define _IMAGE_H_