/*****************************************************************************\
*     File:               I O f i l e . H                                     *
*                                                                             *
*     Defines the IOfile class for file image load and save                   *
*     Standar    C++                                                          *
\*****************************************************************************/

#ifndef _IOfile_H_
#define _IOfile_H_

#include "common.h"
#include "copyrigh.h"

class IOfile
{
// data members
public:
/* IOfile global vars */
     byte          *pic;                   /* ptr to loaded picture */
     u_int          pWIDE,pHIGH;           /* size of 'pic' */
     byte           rMap[256],gMap[256],bMap[256];  /* colormap */
     int            mono;                  /* true if displaying grayscale */
     int            picType;               /* CONV24_8BIT,CONV24_24BIT,etc.*/

     u_int          numcols;       /* # of desired colors in picture */

     char          *picComments;           /* text comments on current pic */
     char           formatStr[80];         /* short-form 'file format' */

     // Name of current file.
     char          filename[MAXPATHLEN]; /* full name of current file */

/* IOfile internal usage vars */
private:
     // Multi-page documents stuff.
     u_int          numPages, curPage;     /* for multi-page files */
     char           pageBaseName[64];      /* basename for multi-page files */

     // Message strings for general info
     char istrs[NISTR][ISTRLEN];

     // Last changed pic and type to able undo option
     byte* origPic;
     int   origPicType;
     u_int origWIDE,origHIGH;           /* size of 'origPic' */
     int   origMono;
     u_int origNumcols;
     byte  rorg[256], gorg[256], borg[256];   /*ORIGINAL colormap */

     // STATIC variables -> their values are shared by all objects IOfile
     static int     conv24;                /* 24to8 algorithm to use (CONV24_*)*/
     static int     SaveMem;               /* if true, don't suport undoes */

public:
     // Settings of TIFF and JPEG writting algorithms */
     static int     DEBUG;                 /* print debugging info */


/* Constructor/Destructor of the class */
public:
    // Constructor of IOfile
    IOfile();
    // Destructor of IOfile
    virtual ~IOfile();

/*************************** COLOR.CPP ***************************/
// functions to sort color table of mapped color images (PIC8 && !mono)
private:
    void   SortColormap         (const byte *, int, int, u_int *,
                                 const  byte*, const byte*, const byte*,
                                 byte *, byte *);
    void   ColorCompress8       (byte *, int, int, u_int, byte*,byte*,byte*,
                                 byte*, byte *);
    
/* Functions to convert 8/24 bits per pixel images
/*********************** CON24TO8.CPP & QUANT.CPP ************************/
public:
    byte *Conv24to8            (const byte *, int, int, byte *, byte *, byte *, int);
    byte *Conv8to24            (const byte *, int, int, 
                                const byte *, const byte *, const byte *);
    byte *Conv24toGrey         (const byte *, int, int);
    byte *ConvGreyto24         (const byte *, int, int);
    byte *Conv8toGrey          (const byte *, int, int, 
                                const byte *, const byte *, const byte *);

    /* Separe components from a 24 bits picture */
    void RedComponent           (byte* component, const byte *pic24 = NULL,
                                 u_int wide = 0, u_int high = 0);
    void GreenComponent         (byte* component, const byte *pic24 = NULL,
                                 u_int wide = 0, u_int high = 0);
    void BlueComponent          (byte* component, const byte *pic24 = NULL,
                                 u_int wide = 0, u_int high = 0);
    int  RGBtoPIC24             (byte* Red, byte* Green, byte* Blue,
                                 u_int wide = 0, u_int high = 0);

private:
    chist_vec mediancut          ( chist_vec chv, int colors, int sum,
                                   int maxval, int newcolors );
    int ppm_quant                (const byte *pic24, int cols,int  rows,
                                  byte *pic8, byte *rmap,byte *gmap, byte *bmap,
                                  int newcolors);
    chash_table ppm_computechash (pixel** pixels, int cols, int rows, 
                                  int maxcolors,int*  colorsP );
    void slow_map_pixels         (byte *pic24, int width, int height, byte *pic8);
    chash_table ppm_allocchash   (void);
    chist_vec   ppm_chashtochist (chash_table, int);
    chist_vec   ppm_computechist (pixel **, int,int,int,int *);

/***************** USER INTERFACE FUNCTIONS *********************/    

/**************************** INTERFAZ.CPP ***************************/
/* This functions are pure virtual, I mean to say, there are only the 
 * declaration, and each derived class must implement this functions
 * depending of the user interface (text or windows).
 */
public:
    virtual void drawStrings    (void) = 0;
    virtual void RedrawPic      (void) = 0; 
    
    virtual int  PopUp          (char *, char **, int) = 0;
    virtual int  InputDialog    (char *, char *, int,char *, int) = 0;

    virtual void OpenAlert      (char *) = 0;
    virtual void CloseAlert     (void) = 0;
    
    virtual int  PopMenu        (char *, char **, int) = 0;

    virtual void Warning        (char *str = NULL) = 0;
    virtual void Error          (char*) = 0;
    virtual void FatalError     (char*) = 0;
    
    virtual void WaitCursor     (void) = 0;
    virtual void ArrowCursor    (void) = 0;

    virtual void TIFFDialog     (void) = 0;
    virtual void JPEGDialog     (void) = 0;

    virtual void ProgressMeter  (int, int, int, char *);
    virtual int  printf         (const char *format, ...) = 0;
    virtual int  scanf          (const char *format, ...) = 0;
    virtual void TextView       (char *) = 0;

/**************************** YESNO.CPP ***************************/
public:
     virtual int YesNo           (char *str) = 0;
     virtual int YesNoCancel     (char *str) = 0;
     virtual int OkCancel        (char *str) = 0;
     virtual int RetryCancel     (char *str) = 0;
     virtual int AbortRetryIgnore(char *str) = 0;

/* End of user interface functions */

// functions for loading data
/****************************** IOFILE.CPP ****************************/
public:
    static int   SaveMemory (void) {return IOfile::SaveMem;}
    static void  SaveMemory (int status) { IOfile::SaveMem=status;}
    int IsUndoAvailable     (void) {return (origPic==NULL);}
    /* Undo stuff */
    virtual void StoreNewPic    (byte *NewPic, int NewPicType=PIC8,
                                 int NewMono=FALSE, u_int w=0, 
                                 u_int h=0, u_int  NewNumCols = 256,
                                 byte *rp=NULL, byte *gp=NULL, byte *bp=NULL);
    virtual void RestorePic     (void);

    virtual void ShowImage      (byte *image,  int  ColorType = F_GREYSCALE,
                                 char* menssage = NULL, int w=-1, int h=-1, 
                                 byte *rp=NULL, byte *gp=NULL, byte *bp=NULL);
    /* General inialization and ending routines for algoritms */
    int  start24bitAlg          (byte **, byte **);
    void end24bitAlg            (byte *, byte *);
    int  startGreyAlg           (byte **, byte **);
    void endGreyAlg             (byte *, byte *);
    /* Show time elapsed */
    void ElapsedTime            ( char* str, clock_t start);


private:
    virtual void StoreNewPic       (void); // frees former origPic && origPic -> pic
    virtual void KillOrigPic       (void);
    virtual void KillOldPics       (void);
    virtual void KillPageFiles     (char *bname, int numpages);

/*************************** MISCELA.CPP ***************************/
public:
    void SetISTR                (int stnum, ...) ;
    char *GetISTR               (int ) ;
    void CropRect2Rect          (int*,int*,int*,int*,int,int,int,int);
    virtual int  HaveSelection  (void) ; 
    virtual void GetSelRCoords  (int *, int *, int *, int *);
private:
    void  FilterString          (char *, char *, int) ;
    int  UncompressFile         (char *, char *);

/****************************** OPENPIC.CPP ****************************/
public:
    virtual int   OpenPic        ( int );
    virtual char *GetFileName    (int bOpenFileDialog = TRUE) ;
     // bOpenFileDialog: TRUE for FileOpen, FALSE for FileSaveAs
private:
    int   ReadFileType      (char *);
    int   ReadPicFile       (char *, int, PICINFO *, int);

/*************************** SAVEPIC.CPP ***************************/
public:
    virtual byte *GetImage      (int *PicTypeP, int  ColorType, int  *freeP, 
                                 u_int  *ncP = NULL, byte **rpp = NULL,
                                 byte **gpp = NULL, byte **bpp = NULL,
                                 u_int MaxAllocCols = 256);
    virtual int SavePic         (int);
    char *GetFullName           ();
    void SetFullName            (char *st);
    int  AskColorType           (void);
    int  GetColorType           (void);
    int  GetFormatFromSuffix    (void);
    void changeSuffix           (int fmt, int col);
    void ChangeImageFormat      (int colType);

private:
    virtual FILE *OpenOutFile   (char *);
    virtual int  CloseOutFile   (FILE *, char *, int);

/**************************** BMP.CPP ***************************/
private:
    int LoadBMP                (char *, PICINFO *);
    int WriteBMP               (FILE *, byte *, int, int, int, byte *, byte *,
                                byte *, int, int );
    int bmpError               (char *fname,char *st);
    
/**************************** FITS.CPP ***************************/
private:
    int  LoadFITS               (char *, PICINFO *, int);
    int  WriteFITS              (FILE *, byte *, int, int, int, byte *,
                                byte *, byte *, int, int, char *);
    int  splitfits              ( byte *, char *, int , int , int ,char *);
    //void ftfixdata              (FITS *fs, void *buffer, int nelem);
    char *rdheader             (FITS *fs);
    int  ftgbyte               (FITS *fs, byte *cbuff, int nelem);
    char *ftopen3d             (FITS *fs, char *file, int *nx, int *ny,
                                int *nz, int *bitpix);

/**************************** GIF.CPP ***************************/
private:
    int   LoadGIF              (char *, PICINFO *);
    int   gifError             (PICINFO *, char *);
    void  gifWarning           (char *);
    int   readImage            (PICINFO *);
    void  compress             (int , FILE *, byte *, int);


/*************************** GIFWR.CPP **************************/
private:
    int WriteGIF               (FILE *, byte *, int, int, int, 
                                byte *, byte *, byte *, int, int, char *);

/**************************** IFF.CPP ***************************/
private:
    int LoadIFF                (char *, PICINFO *);
    int iffError               (char *, char *);

/**************************** IMG.CPP ***************************/
private:
    int LoadIMG                (char *, PICINFO *);

/**************************** IRIS.CPP ***************************/
private:
    int LoadIRIS               (char *, PICINFO *);
    int WriteIRIS              (FILE *, byte *, int, int, int, byte *, 
                                byte *, byte *, int, int);
    byte *getimagedata         (FILE *, IRIS_IMAGE *);
    int  irisError             (char *, char *);

/**************************** JPEG.CPP ***************************/
private:
    int  LoadJFIF              (char *, PICINFO *, int);
    void writeJPEG             (int);
    int  writeJFIF             (FILE *fp, byte *pic, int w, int h, int coltype);
public:
    void SetJPEGquality        (int value);
    void SetJPEGsmooth         (int value);
    int  GetJPEGquality        (void);
    int  GetJPEGsmooth         (void);

/*************************** PAD.CPP ***************************/
private:
    int DoPad(int mode, char *str, u_int wide, u_int high, int opaque,
              int omode);
    int doPadSolid(char *str, int wide, int high, int opaque,int omode);
    int doPadBggen(char *str, int wide, int high, int opaque, int omode);
    int ReadImageFile1(char    *name, PICINFO *pinfo);
    int doPadPaste(byte *pic24, u_int wide, u_int high, int opaque,int omode);
    int doPadLoad(char *str, int wide, int high, int opaque,int omode);

/**************************** PBM.CPP ***************************/
private:
    int LoadPBM                (char *, PICINFO *);
    int WritePBM               (FILE *, byte *, int, int, int, byte *, 
                                byte *, byte *, int, int, int, char *);
    int loadpbm                (FILE *, PICINFO *, int );
    int loadpgm                (FILE *, PICINFO *, int , int );
    int loadppm                (FILE *, PICINFO *, int , int );
    int getintPM               (FILE *, PICINFO *);
    int getbitPM               (FILE *, PICINFO *);
    int getshortPM             (FILE *);
    int pbmError               (char *fname, char *st);

/**************************** PCX.CPP ***************************/
private:
    int LoadPCX                (char *, PICINFO *);
    int pcxError               (char *, char *);
    int pcxLoadImage24         (char *,FILE *, PICINFO *,byte *);
    int pcxLoadImage8          (char *,FILE *,PICINFO *,byte *);

/**************************** PDS.CPP ***************************/
private:
    int LoadPDS                (char *, PICINFO *);
    int Convert16BitImage      (char *fname, PICINFO *pinfo, int swab);

/**************************** PM.CPP ****************************/
private:
    int LoadPM                 (char *, PICINFO *);
    int WritePM                (FILE *, byte *, int, int, int, byte *, 
                                byte *, byte *, int, int, char *);
    int  pmError               (char *, char *);

/*************************** PS.CPP ***************************/
#ifdef GS_PATH
private:
    void  CreatePSD            (char *);
    virtual int PSDialog       (void )=0;
    void  PSResize             (void);
    int   LoadPS               (char *, PICINFO *, int);
#endif

/*************************** RAW.CPP ***************************/
private:
    int   LoadRAW              (char *, PICINFO *);
    int   WriteRAW             (FILE *fp, byte *pic824, int ptype, int w, int h,
                                byte *rmap=NULL, byte *gmap=NULL, byte *bmap=NULL);

/**************************** RLE.CPP ***************************/
private:
    int LoadRLE                (char *, PICINFO *);
    int rleError               (char *fname,char *st);

/**************************** SUNRAS.CPP ***************************/
private:
    int LoadSunRas             (char *, PICINFO *);
    int WriteSunRas            (FILE *, byte *, int, int, int, byte *, 
                                byte *, byte*, int, int, int);
    int sunRasError            (char *, char *);
    int rle_read               (byte *, int , int , FILE *, int );

/**************************** TIFF.CPP ***************************/
private:
    int  LoadTIFF             (char *, PICINFO *);
    int  WriteTIFF(FILE *fp, byte *pic,int ptype,int w,int h, byte *rmap, 
                   byte *gmap, byte *bmap,int numcols,int colorstyle,
                   char *fname, char *comment);
    byte *loadPalette(TIFF *tif, uint32 w, uint32 h, 
                         int photo, int bps, PICINFO *pinfo);
    byte *loadColor(TIFF *tif, uint32 w, uint32 h,
                       int photo, int bps, PICINFO *pinfo);
public:
    void SetTIFFcompression (int value);
    int  GetTIFFcompression (void);

/**************************** XBM.CPP ***************************/
private:
    int LoadXBM                (char *, PICINFO *);
    int WriteXBM               (FILE *, byte *, int, int, byte *, byte *, 
                                byte *, char *);
    int xbmError               (char *, char *);

/**************************** TARGA.CPP ***************************/
private:
    int LoadTarga              (char *, PICINFO *);
    int WriteTarga             (FILE *, byte *, int, int, int, byte *, 
                                byte *, byte *, int, int);

/**************************** XPM.CPP ***************************/
private:
    int LoadXPM                (char *, PICINFO *);
    int WriteXPM               (FILE *, byte *, int, int, int, byte *, 
                                byte *, byte *, int, int, char *, char *);
    int XpmLoadError           (char*, char*);
    int getinit                (FILE *, int*, int*, int*, CARD32 *, 
                                CARD32, PICINFO *);
    CARD32 getpixnum           (FILE *);


/**************************** XWD.CPP ***************************/
private:
    int  LoadXWD             (char *, PICINFO *);
    int  xwdError            (char *);
    void xwdWarning          (char *);
    
};   // End of class IOfile


#endif // #define _IOfile_H_
