/*********************************************************************************\
*     File:                   P A D  . C P P                                      *
*                                                                                 *
*     Standard   C++                                                              *
\*********************************************************************************/

#include <io.h>
#include <math.h>
#include "IOfile.h"

static byte *padPic = (byte *) NULL;
static byte  padmapR[256], padmapG[256], padmapB[256];
static int   padType, padWide, padHigh;

static char *holdcomment = (char *) NULL;
static char *holdfname   = (char *) NULL;

#define NOHUE -1

/*********************/
void rgb2hsv(int r,int g,int b, double *hr, double *sr, double *vr)
{
  double rd, gd, bd, h, s, v, max, min, del, rc, gc, bc;

  /* convert RGB to HSV */
  rd = r / 255.0;            /* rd,gd,bd range 0-1 instead of 0-255 */
  gd = g / 255.0;
  bd = b / 255.0;

  /* compute maximum of rd,gd,bd */
  if (rd>=gd) { if (rd>=bd) max = rd;  else max = bd; }
         else { if (gd>=bd) max = gd;  else max = bd; }

  /* compute minimum of rd,gd,bd */
  if (rd<=gd) { if (rd<=bd) min = rd;  else min = bd; }
         else { if (gd<=bd) min = gd;  else min = bd; }

  del = max - min;
  v = max;
  if (max != 0.0) s = (del) / max;
             else s = 0.0;

  h = NOHUE;
  if (s != 0.0) {
    rc = (max - rd) / del;
    gc = (max - gd) / del;
    bc = (max - bd) / del;

    if      (rd==max) h = bc - gc;
    else if (gd==max) h = 2 + rc - bc;
    else if (bd==max) h = 4 + gc - rc;

    h = h * 60;
    if (h<0) h += 360;
  }

  *hr = h;  *sr = s;  *vr = v;
}



/*********************/
void hsv2rgb(double h, double s, double v, int *rr, int *gr, int *br)
{
  int    j;
  double rd, gd, bd;
  double f, p, q, t;

  /* convert HSV back to RGB */
  if (h==NOHUE || s==0.0) { rd = v;  gd = v;  bd = v; }
  else {
    if (h==360.0) h = 0.0;
    h = h / 60.0;
    j = (int) floor(h);
    if (j<0) j=0;          /* either h or floor seem to go neg on some sys */
    f = h - j;
    p = v * (1-s);
    q = v * (1 - (s*f));
    t = v * (1 - (s*(1 - f)));

    switch (j) {
    case 0:  rd = v;  gd = t;  bd = p;  break;
    case 1:  rd = q;  gd = v;  bd = p;  break;
    case 2:  rd = p;  gd = v;  bd = t;  break;
    case 3:  rd = p;  gd = q;  bd = v;  break;
    case 4:  rd = t;  gd = p;  bd = v;  break;
    case 5:  rd = v;  gd = p;  bd = q;  break;
    default: rd = v;  gd = t;  bd = p;  break;  /* never happen */
    }
  }

  *rr = (int) floor((rd * 255.0) + 0.5);
  *gr = (int) floor((gd * 255.0) + 0.5);
  *br = (int) floor((bd * 255.0) + 0.5);
}


/***********************************/
int IOfile::DoPad(int mode, char *str, u_int wide, u_int high, int opaque, int omode)
{
  /* parses/verifies user-entered string.  If valid, does the thing, and
     installs the new pic and all that...  Returns '0' on failure */

  int   rv;
  //char  loadName[256];

  if (padPic)      free(padPic);
  if (holdcomment) free(holdcomment);
  if (holdfname)   free(holdcomment);
  padPic = (byte *) NULL;
  holdcomment = holdfname = (char *) NULL;

  rv = 1;

  if ((mode != PAD_LOAD) && (wide == pWIDE && high == pHIGH && opaque==100)) {
    Error("Padding to same size as pic while fully opaque has no effect.");
    return 0;
  }

  WaitCursor();
  
  if      (mode == PAD_SOLID) rv = doPadSolid(str, wide, high, opaque,omode);
  else if (mode == PAD_BGGEN) rv = doPadBggen(str, wide, high, opaque,omode); 
  else if (mode == PAD_LOAD)  rv = doPadLoad (str, wide, high, opaque,omode); 

  ArrowCursor();

  if (!rv) return 0;

  if (picComments) {
    holdcomment = (char *) malloc(strlen(picComments) + 1);
    if (holdcomment) strcpy(holdcomment, picComments);
  }

  holdfname = (char *) malloc(strlen(filename) + 1);
  if (holdfname) strcpy(holdfname, filename);

  return 1;
}
        

/***********************************/
int LoadPad(PICINFO *pinfo, char *fname)
{
  /* loads up into XV structures results of last 'pad' command.
     returns 0 on failure, 1 on success */

  int i;

  if (!padPic) return 0;
  pinfo->type = padType;

  for (i=0; i<256; i++) {
    pinfo->r[i] = padmapR[i];
    pinfo->g[i] = padmapG[i];
    pinfo->b[i] = padmapB[i];
  }

  pinfo->pic = padPic;
  pinfo->w   = padWide;
  pinfo->h   = padHigh;
  pinfo->frmType = -1;
  pinfo->colType = -1;

  sprintf(pinfo->fullInfo, "<%s internal>",
    (pinfo->type == PIC8) ? "8-bit" : "24-bit");

  sprintf(pinfo->shrtInfo, "%dx%d image", padWide, padHigh);

  pinfo->comment = holdcomment;

  if (holdfname) {
    strcpy(fname, holdfname);
    free(holdfname);
    holdfname = (char *) NULL;
  }
  else strcpy(fname, "");

  padPic      = (byte *) NULL;
  holdcomment = (char *) NULL;

  return 1;
}



/*******************************/
int IOfile::doPadSolid(char *str, int wide, int high, int opaque,int omode)
{
  /* returns 0 on error, 1 if successful */

  byte *pic24, *pp;
  int   i, solidRGB, r,g,b,rgb;
  char  errstr[256];

  solidRGB = 0;


  /* PARSE STRING:  'r,g,b'  'r g b'  '0xrrggbb' or <name> */

  if ((sscanf(str, "%d,%d,%d", &r,&g,&b) == 3) &&
      (r>=0 && r<=255 && g>=0 && g<=255 && b>=0 && b<=255)) {
    solidRGB = ((r&0xff)<<16) | ((g&0xff)<<8) | (b&0xff);
  }
  else if ((sscanf(str, "%d %d %d", &r,&g,&b) == 3) &&
     (r>=0 && r<=255 && g>=0 && g<=255 && b>=0 && b<=255)) {
    solidRGB = ((r&0xff)<<16) | ((g&0xff)<<8) | (b&0xff);
  }
  else if (sscanf(str, "0x%x", &rgb) && rgb>=0 && rgb<=0xffffff) {
    solidRGB = rgb;
  }
  else {
    sprintf(errstr, "Error:  Color specification '%s' not recognized.",str);
    Error(errstr);
    return 0;
  }
  


  pic24 = (byte *) malloc(wide * high * 3 * sizeof(byte));
  if (!pic24) {
    sprintf(errstr,"Error:  Can't alloc memory for %d x %d image.",
      wide, high);
    Error(errstr);
    return 0;
  }

  
  /* fill pic24 with solidRGB */
  for (i=0,pp=pic24; i<wide*high; i++, pp+=3) {
    pp[0] = (solidRGB>>16) & 0xff;
    pp[1] = (solidRGB>>8)  & 0xff;
    pp[2] = (solidRGB)     & 0xff;
  }

  i = doPadPaste(pic24, wide, high, opaque,omode);

  return i;
}



/*******************************/
int IOfile::doPadBggen(char *str, int wide, int high, int opaque, int omode)
{
  int i;
  byte *bgpic24;
  char syscmd[512], fname[128], errstr[512];
  PICINFO pinfo;

  /* returns 0 on error, 1 if successful */

  if (xv_strstr(str, "-h") || xv_strstr(str, "-w") || xv_strstr(str,"-g") ||
      xv_strstr(str, ">")) {
    Error( "Opciones no vlidas: '-h', '-w' or '-g' y redireccines" );
    return 0;
  }


  sprintf(fname, "xvXXXXXX");
  mktemp(fname);

  /* run bggen to generate the background */
  sprintf(syscmd, "bggen -g %dx%d %s > %s", wide, high, str, fname);
  SetISTR(ISTR_INFO, "Running 'bggen %s'...", str);

  i = system(syscmd);
#ifdef VMS
  i = !i;       /* VMS returns 1 on success */
#endif
  if (i) {
    unlink(fname);
    sprintf(errstr, "Error:  Running '%s' failed, for some reason.", syscmd);
    Error(errstr);
    return 0;
  }
    
  /* read the file that's been created */
  if (!ReadImageFile1(fname, &pinfo)) {
    unlink(fname);
    return 0;
  }

  unlink(fname);

  if (pinfo.comment) free(pinfo.comment);
  pinfo.comment = (char *) NULL;

  if (pinfo.type == PIC24) bgpic24 = pinfo.pic;
  else {
    bgpic24 = Conv8to24(pinfo.pic,pinfo.w,pinfo.h, pinfo.r,pinfo.g,pinfo.b);
    free(pinfo.pic);  pinfo.pic = (byte *) NULL;
  }

  if (!bgpic24) {
    Error("Couldn't generate background image.  malloc() failure?");
    return 0;
  }


  i = doPadPaste(bgpic24, pinfo.w, pinfo.h, opaque,omode);

  return i;
}


/*******************************/
int IOfile::doPadLoad(char *str, int wide, int high, int opaque,int omode)
{
  int i;
  byte *bgpic24;
  char loadName[256];
  PICINFO pinfo;

  /* returns 0 on error, 1 if successful */

  /* use first word as filename to load. */
  if (sscanf(str, "%s", &loadName) != 1) {
    Error("Error:  The entered string is not valid.");
    return 0;
  }

  if (!ReadImageFile1(loadName, &pinfo)) return 0;

  if (pinfo.comment) free(pinfo.comment);
  pinfo.comment = (char *) NULL;

  if (pinfo.type == PIC24) bgpic24 = pinfo.pic;
  else {
    bgpic24 = Conv8to24(pinfo.pic,pinfo.w,pinfo.h, pinfo.r,pinfo.g,pinfo.b);
    free(pinfo.pic);  pinfo.pic = (byte *) NULL;
  }

  if (!bgpic24) {
    Error("Couldn't generate background image.  malloc() failure?");
    return 0;
  }


  i = doPadPaste(bgpic24, pinfo.w, pinfo.h, opaque,omode);

  return i;
}


/*******************************/
int IOfile::doPadPaste(byte *pic24, u_int wide, u_int high, int opaque,int omode)
{
  /* copies 'pic' onto the given 24-bit background image, converts back to
     8-bit (if necessary), and loads up pad* variables.  
     frees pic24 if necessary */

  byte *pp, *p24;
  u_int py,px, p24y,p24x, sx,sy;
  int fg, bg;
  int rval,gval,bval, r, g, b;


  fg = opaque;
  bg = 100 - opaque;


  /* copy 'pic' centered onto pic24.  */

  sx = (wide - pWIDE) / 2;
  sy = (high - pHIGH) / 2;
  
  for (py = 0; py<pHIGH; py++) {
    ProgressMeter(0, pHIGH-1, py, "Pad");
    if ((py & 0x1f)==0) WaitCursor();

    p24y = sy + py;
    if (p24y >= 0 && p24y < high) {
      for (px=0; px<pWIDE; px++) {
  p24x = sx + px;
  if (p24x >= 0 && p24x < wide) {
    p24 = pic24 + (p24y*wide  + p24x)*3;
    
    
    if (picType == PIC24) {                       /* src is PIC24 */
      pp  = pic + (py * pWIDE + px)  *3;
      r = pp[0];  g = pp[1];  b = pp[2];
    }
    else {                                        /* src is PIC8 */
      pp  = pic + (py*pWIDE + px);
      r = rMap[*pp];  g = gMap[*pp];  b = bMap[*pp];
    }
    
    if (omode == PAD_ORGB) {
      rval = (r * fg) / 100 + ((int) p24[0] * bg) / 100;
      gval = (g * fg) / 100 + ((int) p24[1] * bg) / 100;
      bval = (b * fg) / 100 + ((int) p24[2] * bg) / 100;
    }
    else {       /* one of the HSV modes */
      double fh,fs,fv,fw, bh,bs,bv,bw, h,s,v;
      fw = fg / 100.0;  bw = bg / 100.0;
      rgb2hsv(r,g,b, &fh,&fs,&fv);
      rgb2hsv((int) p24[0], (int) p24[1], (int) p24[2], &bh,&bs,&bv);

      h = s = v = 0.0;

      if (omode == PAD_OINT) {
        h = fh;
        s = fs;  
        /* v = (fv * fg) / 100.0 + (bv * bg) / 100.0; */
        v = (fv * bv * bw) + (fv * fw);
      }
      else if (omode == PAD_OSAT) {
        if (fh<0) fs = 0.0;   /* NOHUE is unsaturated */
        if (bh<0) bs = 0.0;
        h = fh;
        /* s = (fs * fg) / 100.0 + (bs * bg) / 100.0; */
        s = (fs * bs * bw) + (fs * fw);
        v = fv;  
      }
      else if (omode == PAD_OHUE) {   /* the hard one! */
        int fdeg,bdeg;
    
        fdeg = (fh<0) ? -1 : (int) floor(fh + 0.5);
        bdeg = (bh<0) ? -1 : (int) floor(bh + 0.5);

        if (fdeg>=0 && bdeg>=0) {           /* both are colors */
    /* convert H,S onto x,y coordinates on the colorwheel for
       constant V */
    
    double fx,fy, bx,by, ox,oy;

    if (fg == 100 || bg == 100) {   /* E-Z special case */
      if (fg==100) { h = fh;  s = fs;  v=fv; }
      else         { h = bh;  s = fs;  v=fv; }
    }
    else {  /* general case */
      
      fh *= (3.14159 / 180.0);    /* -> radians */
      bh *= (3.14159 / 180.0);
      
      fx = fs * cos(fh);  fy = fs * sin(fh);
      bx = bs * cos(bh);  by = bs * sin(bh);
      
      /* compute pt. on line between fx,fy and bx,by */
      ox = (fx * (fg/100.0)) + (bx * (bg/100.0));
      oy = (fy * (fg/100.0)) + (by * (bg/100.0));
      
      /* convert ox,oy back into hue,sat */
      s = sqrt((ox * ox) + (oy * oy));
      if (ox == 0.0) {
        h = (oy>=0.0) ? 90.0 : 270.0;
      }
      else {
        h = atan(oy / ox);
        h *= (180.0 / 3.14159);
        if (ox<0.0) h += 180.0;
        while (h<0.0) h += 360.0;
        while (h>=360.0) h -= 360.0;
      }
      
      v = fv;
    }
        }

        else if (fdeg<0 && bdeg<0) {        /* both are NOHUE */
    h = -1.0;
    s = fs;
    v = fv;
        }

        else if (bdeg<0) {                  /* backgrnd is white */
    h = fh;
    v = fv;
    s = (fs * fg) / 100.0;
        }

        else {                              /* foregrnd is white */
    h = bh;
    v = fv;
    s = (bs * bg) / 100.0;
        }
      }

      v = (fv * bv * bw) + (fv * fw);
      hsv2rgb(h,s,v, &rval,&gval,&bval);
    }
    
    RANGE(rval, 0, 255);  RANGE(gval, 0, 255);  RANGE(bval, 0, 255);
    *p24++ = rval;  *p24++ = gval;  *p24++ = bval;
  }
      }
    }
  }


  /* build 'padPic' appropriately */
  if (picType == PIC8) {   /* put back to 8-bit */
    padPic = Conv24to8(pic24, wide, high, padmapR, padmapG, padmapB,
                       MAX_ALLOCABLE_COLS );
    free(pic24);
    if (!padPic) {
      ArrowCursor();
      Error("Failure occured in 24to8 conversion\n");
      return 0;
    }
    padType = PIC8;
    padWide = wide;
    padHigh = high;
  }
  else {                    /* PIC24 */
    padPic  = pic24;
    padType = PIC24;
    padWide = wide;
    padHigh = high;
  }

  return 1;
}


/*******************************/
int IOfile::ReadImageFile1(char    *name, PICINFO *pinfo) 
{
  int  i, ftype;
#ifdef VMS
  char basefname[128];
#endif
  char uncompname[128], errstr[256], *uncName, *readname;

  ftype = ReadFileType(name);

  if (ftype == RFT_COMPRESS) {    /* handle compressed/gzipped files */
#ifdef VMS
    basefname[0] = '\0';
    strcpy(basefname, name);     /* remove trailing .Z */
    *rindex(basefname, '.') = '\0';
    uncName = basefname;
#else
    uncName = name;
#endif
    
    if (UncompressFile(uncName, uncompname)) {
      ftype = ReadFileType(uncompname);
      readname = uncompname;
    }
    else {
      sprintf(errstr, "Error:  Couldn't uncompress file '%s'", name);
      Error(errstr);
      return 0;
    }
  }


  if (ftype == RFT_ERROR) {
    sprintf(errstr, "Couldn't open file '%s'\n\n  %s.", name, ERRSTR(errno));
    Error(errstr);
    return 0;
  }
  else if (ftype == RFT_UNKNOWN) {
    sprintf(errstr, "Error:  File '%s' not in a recognized format.", name);
    Error(errstr);
    return 0;
  }
  else {                         /* try to read it */
    i = ReadPicFile(name, ftype, pinfo, 0);
    KillPageFiles(pinfo->pagebname, pinfo->numpages);

    if (!i || (i && (pinfo->w<=0 || pinfo->h<=0))) {
      if (i) { 
  if (pinfo->pic)     free(pinfo->pic);
  if (pinfo->comment) free(pinfo->comment);
      }
      sprintf(errstr, "Couldn't load file '%s'.", name);
      Error(errstr);
      return 0;
    }

    /* got it! */
  }

  return 1;
}

