/*=====================================================================*/
/*                 RUTINAS DE VISION POR ORDENADOR BEDEC.C             */
/*                      POR MIGUEL ANGEL MARTINEZ ESTEBAN              */
/*=====================================================================*/

/***********************************************************************/
/* Modificado por Joaqun Mur ( abril 1997) jmur@robot3.cps.unizar.es  */
/***********************************************************************/

/* Contiene el algoritmo multirresolucin completo, incluida la caracte- */
/* rizacin de contornos. Parmetros: Imagen Original, modulo de la      */
/* solucion, columnas, filas                                             */
#include <string.h>
#define IMPLEMENTATION_MIGUELAN
#include "miguelan.h"

#define far
#define farfree(a)  free(a)
#define TRACE( )

static int columnas,filas;
/* Estructura vecino: 1er indice: nmero de direccin (entre 8 posibles)
 * 2  indice: iteracin i (n de intento de encontrar un contorno.
 * valor devuelto: desplazamiento (en x o y) respecto del punto de anlisis
 * del candidato i a ser contorno
 */
                    
const int vecino_x[8][12] =  {{ 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6},
                              { 1,-1, 1,-1, 2,-2, 3,-3, 4,-4, 5,-5},
                              { 0, 0, 1,-1, 2,-2, 3,-3, 4,-4, 5,-5},
                              { 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1},
                              { 0, 0, 0, 0, 0, 0, 0, 0,-1, 1,-1, 1},
                              { 0, 0,-1, 1,-2, 2,-3, 3,-4, 4,-5, 5},
                              {-1, 1,-1, 1,-2, 2,-3, 3,-4, 4,-5, 5},
                              {-1, 1,-2, 2,-3, 3,-4, 4,-5, 5,-6, 6}};

const int vecino_y[8][12] =  {{ 0, 0, 0, 0, 0, 0, 0, 0, 1,-1, 1,-1},
                              { 1,-1, 0, 0, 1,-1, 2,-2, 2,-2, 3,-3},
                              { 1,-1, 1,-1, 2,-2, 3,-3, 4,-4, 5,-5},
                              { 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6},
                              { 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 2,-2, 3,-3},
                              { 1,-1, 0, 0, 1,-1, 2,-2, 2,-2, 3,-3},
                              { 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1}};
/* Maximo nivel de iteracin buscando un contorno cercano en la escala inferior*/
const int max_iteracion[MAX_LEVEL] = { 0,2,4,6,8,10,12};
static struct tp_bandera  banderas;
static unsigned char far *gradtita[MAX_LEVEL];
static unsigned char *supresi[MAX_LEVEL];
static int threshold;
const float sigma[MAX_LEVEL]= {0,0.3,0.6,1.2,2.4,4,7};

/*********************************************************************************/

int deteccion_max_JOAKO(float far *gradxy, unsigned char far *gradtita,
			 int x, int y, unsigned char ang_anterior)
/* Angulo viene escalado entre 0..256 como el rango 0..360, por lo que hay */
/* que desescalarlo */
{
int angulo,sector;
float g1,g2;
float ux,uy,angulo_rad;
int distancia_256;

if (*(gradxy+x+y*columnas) > threshold)
  {
	  /* CALCULO EL SECTOR DEL ANGULO DEL GRADIENTE EN RADIANES*/
    /* Coloco el valor en 8 bits 256 valores escalados (0..256)=(0..360)*/
    angulo=gradtita[x+y*columnas];
	  if (angulo>127) angulo=angulo-128;
	  angulo_rad=angulo*2.0*M_PI/256.0;
    
    ux= (float) fabs(cos((double)angulo_rad));
    uy= (float) fabs(sin((double)angulo_rad));
    
    /* sector = ROUND(angulo_rad/(M_PI/4)); */
	   sector=angulo/32; /* 0..90, 90..180, */
       switch (sector) {
	        case 0: g1=(*(gradxy+(x-1)+(y-1)*columnas)*uy+
			                *(gradxy+(x-1)+y*columnas)*(ux-uy))/ux;
		              g2=(*(gradxy+(x+1)+(y+1)*columnas)*uy+
			                *(gradxy+(x+1)+y*columnas)*(ux-uy))/ux;
			            break;
	        case 1: g1=(*(gradxy+(x-1)+(y-1)*columnas)*ux+
			                *(gradxy+x+(y-1)*columnas)*(uy-ux))/uy;
		              g2=(*(gradxy+(x+1)+(y+1)*columnas)*ux+
			                *(gradxy+x+(y+1)*columnas)*(uy-ux))/uy;
			            break;
	        case 2: g1=(*(gradxy+(x+1)+(y-1)*columnas)*ux+
			                *(gradxy+x+(y-1)*columnas)*(uy-ux))/uy;
		              g2=(*(gradxy+(x-1)+(y+1)*columnas)*ux+
			                *(gradxy+x+(y+1)*columnas)*(uy-ux))/uy;
			          break;     
	        case 3: g1=(*(gradxy+(x+1)+(y-1)*columnas)*uy+
			                *(gradxy+(x+1)+y*columnas)*(ux-uy))/ux;
		              g2=(*(gradxy+(x-1)+(y+1)*columnas)*uy+
			                *(gradxy+(x-1)+y*columnas)*(ux-uy))/ux;
			          break;
	        }
  /* VALIDO COMO CANDIDATO A MAXIMO SI TIENE EL MODULO MAS GRANDE EN LA */
  /* DIRECCION DEL GRADIENTE Y ADEMAS CONSERVA EL ANGULO DEL PUNTO DE   */
  /* ESCALA SUPERIOR DENTRO DEL OCTANTE 45.	(en total, 8 octantes)    */
  /* 1 OCTANTE = 360/8 = 32 grados en escala (0..256)                   */

    distancia_256= (unsigned char) (gradtita[x+y*columnas]-
                              ang_anterior ) ;  
    // distancia_octantes = distancia MOD 8 == distancia & 7
    if ( (gradxy[x+y*columnas]>=g1) && (gradxy[x+y*columnas]>=g2) 
        && ( (distancia_256<16) || (distancia_256>(256-16)) ) )
      return TRUE;
    else
      return FALSE;
  }
 else
   return FALSE;
}

/*********************************************************************************/

int analiza_JOAKO(int x, int y,int contador)
{

  int despla_x=0, despla_y=0, iteracion=0;
  byte distancia_256;
  float diferencia;

  while( 1 ) { // this fuction exits with return  

  /* VALIDO COMO CANDIDATO A MAXIMO SI TIENE EL MODULO MAS GRANDE EN LA */
  /* DIRECCION DEL GRADIENTE Y ADEMAS CONSERVA EL ANGULO DEL PUNTO DE   */
  /* ESCALA SUPERIOR DENTRO DEL OCTANTE 45.	(en total, 8 octantes)    */
  /* 1 OCTANTE = 360/8 = 32 grados en escala (0..256)                   */

  distancia_256= (unsigned char) (gradtita[contador][x+y*columnas]-
            gradtita[contador][x+despla_x+(y+despla_y)*columnas] ) ;  
    // distancia_octantes = distancia MOD 8 == distancia & 7

    
  if ( (supresi[contador][x+despla_x+(y+despla_y)*columnas]
                                          > threshold)
        && ( (distancia_256<16) || (distancia_256>(256-16)) ) )
  {
     if (contador==1) {
        /* SI LA ESCALA ES LA 1 GUARDO LAS COORDENADAS DEL PUNTO A MARCAR     */
        banderas.x = x + despla_x;
        banderas.y = y + despla_y;
     } else { // contador > 1
       /* PUNTO EN LA ESCALA contador ES MAXIMO. ANALISIS A ESCALA contador-1 */
       if ( !analiza_JOAKO(x+despla_x,y+despla_y,contador-1) )
         return FALSE;
     }   
     
     /* PUNTO EN LA ESCALA contador ES MAXIMO ==> ES UN CONTORNO         */
     diferencia=supresi[contador+1][x+y*columnas]
                -supresi[contador][(x+despla_x)+(y+despla_y)*columnas];
     /* normalizacin */
     if (supresi[contador+1][x+y*columnas]>EPSILON) //No dividas por 0
         diferencia /= (*(supresi[contador+1]+x+y*columnas));
     /* marcamos el punto actual como no contorno para no volverlo a estudiar*/
     if (contador >= MIN_LEVEL)
        supresi[contador][(x+despla_x)+(y+despla_y)*columnas]=0;
     
     banderas.c[contador-1] = (diferencia < -MIN_PERCENTAGE) ? '+' :
                              (diferencia >  MIN_PERCENTAGE) ? '-' : '=' ;     
     return TRUE;     
  } /* Fin de posibilidad de que sea maximo                          */

  /* PUNTO NO MAXIMO => ANALIZAMOS OTRO PUNTO DEL ENTORNO              */
  else if ( iteracion<max_iteracion[contador] ) {
    /* divido la circunferencia en ocho direcciones principales        */
    /* Nota: trato igual a tita que a 180+tita (direcciones opuestas)  */
    int aux = (gradtita[contador+1][x + y*columnas]%128)/16;
    despla_x = vecino_x[aux][iteracion];
    despla_y = vecino_y[aux][iteracion];
    iteracion++;
    
    /* PUNTO NO ENCONTRADO EN EL ENTORNO                               */
    /* MODIFICO LOS DESPLAZAMIENTOS PARA NO SALIRME DE IMAGEN          */
    // Check range
    aux = x + despla_x;
    if (aux<1) return FALSE;
      else if (aux>(columnas-2)) return FALSE; 
    aux = y + despla_y;
    if (aux<1) return FALSE;
      else if (aux>(filas-2)) return FALSE; 
  } else 

  /* NO SE HA ENCONTRADO MAXIMO EN PUNTOS DEL ENTORNO => NO ES CONTORNO */      
  return FALSE;
  
  }/* Fin del bucle indefinido */  
}  /* Fin del procedimiento analiza_punto */

/*********************************************************************************/

int IDENTIFICA2(char tipo)
{
  switch (tipo) {

    /* COMPRUEBO TIPO DE CONTORNO ESCALON */
    case 'E': if  ( ( banderas.c[0]=='=' ) 
                &&	( (banderas.c[1]=='=') || (banderas.c[1]=='X') )
                  && ( (banderas.c[2]=='=') || (banderas.c[2]=='X') )
                    &&	( (banderas.c[3]=='=') || (banderas.c[3]=='X') )
                      && ( (banderas.c[4]=='=') || (banderas.c[4]=='X' ) ) )
                      return TRUE;
                return FALSE;

    /* COMPRUEBO TIPO DE CONTORNO RAMPA */
    case 'R' : if ( ( banderas.c[0]=='-' )
                 &&	( (banderas.c[1]=='-') || (banderas.c[1]=='=') )
                   && ( (banderas.c[2]=='-')	|| (banderas.c[2]=='=') ||
                        (banderas.c[2]=='X') )
	                   && ( (banderas.c[3]=='-' ) || (banderas.c[3]=='=') ||
                          (banderas.c[3]=='X' ) )
                        && ( (banderas.c[4]=='-')	|| (banderas.c[4]=='=') ||
                             (banderas.c[4]=='X') ) )
                  		    return TRUE;
               return FALSE;

    /* COMPRUEBO TIPO DE CONTORNO ESCALERA */
    case 'S': if ( ( banderas.c[0]=='=' )
                &&	( (banderas.c[1]=='=' ) || (banderas.c[1]=='-' ) ||
                      (banderas.c[1]=='X') )
                  && ( (banderas.c[2]=='-') || (banderas.c[2]=='X') )
                    &&	( (banderas.c[3]=='-' ) || (banderas.c[3]=='=') ||
                          (banderas.c[3]=='X' ) )
                      && ( (banderas.c[4]=='-')	|| (banderas.c[4]=='=') ||  
                           (banderas.c[4]=='X') ) )
		                    return TRUE;
                return FALSE;

    /* COMPRUEBO TIPO DE CONTORNO PULSO */
    case 'P' : if ( ( (banderas.c[0]=='=' ) || (banderas.c[0]=='+' ) )
                 &&	( (banderas.c[1]=='=') || (banderas.c[1]=='+') || 
                      (banderas.c[1]=='X') )
                   && ( (banderas.c[2]=='=') || (banderas.c[2]=='+') || 
                       (banderas.c[2]=='X') )
                    &&	( (banderas.c[3]=='+') || (banderas.c[3]=='X') )
                      && ( (banderas.c[4]=='+') || (banderas.c[4]=='X') ) )
                		    return TRUE;
                return FALSE;
  }
  return FALSE;
}

/*********************************************************************************/
  
void BEDEC   ( byte *picture, byte *solucion,
               int ncol, int nfil,
               int umbral_grad, float reclass, int minlen )
{
/* EN LAS BANDERAS GUARDO A!L PRINCIPIO LA CARACTERISRICA + - =, Y  */
/* LUEGO LAS COORDENADAS DEL PUNTO DETECTADO EN LA ESCALA MAS BAJA.*/
 /* MAX_LEVEL ha de ser menor o igual a 7 */
 int contador, x, y;
 float far *auxiliar;
 float far *mod_grad;
 float  actual,superior;
 #ifndef IMPLEMENTATION_MIGUELAN
 char mensaje[60];
 #endif

 /* Iniciacin dimensiones */
 columnas = ncol;
 filas = nfil;
 threshold=umbral_grad;

 /* BUSCA LA MEMORIA NECESARIA */
 gradtita[0]=NULL;
 supresi[0]=NULL;
 auxiliar=busca_memoria_Float(filas*columnas);
 if (!auxiliar) return;
 mod_grad=busca_memoria_Float(filas*columnas);
 if (!mod_grad) {
   free(auxiliar);
   return;
 }
 
 for ( contador=1 ; contador<MAX_LEVEL ; contador++ )
   {
    gradtita[contador]=busca_memoria(filas*columnas);
    if (!gradtita[contador]) {
        for(;contador--;) farfree(gradtita[contador]);
        free(auxiliar);
        free(mod_grad);
        TRACE("Fallo requiriendo memoria en BEDEC");
        return;
    }
    supresi[contador]= busca_memoria ( (long)filas*(long)columnas );
    if (!supresi[contador]) {
       // Fallo requiriendo memoria
       farfree(gradtita[contador]);          
       for (; contador-- ;) {
          farfree(supresi[contador]);
          farfree(gradtita[contador]);
       }
       free(auxiliar);
       free(mod_grad); 
       return;
    }
   }
 TRACE("Fin de busqueda de memoria\n");
 

 /* RELLENO CON LA 1 DERIVADA DE LA IMAGEN Y LA 1 DERIVADA DE LAS   */
 /* GAUSIANAS DE LA ORIGINAL CON UNA SIGMA DE 0.6, 1.2, 2.4, 4.8, 9.6 */
 
  for ( contador=1 ; contador<MAX_LEVEL ; contador++ ) {
    
    TRACE("SUAVIZADO ESCALA %d\n",contador);
    if (contador==1)
      ConvByteToFloat(picture,auxiliar,columnas,filas);
    else
      suavizado_gausiana_32(picture,auxiliar,columnas,filas,sigma[contador]);
    
    TRACE("Gradiente de la escala %d \n",contador);
    gradiente_32_exp(auxiliar,mod_grad,gradtita[contador],
                   factor_gaus(contador-1),
                   (char) ((contador==2) ? 'C' : (contador==3) ? 'D' : 'A'),
                   columnas,filas,NO_UMBRALIZA);

    TRACE("Supresion de puntos no maximos escala %d\n",contador);
    supresion_puntos_32(mod_grad,gradtita[contador],auxiliar,
		    columnas,filas,threshold);
    
    /* Conversin de matriz de floats a matiz de bytes */
    ConvFloatToByte(auxiliar,supresi[contador],columnas,filas);
    
    
    #ifndef IMPLEMENTATION_MIGUELAN
    if (contador==1) saturePicGrey(supresi[contador],0,0,columnas,filas,1);
    sprintf(mensaje,"supresion de puntos\nno maximos %d",contador);
    ShowImage(supresi[contador],F_GREYSCALE,mensaje);
    #endif
    
 }
 TRACE("Fin de gradiente y gausianas\n");
  
 /* Slo necesito auxiliar para guardar la imagen suavizada o 
    la supresion de puntos*/
 farfree(auxiliar);
 farfree(mod_grad);

 // La imagen solucin no est limpia, descomentar la siguiente lnea */
 /* LIMPIEZA DE LA MEMORIA DE SOLUCION */                             
 for ( x=filas*columnas ; x ; ) {
      solucion[--x] = 0;
 }
 
 TRACE("Fin de vaciado de la solucin\n");
 
 /* PUNTO HACIA ABAJO HASTA QUE LA ESCALA MAS ALTA ES LA 4         */
 for (contador=(MAX_LEVEL-1); contador >= MIN_LEVEL ; contador-- ) {
    for ( y=filas-2 ; y ; y-- )
      for ( x=columnas-2 ; x ; x-- )
        if ( supresi[contador][x+y*columnas]) {
          
          actual=supresi[contador][x+y*columnas];
          superior=supresi[MAX_LEVEL-1][x+y*columnas];

          banderas.c[0]=banderas.c[1]=banderas.c[2]=banderas.c[3]=
                      banderas.c[4]='X';
          //banderas.x=banderas.y=0;
          if (analiza_JOAKO(x,y,contador-1)) {
            /* COMPRUEBO TIPO DE CONTORNO ESCALON */
            if  ( ( banderas.c[0]=='=' ) 
                &&	( (banderas.c[1]=='=') || (banderas.c[1]=='X') )
                  && ( (banderas.c[2]=='=') || (banderas.c[2]=='X') )
                    &&	( (banderas.c[3]=='=') || (banderas.c[3]=='X') )
                      && ( (banderas.c[4]=='=') || (banderas.c[4]=='X' ) ) )
              solucion[ banderas.x + banderas.y*columnas ] = 127;
            
            /* COMPRUEBO TIPO DE CONTORNO RAMPA */
            else if ( ( banderas.c[0]=='-' )
                 &&	( (banderas.c[1]=='-') || (banderas.c[1]=='=') )
                   && ( (banderas.c[2]=='-')	|| (banderas.c[2]=='=') ||
                        (banderas.c[2]=='X') )
	                   && ( (banderas.c[3]=='-' ) || (banderas.c[3]=='=') ||
                          (banderas.c[3]=='X' ) )
                        && ( (banderas.c[4]=='-')	|| (banderas.c[4]=='=') ||
                             (banderas.c[4]=='X') ) )
                       /* && ( (contador>3) || ( (contador<=3) && 
                          (fabs(( actual-superior)/superior)>MIN_PERCENTAGE )))*/   
              solucion[ banderas.x + banderas.y*columnas ] =255;
            
            /* COMPRUEBO TIPO DE CONTORNO ESCALERA */
            else if ( ( banderas.c[0]=='=' )
                &&	( (banderas.c[1]=='=' ) || (banderas.c[1]=='-' ) ||
                      (banderas.c[1]=='X') )
                  && ( (banderas.c[2]=='-') || (banderas.c[2]=='X') )
                    &&	( (banderas.c[3]=='-' ) || (banderas.c[3]=='=') ||
                          (banderas.c[3]=='X' ) )
                      && ( (banderas.c[4]=='-')	|| (banderas.c[4]=='=') ||  
                           (banderas.c[4]=='X') )
                      /* && ( (contador>3) || ( (contador<=3) && 
                          (fabs(( actual-superior)/superior)>MIN_PERCENTAGE ) )*/ )
              solucion[ banderas.x + banderas.y*columnas ] =63;
            
            /* COMPRUEBO TIPO DE CONTORNO PULSO */
            else if (( (banderas.c[0]=='=' ) || (banderas.c[0]=='+' ) )
                 &&	( (banderas.c[1]=='=') || (banderas.c[1]=='+') || 
                      (banderas.c[1]=='X') )
                   && ( (banderas.c[2]=='=') || (banderas.c[2]=='+') || 
                       (banderas.c[2]=='X') )
                    &&	( (banderas.c[3]=='+') || (banderas.c[3]=='X') )
                      && ( (banderas.c[4]=='+') || (banderas.c[4]=='X') ))
              solucion[ banderas.x + banderas.y*columnas ]=191;
            else {
              solucion[ banderas.x + banderas.y*columnas ]=INDETERMINADO;
            }
          } // fin del marco
        } // fin del if
        #ifndef IMPLEMENTATION_MIGUELAN
        sprintf(mensaje,"Proceso busqueda e identif. escala %d\n",contador);
        ShowImage(solucion,F_GREYSCALE,mensaje);
        #endif
        TRACE("Proceso busqueda e identif. escala %d\n",contador);    
    }
  TRACE("Fin de la busqueda descendente\n");

  /* ACABADO EL PROCESO LIBERO LA MEMORIA DE LOS PUNTEROS USADOS */
  for ( contador=1 ; contador<MAX_LEVEL ; contador++) {
    farfree(supresi[contador]);
    farfree(gradtita[contador]);
  }
  
  if (reclass > 0.0) 
    RECONVERSION_TIPOS_TOTAL(solucion,reclass,
                             minlen,columnas,filas);
  return solucion;
}
/*********************************************************************************/
