
/*=====================================================================*/
/*                      RUTINAS DE VISION POR ORDENADOR                */
/*                      POR MIGUEL ANGEL MARTINEZ ESTEBAN              */
/*=====================================================================*/

#define IMPLEMENTATION_MIGUELAN  
#include "miguelan.h"

/*=====================================================================*/
/*                      Cuerpo del paquete CHANI.C                     */
/*=====================================================================*/


void discrete_2d_convolution(unsigned char far *picture,
			     unsigned char far *convolucion,
			     int columnas,int filas,int m)
{
 int filtro=1; /* Filtro Funcion de filtro de la forma f(i,j) */
 int contador;
 int salida;
 /* Guarda el valor temporal de la convolucion de cada punto */

 int x,y,i,j;
 /* Los indices x y recorren cada uno de los pixels de la imagen */
 /* Los indices i j sirven para realizar la convolucion */
 /* El valor de la variable M ha de ser par */

 for ( x=m/2 ; x<(columnas+1-(m/2)) ; x++ )
    for ( y=m/2 ; y<(filas+1-(m/2)) ; y++ )
       {
	salida=0;
	for ( i=-m/2 ; i<(m/2)+1 ; i++ )
	   for ( j=-m/2 ; j<(m/2)+1 ; j++ )
	     /* salida = salida + *(picture + (i+x) + (j+y)*columnas) * filtro; */
		salida = salida + *(picture + (i+x) + (j+y)*columnas) * 1/(1+abs(i)+abs(j));
	*(convolucion + x + y * columnas)=salida/11;
       }

 for ( x=0 ; x<(m/2) ; x++)
    for ( contador=0 ; contador<columnas ; contador++ )
      {
       *(convolucion + contador + x*columnas)=31;
       *(convolucion + contador + (filas-x-1)*columnas)=31;
       *(convolucion + x + contador*columnas)=31;
       *(convolucion + 255-x + contador*columnas)=31;
      }
}


void suavizado_gausiana(unsigned char far *picture,
			unsigned char far *gausiana,
			int columnas,int filas,float sigma)
{
/* Suavizado de imagen con Gausiana f=exp(-xx/2ss) y orlada de repeticion */
unsigned char far *aux;      /* Imagen auxiliar para Gausiana     */
int indice1,indice2,posicion;
int ancho,ancho_orla;
float valor_1,factor;
float gaus[MAX_SIGMA];

 aux=busca_memoria(filas*columnas);

/* Calculo la gausiana para no tener que repetir los calculos cada punto. */
/* Est solo la parte de (0..2sigma) */
for ( indice1=0 ; indice1<ROUND(2*sigma+1) ; indice1++ )
   gaus[indice1]=exp(-indice1*indice1/(2*sigma*sigma));

/* Calculo del factor de correccion para los colores */
factor=0;
for ( ancho=ROUND(-2*sigma) ; ancho<ROUND(2*sigma+1) ; ancho++)
     factor=factor+gaus[abs(ancho)];

/* Suavizado en x.*/
for ( indice1=0 ; indice1<filas ; indice1++ )
   for ( indice2=0 ; indice2<columnas ; indice2++ )
    {
     valor_1=0;
     for ( ancho=ROUND(-2*sigma) ; ancho<ROUND(2*sigma+1) ; ancho++)
      {
       posicion=indice2+ancho;
       if (posicion<0) ancho_orla=-(indice2+ancho);
	 else ancho_orla=ancho;
       if (posicion>255) ancho_orla=511-2*indice2-ancho;
	 else ancho_orla=ancho;
       valor_1=valor_1+*(picture+indice2+ancho_orla+indice1*columnas)*
			 gaus[abs(ancho)];
      }
       *(aux+ indice2 + indice1*columnas)=abs(ROUND(valor_1/factor));
   }


/* Suavizado en y.*/
for ( indice1=0 ; indice1<filas ; indice1++ )
   for ( indice2=0 ; indice2<columnas ; indice2++ )
    {
     valor_1=0;
     for ( ancho=ROUND(-2*sigma) ; ancho<ROUND(2*sigma+1) ; ancho++)
      {
       if (indice1+ancho<0) ancho_orla=-(indice1+ancho);
	else ancho_orla=ancho;
       if (indice1+ancho>255) ancho_orla=511-2*indice1-ancho;
	else ancho_orla=ancho;
       valor_1=valor_1+*(aux+indice2+(ancho_orla+indice1)*columnas)*
			 gaus[abs(ancho)];
      }
     *(gausiana + indice2 + indice1*columnas)=abs(ROUND(valor_1/factor));
    }

farfree(aux);

}


void supresion_puntos(unsigned char far *gradxy,
		      unsigned char far *gradtita,
		      unsigned char far *solucion,
		      int columnas, int filas,int minimo)
/* Angulo viene escalado entre 0..255 como el rango 0..360, por lo que hay */
/* que desescalarlo */
{
int indice1,indice2;
int angulo,sector;
float g1,g2;
float ux,uy,angulo_rad;

for ( indice1=0 ; indice1<=filas ; indice1++ )
   for ( indice2=0 ; indice2<=columnas ; indice2++ )
      {
       if (*(gradxy+indice2+indice1*columnas) > minimo)
	  {
	  /* CALCULO EL SECTOR DEL ANGULO DEL GRADIENTE EN RADIANES*/
    /* Coloco el valor en 8 bits 256 valores escalados (0..256)=(0..360)*/
     angulo=*(gradtita+indice2+indice1*columnas);
	   if (angulo>127) angulo=angulo-128;
	   angulo_rad=angulo*2.0*M_PI/256.0;

	   ux=fabs(cos((double)angulo_rad));
	   uy=fabs(sin((double)angulo_rad));
	   sector=angulo_rad/(M_PI*4);
	   switch (sector) {
	      case 0: g1=((float)*(gradxy+(indice2+1)+(indice1-1)*columnas)*uy+
			  (float)*(gradxy+(indice2+1)+indice1*columnas)*(ux-uy))/ux;
		      g2=((float)*(gradxy+(indice2-1)+(indice1+1)*columnas)*uy+
			  (float)*(gradxy+(indice2-1)+indice1*columnas)*(ux-uy))/ux;
			 break;
	      case 1: g1=((float)*(gradxy+(indice2+1)+(indice1-1)*columnas)*ux+
			  (float)*(gradxy+indice2+(indice1-1)*columnas)*(uy-ux))/uy;
		      g2=((float)*(gradxy+(indice2-1)+(indice1+1)*columnas)*ux+
			  (float)*(gradxy+indice2+(indice1+1)*columnas)*(uy-ux))/uy;
			 break;
	      case 2: g1=((float)*(gradxy+(indice2-1)+(indice1-1)*columnas)*ux+
			  (float)*(gradxy+indice2+(indice1-1)*columnas)*(uy-ux))/uy;
		      g2=((float)*(gradxy+(indice2+1)+(indice1+1)*columnas)*ux+
			  (float)*(gradxy+indice2+(indice1+1)*columnas)*(uy-ux))/uy;
			 break;
	      case 3: g1=((float)*(gradxy+(indice2-1)+(indice1-1)*columnas)*uy+
			  (float)*(gradxy+(indice2-1)+indice1*columnas)*(ux-uy))/ux;
		      g2=((float)*(gradxy+(indice2+1)+(indice1+1)*columnas)*uy+
			  (float)*(gradxy+(indice2+1)+indice1*columnas)*(ux-uy))/ux;
			 break;
	      }
	   if (( (float)*(gradxy+indice2+indice1*columnas) >= g1 )&&
	       ( (float)*(gradxy+indice2+indice1*columnas) >= g2 ))
		*(solucion+indice2+indice1*columnas)=255;
		else *(solucion+indice2+indice1*columnas)=0;

	  }
	  else *(solucion+indice2+indice1*columnas)=0;
      }
}



int cuenta_pixels(unsigned char far *picture, int columnas,int filas, int color)
{
int indice1,indice2;
int numero=0;

for ( indice1=0 ; indice1<=filas ; indice1++ )
   for ( indice2=0 ; indice2<=columnas ; indice2++ )
      if (*(picture+indice2+indice1*columnas)==color) numero++;
return numero;
}




