/***********************************************************************/
/* VERSIONES ACELERADAS DE LOS ALGORITMOS PARA VISIN ARTIFICIAL                               */
/* (c) Joaqun Mur ( abril 1997) jmur@robot3.cps.unizar.es  */
/***********************************************************************/

#include <string.h>

#define IMPLEMENTATION_MIGUELAN
#include "miguelan.h"
#include "tabla.h"

/******************************************************************************/

void FastSupresionPuntos (int *gradxy,
			 unsigned char *gradtita,
			 byte *solucion,
			 int columnas,int filas, int umbral_gradiente, int escala)
/* Angulo viene escalado entre 0..255 como el rango 0..360, por lo que hay */
/* que desescalarlo */
{
int indice1,indice2,angulo,caso;
int vecino_a[4];
int vecino_b[4];

/* precalculated neigbours to avoid a switch statement */
vecino_a[0] = -1;
vecino_a[2] = vecino_a[1] = -columnas;
vecino_a[3] = 1; 

vecino_b[1] = vecino_b[0] = -1-columnas;
vecino_b[3] = vecino_b[2] =  1-columnas;

/* default value: non contour */
memset((void*) solucion, 0, filas*columnas*sizeof(byte));

/* Modified by Joako to avoid excess index range */
for ( indice1=(filas-1)*columnas ; indice1 -= columnas; ) {
   indice2=columnas-1;
   {
     register int *ptr = gradxy+indice1+indice2;   
     for (  ; --indice2 ; ) { //ROW for (except borders)
      ptr--;
      if (*ptr > umbral_gradiente) {
       register float g2;
	     /* CALCULO EL SECTOR DEL ANGULO DEL GRADIENTE EN RADIANES*/
       /* Coloco el valor en 8 bits 256 valores escalados (0..127)=(0..180)*/
       angulo = (gradtita[indice2+indice1] & 127);
	     /* if (angulo>127) angulo=angulo-128; */
       /*sector=angulo/32;  0..90, 90..180, */
       caso = angulo>>5;
       /* slo marco como vlido el punto si el punto es mximo local en direccin
        * del gradiente (pongo el 1 para marcar contornos dudosos a bajas resoluciones
        * que si no no se marcaran) */
       g2 =  (float)(ptr[vecino_a[caso]] - ptr[vecino_b[caso]]);
       g2 *= Tabla_Tag_Cotg[angulo];
       // compare  float values to avoid errors in low scales
       if ((*ptr - ptr[vecino_a[caso]] ) < (int) g2 ) continue;
       g2 =  (float)(ptr[-vecino_a[caso]] - ptr[-vecino_b[caso]] + 1);
       g2 *= Tabla_Tag_Cotg[angulo];
       if ((*ptr - ptr[-vecino_a[caso]] ) < (int) g2 ) continue;
		   solucion[indice2+indice1] = (byte) (*ptr/escala);
      } //(*ptr <= umbral_gradiente)
     } /* Row for */
    }
  }    /* Column for */  
} /* End of FastSupresionPuntos */


/******************************************************************************/

void supresion_puntos_32(float *gradxy,
			 unsigned char *gradtita,
			 byte *solucion,
			 int columnas,int filas, float umbral_gradiente, float escala)
/* Angulo viene escalado entre 0..255 como el rango 0..360, por lo que hay */
/* que desescalarlo */
{
int indice1,indice2,angulo,caso;
int vecino_a[4];
int vecino_b[4];

memset((void*) solucion, 0, filas*columnas*sizeof(byte));

/* precalculated neigbours to avoid a switch statement */
vecino_a[0] = -1;
vecino_a[2] = vecino_a[1] = -columnas;
vecino_a[3] = 1; 

vecino_b[1] = vecino_b[0] = -1-columnas;
vecino_b[3] = vecino_b[2] =  1-columnas;

/* Modified by Joako to avoid excess index range */
for ( indice1=(filas-1)*columnas ; indice1 -= columnas; ) {
   register float *ptr;
   indice2=columnas-1;
   ptr = gradxy+indice1+indice2;   
   for (  ; --indice2 ; ) { //ROW for (except borders)
    ptr--;
    if (*ptr > umbral_gradiente) {
     register float g1;
     /* CALCULO EL SECTOR DEL ANGULO DEL GRADIENTE EN RADIANES*/
     /* Coloco el valor en 8 bits 256 valores escalados (0..127)=(0..180)*/
     angulo = (gradtita[indice2+indice1] & 127);
     /* if (angulo>127) angulo=angulo-128; */
     /*sector=angulo/32;  0..90, 90..180, */
     caso = angulo>>5;
     /* slo marco como vlido el punto si ptr >= g1 y ptr >= g2 */
     g1  = ptr[vecino_a[caso]];
     g1 += ( g1 - ptr[vecino_b[caso]] )*Tabla_Tag_Cotg[angulo];
     if ( *ptr < g1 ) continue;
     g1  = ptr[-vecino_a[caso]];
     g1 += ( g1 - ptr[-vecino_b[caso]] )*Tabla_Tag_Cotg[angulo];
     if ( *ptr < g1 ) continue;
		 solucion[indice2+indice1] = (byte) (*ptr/escala);
     } //(*ptr <= umbral_gradiente)
   } /* Row for */
   /* Added by Joako to fill with cero borders ( supress borders) */
  }    /* Column for */  
} /* End of supresion_puntos_32 */

/*********************************************************************************/

void SupresionPuntosCannyMRes(float *gradxy,
			 unsigned char *gradtita,
			 float *solucion,
			 int columnas,int filas, float umbral_gradiente)
/* Angulo viene escalado entre 0..255 como el rango 0..360, por lo que hay */
/* que desescalarlo */
{
int indice1,indice2,angulo,caso;
int vecino_a[4];
int vecino_b[4];

/* precalculated neigbours to avoid a switch statement */
vecino_a[0] = -1;
vecino_a[2] = vecino_a[1] = -columnas;
vecino_a[3] = 1; 

vecino_b[1] = vecino_b[0] = -1-columnas;
vecino_b[3] = vecino_b[2] =  1-columnas;

// Initialize SupressCanny. Default: no edges in image
memset((void*) solucion, 0, filas*columnas*sizeof(float));

/* Modified by Joako to avoid excess index range */
for ( indice1=(filas-1)*columnas ; indice1 -= columnas; ) {
   register float *ptr;
   indice2=columnas-1;
   ptr = gradxy+indice1+indice2;   
   for (  ; --indice2 ; ) { //ROW for (except borders)
    ptr--;
    if (*ptr > umbral_gradiente) {
     register float g1;
     /* CALCULO EL SECTOR DEL ANGULO DEL GRADIENTE EN RADIANES*/
     /* Coloco el valor en 8 bits 256 valores escalados (0..127)=(0..180)*/
     angulo = (gradtita[indice2+indice1] & 127);
     /* if (angulo>127) angulo=angulo-128; */
     /*sector=angulo/32;  0..90, 90..180, */
     caso = angulo>>5;
     /* slo marco como vlido el punto si ptr >= g1 y ptr >= g2 */
     g1  = ptr[vecino_a[caso]];
     g1 += ( g1 - ptr[vecino_b[caso]] )*Tabla_Tag_Cotg[angulo];
     if ( *ptr < g1 ) continue; // No local maximum detected
     g1  = ptr[-vecino_a[caso]];
     g1 += ( g1 - ptr[-vecino_b[caso]] )*Tabla_Tag_Cotg[angulo];
     if ( *ptr < g1 ) continue; // No local maximum detected
        solucion[indice2+indice1] = *ptr;
    } //(*ptr <= umbral_gradiente)
   } /* Row for */
   /* Added by Joako to fill with cero borders ( supress borders) */
  }    /* Column for */  
} /* End of supresion_puntos_32 */

/*********************************************************************************/



void FastGradientGrey_max (int *pt_imagen,float *pt_gradientexy,
		                   byte *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN GREY */
{
 int gx,abs_gx;                          /* Almacena el valor pasada a x */
 int gy,abs_gy;                          /* Almacena el valor pasada a y */
 int x,y;
 
  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {    
      gx= pt_imagen[x+y] - pt_imagen[x-1+y-ncol];
      abs_gx= (gx>=0) ? gx : -gx; 
      gy= pt_imagen[x-1+y]-pt_imagen[x+y-ncol] ;
      abs_gy= (gy>=0) ? gy : -gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* Calculo del modulo del gradiente : maximo(abs(gx),abs(gy))*/
      pt_gradientexy[x+y] = (float) MAX(abs_gx,abs_gy);

    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE IZQUIERDO (x=0) */
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y] =  (float)abs(pt_imagen[y] - pt_imagen[y-ncol]);
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    /* Almaceno el valor de G(x,0) */       
    pt_gradientexy[x] =  (float)abs(pt_imagen[x] - pt_imagen[x-1]);
    pt_gradientetita[x+1] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]= 0;     

}     /* fin del prodedimiento FastGradientGrey */

/******************************************************************************/

void FastGradientGrey_scale2 (int *pt_imagen,float *pt_gradientexy,
		                   byte *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN GREY */
{
 int gx,abs_gx;                          /* Almacena el valor pasada a x */
 int gy,abs_gy;                          /* Almacena el valor pasada a y */
 int x,y;
 
  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {    
      gx= pt_imagen[x+y] - pt_imagen[x-1+y-ncol];
      abs_gx= (gx>=0) ? gx : -gx; 
      gy= pt_imagen[x-1+y]-pt_imagen[x+y-ncol] ;
      abs_gy= (gy>=0) ? gy : -gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* gradxy=sqrt((media^2+8*maximo^2)/9)*escala; */
      /* variable auxiliar: maximo -> gx */
      /* variable auxiliar: media -> gy */
      gy = abs_gx+abs_gy;
      gx = MAX(abs_gx,abs_gy);
      /* El factor de escala no es necesario y podemos utilizar enteros pues 
       * este gradiente se utiliza en la escala 2 y todava no tenemos un 
       * factor de escalado de la imagen muy grande (no se produce desbordamiento
       * al calcular el producto gx*gx */
      pt_gradientexy[x+y] = (float) sqrt( ((gx*gx)<<3) + gy*gy);
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE IZQUIERDO (x=0) */
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y] =  (float)abs(pt_imagen[y] - pt_imagen[y-ncol]);
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    /* Almaceno el valor de G(x,0) */       
    pt_gradientexy[x] =  (float)abs(pt_imagen[x] - pt_imagen[x-1]);
    pt_gradientetita[x+1] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]= 0;     

}     /* fin del prodedimiento FastGradientGrey */

/*********************************************************************************/
/******************************************************************************/

void FastGradientGrey_scale3 (int *pt_imagen,float *pt_gradientexy,
		                   byte *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN GREY */
{
 int gx,abs_gx;                          /* Almacena el valor pasada a x */
 int gy,abs_gy;                          /* Almacena el valor pasada a y */
 int x,y;
 
  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {    
      gx= pt_imagen[x+y] - pt_imagen[x-1+y-ncol];
      abs_gx= (gx>=0) ? gx : -gx; 
      gy= pt_imagen[x-1+y]-pt_imagen[x+y-ncol] ;
      abs_gy= (gy>=0) ? gy : -gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* modulo_gradiente =(2*media+maximo)/3;*/
      /* (el factor de escala no es necesario) */
      pt_gradientexy[x+y] = (float)
        ( (abs_gx>abs_gy) ? ((abs_gx<<1)+abs_gy) :((abs_gy<<1)+abs_gx) );
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE IZQUIERDO (x=0) */
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y] =  (float)abs(pt_imagen[y] - pt_imagen[y-ncol]);
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    /* Almaceno el valor de G(x,0) */       
    pt_gradientexy[x] =  (float)abs(pt_imagen[x] - pt_imagen[x-1]);
    pt_gradientetita[x+1] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]= 0;     

}     /* fin del prodedimiento FastGradientGrey */

/*********************************************************************************/

void FastGradientGrey_sqrt (int *pt_imagen,float *pt_gradientexy,
		                   byte *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN GREY */
{
 int gx;                                  /* Almacena el valor pasada a x */
 int gy;                                  /* Almacena el valor pasada a y */
 int x,y;
 
  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {    
      gx= pt_imagen[x+y] - pt_imagen[x-1+y-ncol];
      gy= pt_imagen[x-1+y]-pt_imagen[x+y-ncol] ;

      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs(gx), gy, abs(gy));      
      
      /* Calculo del modulo del gradiente A : sqrt((gx*gx+gy*gy)/2)         */
      /* Paso enteros a double para evitar desbordamientos de enteros (estamos
         a altas escalas y con factores de imagen altos */
      pt_gradientexy[x+y] = (float)sqrt((double)gx*(double)gx+(double)gy*(double)gy);

    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE IZQUIERDO (x=0) */
    pt_gradientexy[y] =  (float)  /* Almaceno el valor de G(0,y) */
         abs(pt_imagen[y] - pt_imagen[y-ncol]);    
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    pt_gradientexy[x] =  (float)  /* Almaceno el valor de G(x,0) */   
         abs(pt_imagen[x] - pt_imagen[x-1]);    
    pt_gradientetita[x+1] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]= 0;     

}     /* fin del prodedimiento FastGradientGrey */

/*********************************************************************************/
/******************************************************************************/

void FastGradientGrey_suma (int *pt_imagen,float *pt_gradientexy,
		                   byte *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN GREY */
{
 int gx,abs_gx;                          /* Almacena el valor pasada a x */
 int gy,abs_gy;                          /* Almacena el valor pasada a y */
 int x,y;
 
  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {    
      gx= pt_imagen[x+y] - pt_imagen[x-1+y-ncol];
      abs_gx= (gx>=0) ? gx : -gx; 
      gy= pt_imagen[x-1+y]-pt_imagen[x+y-ncol] ;
      abs_gy= (gy>=0) ? gy : -gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* modulo_gradiente = media = suma de las componentes/2;*/
      /* (el factor de escala no es necesario) */
      pt_gradientexy[x+y] = (float) ( abs_gx + abs_gy );
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE IZQUIERDO (x=0) */
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y] =  (float)abs(pt_imagen[y] - pt_imagen[y-ncol]);
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    /* Almaceno el valor de G(x,0) */       
    pt_gradientexy[x] =  (float)abs(pt_imagen[x] - pt_imagen[x-1]);
    pt_gradientetita[x+1] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]= 0;     

}     /* fin del prodedimiento FastGradientGrey */

/*********************************************************************************/
/*********************************************************************************/

void FastGradient24_max (int *pt_imagen, float *pt_gradientexy,
		                 byte  *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN RGB */
{
 int gx,abs_gx;                       /* Almacena el valor pasada a x */
 int gy,abs_gy;                       /* Almacena el valor pasada a y */
 int x, y, *p1, *p2;
 int bpline=ncol*3;                       /* bytes per line */

  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {
      
      p1=pt_imagen+(x-1+y)*3 - bpline;
      p2=p1 + 3 + bpline;
      abs_gx= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gx = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gx : -abs_gx;

      p1=pt_imagen+ (x+y)*3 -bpline;
      p2=p1 - 3 + bpline;
      abs_gy= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gy = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gy : -abs_gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* Calculo del modulo del gradiente : maximo(abs(gx),abs(gy))*/
      pt_gradientexy[x+y] = (float) abs(MAX(abs_gx,abs_gy));
	    
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE izquierdo (x=0) */
    p1=pt_imagen+y*3;
    p2=p1-bpline;
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y]  =  (float) 
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    p1=pt_imagen+x*3;
    p2=p1-3;
    pt_gradientexy[x] =  (float)    /* Almaceno el valor de G(x,0) */   
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[x] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]=0;     

}     /* fin del prodedimiento FastGradient24 */

/******************************************************************************/
/*********************************************************************************/

void FastGradient24_scale2 (int *pt_imagen, float *pt_gradientexy,
		                 byte  *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN RGB */
{
 int gx,abs_gx;                       /* Almacena el valor pasada a x */
 int gy,abs_gy;                       /* Almacena el valor pasada a y */
 int x, y, *p1, *p2;
 int bpline=ncol*3;                       /* bytes per line */

  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {
      
      p1=pt_imagen+(x-1+y)*3 - bpline;
      p2=p1 + 3 + bpline;
      abs_gx= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gx = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gx : -abs_gx;

      p1=pt_imagen+ (x+y)*3 -bpline;
      p2=p1 - 3 + bpline;
      abs_gy= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gy = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gy : -abs_gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* gradxy=sqrt(0.2*media*media+1.6*maximo)*escala/sqrt(2.0); */
      /* variable auxiliar: maximo -> gx */
      /* variable auxiliar: media -> gy */
      gy = abs_gx+abs_gy;
      gx = MAX(abs_gx,abs_gy);
      /* El factor de escala no es necesario y podemos utilizar enteros pues 
       * este gradiente se utiliza en la escala 2 y todava no tenemos un 
       * factor de escalado de la imagen muy grande (no se produce desbordamiento
       * al calcular el producto gx*gx */
      pt_gradientexy[x+y] = (float) sqrt( ((gx*gx)<<3) + gy*gy );
	    
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE izquierdo (x=0) */
    p1=pt_imagen+y*3;
    p2=p1-bpline;
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y]  =  (float) 
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    p1=pt_imagen+x*3;
    p2=p1-3;
    pt_gradientexy[x] =  (float)    /* Almaceno el valor de G(x,0) */   
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[x] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]=0;     

}     /* fin del prodedimiento FastGradient24 */

/******************************************************************************/
/*********************************************************************************/

void FastGradient24_scale3 (int *pt_imagen, float *pt_gradientexy,
		                 byte  *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN RGB */
{
 int gx,abs_gx;                       /* Almacena el valor pasada a x */
 int gy,abs_gy;                       /* Almacena el valor pasada a y */
 int x, y, *p1, *p2;
 int bpline=ncol*3;                       /* bytes per line */

  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {
      
      p1=pt_imagen+(x-1+y)*3 - bpline;
      p2=p1 + 3 + bpline;
      abs_gx= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gx = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gx : -abs_gx;

      p1=pt_imagen+ (x+y)*3 -bpline;
      p2=p1 - 3 + bpline;
      abs_gy= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gy = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gy : -abs_gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* modulo_gradiente =(2*media+maximo)/3;*/
      /* (el factor de escala no es necesario) */
      pt_gradientexy[x+y] = (float)
        ( (abs_gx>abs_gy) ? ((abs_gx<<1)+abs_gy) :((abs_gy<<1)+abs_gx) );
	    
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE izquierdo (x=0) */
    p1=pt_imagen+y*3;
    p2=p1-bpline;
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y]  =  (float) 
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    p1=pt_imagen+x*3;
    p2=p1-3;
    pt_gradientexy[x] =  (float)    /* Almaceno el valor de G(x,0) */   
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[x] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]=0;     

}     /* fin del prodedimiento FastGradient24 */

/******************************************************************************/
/*********************************************************************************/

void FastGradient24_sqrt (int *pt_imagen, float *pt_gradientexy,
		                 byte  *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN RGB */
{
 int gx,abs_gx;                       /* Almacena el valor pasada a x */
 int gy,abs_gy;                       /* Almacena el valor pasada a y */
 int x, y, *p1, *p2;
 int bpline=ncol*3;                       /* bytes per line */

  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {
      
      p1=pt_imagen+(x-1+y)*3 - bpline;
      p2=p1 + 3 + bpline;
      abs_gx= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gx = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gx : -abs_gx;

      p1=pt_imagen+ (x+y)*3 -bpline;
      p2=p1 - 3 + bpline;
      abs_gy= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gy = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gy : -abs_gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* Calculo del modulo del gradiente A : sqrt((gx*gx+gy*gy)/2)         */
      /* Paso enteros a double para evitar desbordamientos de enteros (estamos
       a altas escalas y con factores de imagen altos */

      pt_gradientexy[x+y] = (float)sqrt((double)gx*(double)gx+(double)gy*(double)gy);

	    
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE izquierdo (x=0) */
    p1=pt_imagen+y*3;
    p2=p1-bpline;
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y]  =  (float) 
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    p1=pt_imagen+x*3;
    p2=p1-3;
    pt_gradientexy[x] =  (float)    /* Almaceno el valor de G(x,0) */   
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[x] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]=0;     

}     /* fin del prodedimiento FastGradient24 */

/******************************************************************************/
/*********************************************************************************/

void FastGradient24_suma (int *pt_imagen, float *pt_gradientexy,
		                 byte  *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN RGB */
{
 int gx,abs_gx;                       /* Almacena el valor pasada a x */
 int gy,abs_gy;                       /* Almacena el valor pasada a y */
 int x, y, *p1, *p2;
 int bpline=ncol*3;                       /* bytes per line */

  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {
      
      p1=pt_imagen+(x-1+y)*3 - bpline;
      p2=p1 + 3 + bpline;
      abs_gx= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gx = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gx : -abs_gx;

      p1=pt_imagen+ (x+y)*3 -bpline;
      p2=p1 - 3 + bpline;
      abs_gy= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gy = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? abs_gy : -abs_gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
	    pt_gradientetita[x+y]= FastAtan2(gx, abs_gx, gy, abs_gy);

      /* modulo_gradiente = media = suma de las componentes/2;*/
      /* (el factor de escala no es necesario) */
      pt_gradientexy[x+y] = (float) ( abs_gx + abs_gy );
	    
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE izquierdo (x=0) */
    p1=pt_imagen+y*3;
    p2=p1-bpline;
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y]  =  (float) 
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    p1=pt_imagen+x*3;
    p2=p1-3;
    pt_gradientexy[x] =  (float)    /* Almaceno el valor de G(x,0) */   
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[x] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float) 0.0;
 pt_gradientetita[0]=0;     

}     /* fin del prodedimiento FastGradient24 */

/******************************************************************************/

/* TABLA DE LA FUNCIN ARCOTANGENTE */
/* ( M_PI/2 - atan2( 1, i/128 ) )*256.0/( 2*M_PI) */
static const byte Tabla_Atan[129] = {
      0,   0,   1,   1,   1,   2,   2,   2,   3,   3,   3,   3, 
      4,   4,   4,   5,   5,   5,   6,   6,   6,   7,   7,   7, 
      8,   8,   8,   8,   9,   9,   9,  10,  10,  10,  11,  11, 
     11,  11,  12,  12,  12,  13,  13,  13,  13,  14,  14,  14, 
     15,  15,  15,  15,  16,  16,  16,  17,  17,  17,  17,  18, 
     18,  18,  18,  19,  19,  19,  19,  20,  20,  20,  20,  21, 
     21,  21,  21,  22,  22,  22,  22,  23,  23,  23,  23,  23, 
     24,  24,  24,  24,  25,  25,  25,  25,  25,  26,  26,  26, 
     26,  26,  27,  27,  27,  27,  27,  28,  28,  28,  28,  28, 
     29,  29,  29,  29,  29,  29,  30,  30,  30,  30,  30,  31, 
     31,  31,  31,  31,  31,  32,  32,  32,  32
};

/* this function return arc tangent of x / y in range 0..256,
 * meaning 0..360. In fact, it returns:
 * (byte) ( atan2( gy, gx ) )*256.0/( 2*M_PI) 
 */
byte __fastcall FastAtan(int x, int absx, int y, int absy)
{
  register int result;

  if (absx>=absy) result = (absx) ? Tabla_Atan[(absy<<7)/absx] : 0;
  else           result = (absy) ? (64 - Tabla_Atan[(absx<<7)/absy]) : 64;

  /* return atan2 */
  if (y>=0) 
    if (x>=0) return  result;
    else return 128 - result;
  else /* y<0 */ /* return range 0..255 */
    if (x>=0) return (byte) ((int)256-result); 
    else return (byte) ((int)128 + result);
}


/* this function return arc tangent of x / y in range 0..256,
 * meaning 0..360. In fact, it returns:
 * (byte) ( atan2( gy, gx ) + M_PI/4 )*256.0/( 2*M_PI) 
 * because the axes in Robert's gradient are rotated 45.
 */
byte __fastcall FastAtan2(int x, int absx, int y, int absy)
{
  register int result;

  if (absx>=absy) result = (absx) ? Tabla_Atan[(absy<<7)/absx] : 0;
  else           result = (absy) ? (64 - Tabla_Atan[(absx<<7)/absy]) : 64;

  /* return atan2 + 32 */
  if (y>=0) 
    if (x>=0) return 32 + result;
    else return 160 - result;
  else /* y<0 */ /* return range 0..255 */
    if (x>=0) return (byte) ((int)288-result); 
    else return (byte) ((int)160 + result);
}

/******************************************************************************/

void FastGradientGrey (int *pt_imagen,float *pt_gradientexy,
		                   byte *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN GREY */
{
 int gx,maximo /* abs maximo */;                          /* Almacena el valor pasada a x */
 int gy,minimo /* abs minimo */;                          /* Almacena el valor pasada a y */
 int x,y;
 
  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {    
      gx= pt_imagen[x+y] - pt_imagen[x-1+y-ncol];
      maximo= (gx>=0) ? gx : -gx; 
      gy= pt_imagen[x-1+y]-pt_imagen[x+y-ncol] ;
      minimo= (gy>=0) ? gy : -gy;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
      {
        register byte result ;
        /* El mximo valor absoluto lo guardo en gx,
         * El mnimo valor absouluto lo guardo en y */
        /* resultado en el primer octante (0..64)*/
        if (maximo /*abs(gx)*/ >= minimo /*abs(gy)*/) {
           if (maximo) {
              result = Tabla_Atan[((maximo-minimo)<<7)/(maximo+minimo)];
              /* Calculo el cuadrante de tita */
              if (gx>=0) { /* cuadrante I (33..64)*/ 
                if (gy>0) result = 64 - result;
              } else {
                if (gy>=0) result += 128;  /* III (128..150) */ 
                else result = 192 - result; /* III (151..192) */
              }
           } else result = 0; /* los cuatro puntos tienen el mismo valor */
        } else {
            int aux=minimo;
            minimo=maximo;
            maximo=aux;
            result = Tabla_Atan[((maximo-minimo)<<7)/(maximo+minimo)];
            /* Calculo el cuadrante de tita */
            if (gy>0) { /* cuadrante II */ 
              if (gx>=0) result += 64; /* II (64 .. 96 ) */
              else result = 128 - result; /* II (97 .. 128 ) */
            } else {
              if (gx>0) result = /*(byte)256==0*/ -result;  /* III (225..255) */ 
              else result += 192; /* III (192..224) */
            }
        }
	      pt_gradientetita[x+y]= result;
      }
  
      /* Calculo el mdulo y la direccin del gradiente mediante el
       * ajuste de mnimos cuadrados a un plano que pasase por
       * los cuatro puntos. De esta forma mejoro el comportamiento
       * frente a contornos en la cercana de 45 y me evito tener
       * varios gradientes para cada resolucin
      {
        float aux = (maximo) ? ((float) minimo)/ ((float) maximo) : (float) 0.0;
        // factor de escala: sqrt(2) 
        pt_gradientexy[x+y] = maximo * sqrt ( 1 + aux*aux );
      }        */

      /* Mismo clculo ajustado con un polinomio de tercer orden
       * de Taylor centrado en minimo=0 */
      /* MODULO = RAIZ(1+(minimo/MAXIMO)^2) * MAXIMO / RAIZ(2)						
       * Frmula aproximada con un polinomio de taylor de 5 orden
       *         centrado en x=0, (llamo x=minimo/MAXIMO)						
       * MODULO aprx = (1 + 0,5*x^2 - (3/4!)*X^4 )*MAXIMO/RAIZ(2)						
       */
        // factor de escala: 1.5 
        pt_gradientexy[x+y] = (maximo)  
          ? ( (float)maximo  + (((float)minimo) * ((float)minimo) / (float)(maximo<<1)) ) 
          :  (float) 0 ;

    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE IZQUIERDO (x=0) */
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y] =   (float) abs(pt_imagen[y] - pt_imagen[y-ncol]);
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    /* Almaceno el valor de G(x,0) */       
    pt_gradientexy[x] =   (float) abs(pt_imagen[x] - pt_imagen[x-1]);
    pt_gradientetita[x+1] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]= (float)   0;
 pt_gradientetita[0]= 0;     

}     /* fin del prodedimiento FastGradientGrey */

/*********************************************************************************/
/*********************************************************************************/

void FastGradient24 (int *pt_imagen, float *pt_gradientexy,
		                 byte  *pt_gradientetita, int ncol,int nfil)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* IMAGEN RGB */
{
 int gx,maximo;                       /* Almacena el valor pasada a x */
 int gy,minimo;                       /* Almacena el valor pasada a y */
 int x, y, *p1, *p2;
 int bpline=ncol*3;                       /* bytes per line */

  for ( y=(nfil-1)*ncol; y ; y -= ncol) {
    for ( x=ncol-1; x ; x--) {
      
      p1=pt_imagen+(x-1+y)*3 - bpline;
      p2=p1 + 3 + bpline;
      maximo= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gx = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? maximo : -maximo;

      p1=pt_imagen+ (x+y)*3 -bpline;
      p2=p1 - 3 + bpline;
      minimo= MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
      gy = (MONO(p1[0],p1[1],p1[2]) < MONO(p2[0],p2[1],p2[2])) ? minimo : -minimo;
      
      /* CALCULO NORMAL DEL ARGUMENTO */
      {
        byte result ;
        /* El mximo valor absoluto lo guardo en gx,
         * El mnimo valor absouluto lo guardo en y */
        /* resultado en el primer octante (0..64)*/
        if (maximo /*abs(gx)*/ >= minimo /*abs(gy)*/) {
           if (maximo) {
              result = Tabla_Atan[((maximo-minimo)<<7)/(maximo+minimo)];
              /* Calculo el cuadrante de tita */
              if (gx>=0) { /* cuadrante I (33..64)*/ 
                if (gy>0) result = 64 - result;
              } else {
                if (gy>=0) result += 128;  /* III (128..150) */ 
                else result = 192 - result; /* III (151..192) */
              }
           } else result = 0; /* los cuatro puntos tienen el mismo valor */
        } else {
            int aux=minimo;
            minimo=maximo;
            maximo=aux;
            result = Tabla_Atan[((maximo-minimo)<<7)/(maximo+minimo)];
            /* Calculo el cuadrante de tita */
            if (gy>0) { /* cuadrante II */ 
              if (gx>=0) result += 64; /* II (64 .. 96 ) */
              else result = 128 - result; /* II (97 .. 128 ) */
            } else {
              if (gx>0) result = /*(byte)256==0*/ -result;  /* III (225..255) */ 
              else result += 192; /* III (192..224) */
            }
        }
	      pt_gradientetita[x+y]= result;
      }

  
      /* Calculo el mdulo y la direccin del gradiente mediante el
       * ajuste de mnimos cuadrados a un plano que pasase por
       * los cuatro puntos. De esta forma mejoro el comportamiento
       * frente a contornos en la cercana de 45 y me evito tener
       * varios gradientes para cada resolucin
      {
        float aux = (maximo) ? ((float) minimo)/ ((float) maximo) : (float) 0.0;
        // factor de escala: sqrt(2) 
        pt_gradientexy[x+y] = maximo * sqrt ( 1 + aux*aux );
      }        */

      /* Mismo clculo ajustado con un polinomio de tercer orden
       * de Taylor centrado en minimo=0 */
      /* MODULO = RAIZ(1+(minimo/MAXIMO)^2) * MAXIMO / RAIZ(2)						
       * Frmula aproximada con un polinomio de taylor de 5 orden
       *         centrado en x=0, (llamo x=minimo/MAXIMO)						
       * MODULO aprx = (1 + 0,5*x^2 - (3/4!)*X^4 )*MAXIMO/RAIZ(2)						
       */
        // factor de escala: sqrt(2) 
        pt_gradientexy[x+y] = (maximo)  
          ? ( (float)maximo  + (((float)minimo) * ((float)minimo) / (float) (maximo<<1) )) 
          :  (float) 0 ;

     } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE izquierdo (x=0) */
    p1=pt_imagen+y*3;
    p2=p1-bpline;
    /* Almaceno el valor de G(0,y) */
    pt_gradientexy[y]  =  (float)
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[y] = 0;    /* 0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE SUPERIOR (y=0) */
 for (x=ncol-1; x ;x--) {
    p1=pt_imagen+x*3;
    p2=p1-3;
    pt_gradientexy[x] =    (float)/* Almaceno el valor de G(x,0) */   
      MONO(abs(p2[0]-p1[0]),abs(p2[1]-p1[1]),abs(p2[2]-p1[2]));
    pt_gradientetita[x] = 0;    /* 0 por defecto*/
 }
 
 /* Almaceno el valor de G(0,0)=0 */
 pt_gradientexy[0]=  (float) 0;
 pt_gradientetita[0]=0;     

}     /* fin del prodedimiento FastGradient24 */

/******************************************************************************/

