/*=====================================================================*/
/*  DEFINICIN FUNCIONES DE ALGORITMO MULTIRRESOLUCIN POR GAUSIANAS   */
/*  PARA LA DETECCIN Y CARACTERIZACIN DE CONTORNOS EN IMGENES 2-D   */
/*                      POR MIGUEL ANGEL MARTINEZ ESTEBAN              */
/*=====================================================================*/

/***********************************************************************/
/* Modificado por Joaqun Mur ( abril 1997) jmur@robot3.cps.unizar.es  */
/***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Tipedefs of this proyect */
#include "visionty.h"

/*=====================================================================*/
/*                 Funciones contenidas en VISITOOL.C                  */
/*=====================================================================*/

/* Si nos encontramos en la definicion INT,INT corresponde COLUMNAS, FILAS */

/* GRABA UNA IMAGEN EN DISCO EN FORMA BINARIA (tipo RAW).
 * Devuelve 0 si no ha habido fallos. */
EXTERN int graba_imagen_RAW(byte *picture,
		                     unsigned long filas_x_columnas,
		                     char nombre[]);           

/* GRABA UNA IMAGEN EN DISCO EN FORMA DE TEXTO ASCII.
 * Devuelve 0 si no ha habido fallos. */
EXTERN int graba_imagen_txt(byte *picture,
		      int columnas ,int filas,
		      char nombre[]);       /* GRABA UNA IMAGEN EN DISCO   */

/* LEE UNA IMAGEN DEL DISCO EN FORMA BINARIA (tipo RAW). 
 * Devuelve el nmero de bytes ledos o un valor negativo en caso de error.*/
EXTERN int lee_imagen_RAW(byte *picture,
                      unsigned long filas_x_columnas,
		                  char nombre[]);

/* Paso de formato de una imagen con resolucion de 8 bits a 32 bits y viceversa */
EXTERN void ConvFloatToByte(float *original,byte *copia,int columnas,int filas);

EXTERN void ConvByteToFloat(byte *original,float *copia,int columnas,int filas);


EXTERN void histograma(byte *pt_imagen,
		unsigned long filas_x_columnas,
		char file_out[]);           /* CALCULA EL HISTOGRAMA DE UNA*/
					    /* IMAGEN DE 256 COLORES       */

EXTERN void superpone(byte *original,
	       byte *mascara,unsigned long tamano);

/* BUSCA ESPACIO BYTES EN LA MEMORIA Y DEVUELVE PUNTERO */
EXTERN byte * busca_memoria(unsigned long espacio); 

/* BUSCA ESPACIO FLOATS EN LA MEMORIA Y DEVUELVE PUNTERO */
EXTERN float * busca_memoria_Float(unsigned long espacio);

EXTERN void CopySpacedIntsToBuffer ( int* src, int* dst, int total, int space);
EXTERN void CopyBufferToSpacedInts ( int* src, int* dst, int total, int space);

EXTERN void ConvByteToInt(byte *src, int *dst,int contador);
EXTERN void ConvIntToByte(int *src, byte *dst, int contador, int factor);

EXTERN void FitFltInByte(float *src, byte *dst, int contador, float factor);
EXTERN void FitFltInByteThreshold(float *src, byte *dst, int contador,
                                  float factor, byte threshold);
EXTERN float MaximumFlt (float *src, int contador);

EXTERN void FitIntInByte(int *src, byte *dst, int contador, float factor);
EXTERN int  MaximumInt (int *src, int contador);

EXTERN void FastBlur24  ( int *pic24, int columnas, int filas,int n);

EXTERN void FastBlurGrey( int *picGrey, int columnas, int filas,int n);



/*=====================================================================*/
/*                 Funciones contenidas en VISIO208.C                  */
/*=====================================================================*/

 		/* Calcula el valor mximo del mdulo del gradiente */
EXTERN float valor_max_gradiente(byte far *pt_imagen,int columnas,int filas);


EXTERN void gradiente(byte far *pt_imagen,
		byte far *pt_gradientexy,
		byte far *pt_gradientetita,
		float escala,
		int ncol,int nfil, int modo);  /* GRADIENTE DE INA IMAGEN     */


/*=====================================================================*/
/*                 Funciones contenidas en VISIO232.C                  */
/*=====================================================================*/

/* Calcula el valor mximo del mdulo del gradiente */
EXTERN float valor_max_gradiente_32(float far *pt_imagen,int columnas,int filas);

EXTERN void gradiente_32_exp(float far *pt_imagen ,
		  float far *pt_gradientexy,
		  byte far *pt_gradientetita,
		  float escala,char tipo,
		  int columnas,int filas, int modo);      /* GRADIENTE DE UNA IMAGEN */
/* NOTA: la imagen original y el resultado de la opearacin pueden ser  */
/* el mismo, es decir, si pasamos en pt_imagen y pt_gradientexy el mismo*/
/* puntero, el algoritmo calcular de forma correcta el gradiente, sobre*/
/* escribiendo la imagen original.                                      */

/* Versin incremental de gradiente_32_exp
 * Inicializar a cero pr_gradientexy antes de llamar por pirmera vez
 * a doGradientComponent. Suma el gradiente al que ya haba en pt_gradientexy
 */

EXTERN void gradiente_32_component(float far *pt_imagen,
		  float *pt_gradientexy,
		  byte far *pt_gradientetita,
		  float escala,char tipo,
		  int ncol,int nfil, int modo);

/* versin de gradiente_32_component que toma como argumento para hacer el 
 * gradiante la imagen directamente. Inicializar a cero pr_gradientexy antes 
 * de llamar por pirmera vez a doGradientComponent.
 * Suma el gradiente al que ya haba en pt_gradientexy
 */
EXTERN void doGradientComponent(byte *pt_imagen,
		  float *pt_gradientexy, byte *pt_gradientetita,
		  int ncol,int nfil, float factor);



/*=====================================================================*/
/*                 Funciones contenidas en CHANI.C                     */
/*=====================================================================*/

/* Rutinas implementadas para una resolucin en todo el proceso de 8 bits */

/* Convolucion discreta en 2D de una imagen */
EXTERN void discrete_2d_convolution(byte far *picture, byte far *convolucion,
			     int columnas,int filas,int m);

/* Suavizado de imagenes por medio de una Gausiana con ancho de 4SIGMA+1 */
EXTERN void suavizado_gausiana(byte far *picture,	byte far *gausiana,
			int columnas,int filas,float sigma);

/* Supresion de puntos no maximos */
EXTERN void supresion_puntos(byte far *gradxy,
		      byte far *gradtita, byte far *solucion,
		      int columnas, int filas,int minimo);

/* Cuenta los pixels del color COLOR en la imagen PICTURE y lo
   devuelve en un entero.*/
EXTERN int cuenta_pixels(byte far *picture,int columnas,int filas,
		  int color);

/*=====================================================================*/
/*                 Funciones contenidas en CHANI32.C                   */
/*=====================================================================*/
/* Rutina inplementadas para una de rosulucion en todo el proceso de   */
/* 32bit.                                         							       */

/* Definiciones de las rutinas usadas en este paquete                  */

/* Suavizado de imagenes por medio de una Gausiana con ancho de 4SIGMA+1*/
/* Recibe una imagen PICTURE leida en resolucion 8bits y la devuelve en */
/* 32bits (float)							 */
/* Parametros: Imagen Original, Imagen resultado, Filas, Columnas, Sigma*/
EXTERN void suavizado_gausiana_32(byte far *picture, float far *gausiana,
			  int columnas, int filas,float sigma);


/* Nos da el valor del area de una gausiana de ancho 4sigma+1 segun la */
/* escala que use .							                                        */
EXTERN float factor_gaus(int);                                                 


/*=====================================================================*/
/*                 Funciones contenidas en FAST.C                     */
/*=====================================================================*/
/* Supresion de puntos no maximos                                       */
/* Rutina acelerada para eliminar puntos que no son mximo locales
 * toma como argumentos el gradiente (mdulo y argumento)  en forma de 
 * matriz de ints, y  y el argumento en matriz de bytes (0..256)
 * equivalente a (0..360). Devuelve una matriz de bytes equivalente
 * al mdulo del gradiente dividido por escala.
 * El valor gradient_threshold indica a partir de que valor se busca un
 * mximo local. */
/* Parmetros: Gradiente mdulos, Gradiente Fases, Solucin, columnas,  */
/* filas, umbral mnimo.                                                */
EXTERN void FastSupresionPuntos (int *gradxy, byte *gradtita,
			 byte *solucion, int columnas, int filas, int gradient_threshold,
       int escala);

/* Supresion de puntos no maximos                                       */
/* Rutina standar para eliminar puntos que no son mximo locales
 * toma como argumentos el gradiente (mdulo y argumento)  en forma de 
 * matriz de floats, y  y el argumento en matriz de bytes (0..256)
 * equivalente a (0..360). Devuelve una matriz de bytes equivalente
 * al mdulo del gradiente dividido por escala.
 * El valor gradient_threshold indica a partir de que valor se busca un
 * mximo local. */
EXTERN void supresion_puntos_32(float *gradxy, byte *gradtita,
			 byte *solucion, int columnas, int filas, float gradient_threshold,
       float escala);

/* Supresion de puntos no maximos                                       */
/* Slo se diferencia respecto supresion_puntos_32 en que devuelve una
 * matriz de floats, sin el factor de escala.
 */
EXTERN void SupresionPuntosCannyMRes(float *gradxy, byte *gradtita,
			        float *solucion, int columnas,int filas, float umbral_gradiente);

/* FastAtan return arc tangent of x / y in range 0..256,
 * meaning 0..360. In fact, it returns:
 * (byte) ( atan2( gy, gx ) )*256.0/( 2*M_PI) 
 */
EXTERN byte __fastcall FastAtan(int x, int absx, int y, int absy);

/* FastAtan2 return arc tangent of x / y in range 0..256,
 * meaning 0..360. In fact, it returns:
 * (byte) ( atan2( gy, gx ) + M_PI/4 )*256.0/( 2*M_PI) 
 * because the axes in Robert's gradient are rotated 45.
 */
EXTERN byte __fastcall FastAtan2(int x, int absx, int y, int absy);

/* Rutinas para calcular el gradiente (mdulo y argumento) de forma rpida
 * y eficaz. Toman una imagen en forma de matriz de floats, y devuelven
 * el mdulo como matriz de floats y el argumento en matriz de bytes (0..256)
 * equivale a (0..360) */

EXTERN void FastGradientGrey_max (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradientGrey_scale2 (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradientGrey_scale3 (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradientGrey_sqrt (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradientGrey_suma (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);


EXTERN void FastGradient24_max  (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradient24_scale2(int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradient24_scale3(int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradient24_sqrt  (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradient24_suma (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);

/* Funciones mejoradas para trabajar nicamente con enteros a toda pastilla*/
/* Toman una imagen en forma de matriz de floats, y devuelven
 * el mdulo como matriz de enteros y el argumento en matriz de bytes (0..256)
 * equivale a (0..360).
 * Nota: se ha probado guardar el resultado en una matriz de int's en vez de float's,
 * pero a bajas resoluciones da muy malos resultados en la eliminacin de puntos
 * contorno, por lo que al final se ha dejado como matriz de float's. El incremento
 * de velocidad es muy pequeo en comparacin con el empeoramiento de los
 * resultados.*/

EXTERN void FastGradient24  (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradientGrey  (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil);
EXTERN void FastGradient24_4x4  (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil,int noise_comp);
EXTERN void FastGradientGrey4x4  (int *pt_imagen, float *pt_modulo,byte *pt_tita,
                       int ncol,int nfil,int noise_comp);

/*=====================================================================*/
/*                 Funciones contenidas en BEDEC.C                     */
/*=====================================================================*/

/* Contiene el algoritmo multirresolucin completo, incluida la caracte- */
/* rizacin de contornos. Parmetros: Imagen Original, modulo de la      */
/* solucion, columnas y filas.                                           */

EXTERN void BEDEC   ( byte *picture, byte *solucion,
                      int ncol, int nfil,
                      int umbral_grad, float reclass, int minlen );

/*=====================================================================*/
/*                 Funciones contenidas en RECALIFI.C                  */
/*=====================================================================*/

EXTERN void RECONVERSION_TIPOS_TOTAL(byte *picture,
			      float porcentaje,int long_min, int columnas,int filas);

/* Realiza un seguimiento de contornos con histresis desarrollado por   */
/* Canny. Parmetros: Imagen Original, Solucin, Umbral bajo, Umbral alto*/
/* columnas, filas.                                                      */

EXTERN void FAST_UMBRALIZA_HISTERESIS ( byte *pic, byte *soluc,
              byte **almac,int u1,int u2, int min_len, int ncol,int nfil);

/* Seguimiento utilizando la direccin del gradiente para rechazar puntos
 * espreos (ngulo codificado de 0..256 -> 0..360 */
EXTERN void ANGLE_UMBRALIZA_HISTERESIS ( 
                     byte *pic, byte *buf_fil_col /*buffer intermedio filas*columnas*/,
                     byte *tita, byte **almac,
                     byte* final_dst /* donde se copia la imagen resultado del
                             algoritmo, puede ser buf_fil_col, tita u otra zona de memoria
                             fil*col (pero no puede ser almac ni pic) */,
			               int threshold,int min_suma, int min_len,
                     int columnas,int filas);

/* Borra puntos aislados (slo til para imgenes con mucho ruido) */
EXTERN void ERASE_LONELY_POINTS_OLD ( byte *pic,  int ncol,int nfil);

EXTERN void ERASE_LONELY_POINTS ( byte *pic, byte *tita,  int ncol,int nfil);

/*=====================================================================*/
/*                 Funciones contenidas en NOISE.C                     */
/*=====================================================================*/

/* Algoritmo para perturbar con ruido gausiano una imagen.             */
/* Parmetros: imagen original (tamao filas*columnas), imagen final   */
/*            (con ruido aadido), relacin seal/ruido en dB,         */
/*            nmero de columnas, nmero de filas.                     */
EXTERN void ensucia(byte *picture,byte *solucion,float snr,
	     int columnas,int filas);

#define FAST_MONO

#ifndef MONO
    #ifdef FAST_MONO
     /* Aproximation: each color chanel, red, green and blue have the same weight
      * and the value is in range 0..3*255 = 0..765 instead of 0..255           */
      #define MONO(rd,gn,bl) ((int)(rd) + (int)(gn) + (int)(bl))
    #else
      /* MONO returns total intensity of r,g,b triple (i = .33R + .5G + .17B) */
      #define MONO(rd,gn,bl) ( ((int)(rd)*11 + (int)(gn)*16 + (int)(bl)*5) >> 5)
    #endif //FAST_MONO
#endif //MONO