/*=====================================================================*/
/*                RUTINA DISENADA POR MIGUEL ANGEL MARTINEZ            */
/*=====================================================================*/

/***********************************************************************/
/* Modificado por Joaqun Mur ( abril 1997) jmur@robot3.cps.unizar.es  */
/***********************************************************************/

#include <memory.h>


#define IMPLEMENTATION_MIGUELAN  
#include "miguelan.h"


/*=====================================================================*/
/*   RUTINAS PARA LA RECALIFICACION DEL TODOS LOS TIPOS.               */
/*=====================================================================*/

void RECONVERSION_TIPOS_TOTAL(byte *pic,
			      float porcentaje,int min_lon ,int ncol,int nfil);
/* Local functions */
void __fastcall marca_punto_contorno_total (register byte *offset );
__inline void __fastcall inicia_segmento_total      (byte color);
__inline byte __fastcall maximo_tipo_en_seg         (int  len);
__inline void __fastcall inicializa_vecinos(int filas, int columnas);
/* Global variables (used to speed up algorithms and decrease function
   parameters int RECONVERSION_TIPOS_TOTAL) */
static byte **pFinSegmento;
static int tipo_INDETERMINADO,tipo_ESCALERA,tipo_RAMPA,tipo_PULSO,tipo_ESCALON;
static int porcentaje;

/*********************************************************************************/
#define MAX_SEGS  16384
#define REDUCE_MEM_USAGE 8
#define PORCEN_SHIFT 8
#define SECURITY_ALMAC_MARGIN 256

typedef struct segTotal {
  int   nptos;
  byte  color;
  byte  **offset;
} TypeSegTotal;

typedef TypeSegTotal TypePicSegTotal[MAX_SEGS];

/*********************************************************************************/

void RECONVERSION_TIPOS_TOTAL(byte *soluc,
			      float porcent,int min_len,int columnas,int filas)
{
 int x,y, nseg=0;
 byte **pPtoEstudiado, **pInicioSegmento, **almac_limit;
 byte **almac;
 TypeSegTotal *seg;
 
 /* Inicializacin de datos */
 almac = (byte**) busca_memoria((filas*columnas*sizeof(byte*))/REDUCE_MEM_USAGE);
 if (!almac) return;
 almac_limit=almac + (filas*columnas*sizeof(byte*))/REDUCE_MEM_USAGE
                   - SECURITY_ALMAC_MARGIN;
 seg   = (TypeSegTotal*) busca_memoria(sizeof(TypePicSegTotal));
 if (!seg)  {free(almac); return;}
 pInicioSegmento=almac;
 porcentaje=(int)(porcent*(1<<PORCEN_SHIFT));
 
 for (y=(filas-1)*columnas; y -= columnas; ) {
   register byte *punto;
   x=columnas-1;
   punto = soluc+x+y;
   for (;--x;) {
     punto--;
     if ( *punto ) {
        inicia_segmento_total ( *punto );
        *punto=0;
        pFinSegmento=pPtoEstudiado=pInicioSegmento;
	      *pFinSegmento=punto;
        pFinSegmento++;
        while (pPtoEstudiado<pFinSegmento) {
          register byte* ptoVecino = (*pPtoEstudiado) + 1;
          /* (offset+1) ;          Vecino 0 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino -= columnas;      /* (offset+1-columnas) ; Vecino 1 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino--;                /* (offset  -columnas) ; Vecino 2 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino--;                /* (offset-1-columnas) ; Vecino 3 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino += columnas;      /* (offset-1)          ; Vecino 4 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino += columnas;      /* (offset-1+columnas) ; Vecino 5 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino++;                /* (offset  +columnas) ; Vecino 6 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          ptoVecino++;                /* (offset+1+columnas) ; Vecino 7 */
          if ( *ptoVecino )  marca_punto_contorno_total( ptoVecino );
          pPtoEstudiado++;
        }
        /* "Elemento extraido \n" */
        /* Update pointers depending if it's segment */
        {
          int len = (pPtoEstudiado-pInicioSegmento); 	        
          if (len > min_len) {
            /* Acept segment */
            seg[nseg].nptos=len;
            seg[nseg].color=maximo_tipo_en_seg (len);
            seg[nseg].offset=pInicioSegmento;
            nseg++;
            /* Don't excess nseg limits */
            if (nseg == MAX_SEGS) break;
            pInicioSegmento=pPtoEstudiado;
          }
        }

        /* Don't excess almac limits */
        if ( pPtoEstudiado > almac_limit ) break;          
	   }/* no es un punto de contorno (*punto == 0) */
   }
 }

  /* Copy segments detected to solution */
  for ( ; nseg ; ) {
    int counter;
    nseg--;
    counter=seg[nseg].nptos;
    pPtoEstudiado=seg[nseg].offset;
    for (; counter;) {
      counter--;
      *(pPtoEstudiado[counter])=seg[nseg].color;
    }
  }

  /* free storage memory */
  free(almac);
  free(seg);
}   /* end of RECONVERSION_TIPOS_TOTAL */

/*********************************************************************************/

void __fastcall marca_punto_contorno_total( register byte *offset )
{
  switch (*offset) {
    case INDETERMINADO:  tipo_INDETERMINADO +=2; tipo_ESCALERA++; tipo_RAMPA++; 
              tipo_PULSO++; tipo_ESCALON++; break;
    case ESCALERA:  tipo_ESCALERA++;  break;
    case RAMPA: tipo_RAMPA++; break;
    case PULSO: tipo_PULSO++; break;
    case ESCALON: tipo_ESCALON++; break;
  }
  *offset=0;
  *pFinSegmento=offset;
  pFinSegmento++;
}

/*********************************************************************************/

__inline void __fastcall inicia_segmento_total (byte color)
{
  tipo_INDETERMINADO=tipo_ESCALERA=tipo_RAMPA=tipo_PULSO=tipo_ESCALON=0;
	switch (color) {
    case INDETERMINADO: tipo_INDETERMINADO++; break;
    case ESCALERA:      tipo_ESCALERA++;      break;
	  case RAMPA:         tipo_RAMPA++;         break;
	  case PULSO:         tipo_PULSO++;         break;
	  case ESCALON:       tipo_ESCALON++;       break;
  }
}

/*********************************************************************************/
__inline byte __fastcall maximo_tipo_en_seg (int len)
{
  /* Miro secuencialmente si es escaln, pulso, rampa o escalera */
  int maximo = tipo_ESCALERA;
  byte color = ESCALERA;
  if (tipo_RAMPA>maximo)  
     {color=RAMPA;  maximo=tipo_RAMPA;}
  if (tipo_PULSO>maximo)
     {color=PULSO;  maximo=tipo_PULSO;}
  if (tipo_ESCALON>maximo)
     {color=ESCALON; maximo=tipo_ESCALON;}
  if ( (maximo<<PORCEN_SHIFT)/len < porcentaje ) 
	    color=INDETERMINADO;
  return color;
}

/*********************************************************************************/
typedef struct segAngle {
  int   nptos;
  byte  **offset;
  int   suma;
} TypeSegAngle;

typedef TypeSegAngle TypePicSegAngle[MAX_SEGS];

#ifdef _DEBUG
void  SigueContorno(  byte* punto, struct segAngle *pSeg,
                                         byte **pInicioSegmento );
#else
/* faster version */
__inline void __fastcall SigueContorno(  byte* punto, struct segAngle *pSeg,
                                         byte **pInicioSegmento );
#endif

static int SolucToTita;

/* Use a once-calculated offset: 
 * vecino = vecino_x + vecino_Y*columnas;
 */
#define _N_VECINOS_1_ 10
static int vecino[2][_N_VECINOS_1_];


/*********************************************************************************/

void ANGLE_UMBRALIZA_HISTERESIS ( byte *pic, byte *soluc, byte *tita,
                                 byte **almac, byte* final_dst,
			                           int threshold,int min_suma, int min_len,
                                 int columnas,int filas)
{
/* final_dst es una zona de memoria de dimensin filas*columnas
 * en donde se guardar la imagen. De los punteros que se pasan,
 * puede ser soluc (lo ms normal) o tita. soluc es una zona de memoria
 * que se utiliza
 * como buffer intermedio, pero despus de efectuar el algoritmo por
 * toda la imagen, el resultado se guarda en el puntero final_dst, que
 * por lo general coincidir con soluc, pero que puede ser otra zona
 * de memoria y de esta forma se puede ahorrar tener que copiar
 * la imagen (es un parmetro aadido por eficiencia.
 */
 int x, y, nseg=0, SolucToPic = pic-soluc, SolucToFinalDst = final_dst - soluc;
 byte **pInicioSegmento;
 TypePicSegAngle seg;

 /* Inicializacin de datos */
 pInicioSegmento=almac;
 SolucToTita = tita-soluc;
 inicializa_vecinos(filas,columnas);
 
 //Copy original pic in solution 
 memcpy(soluc,pic,columnas*filas);

 for (y=(filas-1)*columnas;  y -= columnas; ) {   
   register byte *punto;
   if (nseg == MAX_SEGS) break;
   x=columnas-1;
   punto = soluc+x+y;
   for (; --x ;) {
     punto--;
     if ( *punto > threshold ) {
        SigueContorno( punto, &seg[nseg], pInicioSegmento);
          /* Update pointers depending if it's segment */
        if ( (seg[nseg].suma > min_suma) && (seg[nseg].nptos > min_len) ) {
          /* Actualiza el estado del almacn */
          pInicioSegmento += seg[nseg].nptos;
          nseg++;
          /* Don't excess nseg limits */
          if (nseg == MAX_SEGS) break;
        } else {
          int npto = seg[nseg].nptos;
          /* CLEAR NON-MARKED SEGMENT */
          for(;npto;) {
            npto--;
			/* erase points in segment analized to
			 * avoid process again 
			 */
            *pInicioSegmento[npto] = 0;
          }
        }
	   }
   }
 }

  /* Clear solution (to erase lonely points with [1..threshold] values) */  
  memset(final_dst, 0, filas*columnas);

  /* pointer to orig point in pic: (*pptr+OffsetSolToPic)
   * value of orig point in pic  : *(*pptr+OffsetSolToPic)
   * pointer to dest point       : *pptr or *pptr+SolucToFinalDs
   * value of point to be marked : **pptr or *(*pptr+SolucToFinalDst)
   */

  /* Copy segments detected to solution */
  if (final_dst == soluc)
	  for ( ; nseg ; ) {
		  nseg--;	{
		    int nptos = seg[nseg].nptos;
		    byte **pptr= seg[nseg].offset;
		    for ( ; nptos ;nptos--) {
			    /* a little optimization */
		      **pptr = *(*pptr+SolucToPic);
          pptr++;
		    }
		  }
	  }
  else
	  for ( ; nseg ; ) {
		  nseg--; {
		    int nptos = seg[nseg].nptos;
		    byte **pptr= seg[nseg].offset;
  		  for ( ; nptos ;nptos--) {
			    *(*pptr+SolucToFinalDst) = *(*pptr+SolucToPic);
          pptr++;
		    }
		  }
	  }
}   /* end of ANGLE_UMBRALIZA_HISTERESIS */

/*********************************************************************************/

#ifdef _DEBUG
void  SigueContorno(  byte* punto, struct segAngle *pSeg,
                                         byte **pInicioSegmento )
#else
/* faster version */
__inline void __fastcall SigueContorno(  byte* punto, struct segAngle *pSeg,
                                         byte **pInicioSegmento )
#endif
{
  byte **pPtoEstudiado=pInicioSegmento, **pFinSegmento=pInicioSegmento;
  /* Sentido = 1  2, en cual se buscan puntos contiguos en las dos posibles
   * direcciones (es decir, a la izda y derecha del punto) */ 
  int sentido=2, suma=*punto;
  *punto=0;
  *pFinSegmento=punto;

  /* condiciones necesarias aadir un nuevo punto a uno considerado en el
   * segmento:
   * - que sean puntos contiguos y estn aliniados en direccin perpendicular
   *   a la direccin del gradiente (hay 2 direcciones perpendiculares de sen-
   *   tido contrario -> hay 2 segmentos ptos1 y ptos2 a partir de una semilla).
   * - que la direccin del gradiente se diferencie en menos de HOLGURA_ANG
   *   grados en rango 0..256.
   */
  do {
    /* actualizo los vecinos que tomo para la segunda pasada (sentido=0) */
    int* pVecinos = (--sentido) ? vecino[0] : vecino[1];
    /* Al inicio de cada bucle de sentido, pFinSegmento debe ser una unidad
     * mayor que pPtoEstudiado */
    pFinSegmento++;
    do {
      byte tita0 = *(SolucToTita + *pPtoEstudiado);
      /* aumento un poco tita para que los dos vecinos estudiados estn ms
       * centrados respecto la direccin del gradiente */
      int *vecino_tita = pVecinos + ( (tita0+(byte)8) >> 5);
      byte* pto = *pPtoEstudiado + *vecino_tita;
      /* divido la circunferencia en 8 partes con ocho vecinos distintos */
      /* 0..256 -> 0..8 . Por tanto divido por 32, o equivalentemente >>5*/
      if ( *pto ) { /* first try */
        byte distancia =tita0 - *(SolucToTita + pto);
        if ( distancia < HOLGURA_ANG || distancia >(256-HOLGURA_ANG) ) {
          suma += *pto; *pto=0; 
          *pFinSegmento=pto; pFinSegmento++;
        }
      
      } else { /* second try */
        vecino_tita++;
        pto = *pPtoEstudiado + *vecino_tita;
        if ( *pto ) {
          byte distancia = tita0 - *(SolucToTita + pto);
          if ( distancia < HOLGURA_ANG || distancia >(256-HOLGURA_ANG) ) {
            suma += *pto; *pto=0; 
            *pFinSegmento=pto; pFinSegmento++;
          }
        }
      }
      pPtoEstudiado++;
    /* Se termina el bucle cuando pFinSegmento == pPtoEstudiado; */
    } while (pPtoEstudiado<pFinSegmento);
  
    /* Incluimos el punto de inicio de la busqueda en el semisegmento de
     * sentido contrario (dos veces)
     * para simplificar el bucle (mismo comportamiento para cada sentido)
     * En realidad el punto de inicio se incluye dos veces en el segmento
     */
    *pFinSegmento = *pInicioSegmento;
  } while (sentido);
  /* Guarda el segmento en la estructura*/
  pSeg->nptos = pPtoEstudiado - pInicioSegmento;
  pSeg->offset = pInicioSegmento;
  /* Guarda la suma total */
  pSeg->suma=suma;
}

/*********************************************************************************/
#define _OLD_ANGLES_
__inline void __fastcall inicializa_vecinos(int filas, int columnas)
{
   int *pVecino=vecino[0];
#ifdef _OLD_ANGLES_
   /* vecino 0*/
   *pVecino = columnas;        
   pVecino++;
   /* vecino 1*/
   *pVecino = columnas-1;        
   pVecino++;
   /* vecino 2*/
   *pVecino = -1;        
   pVecino++;
   /* vecino 3*/
   *pVecino = -1-columnas;        
   pVecino++;
   /* vecino 4*/
   *pVecino=  -columnas;        
   pVecino++;
   /* vecino 5*/
   *pVecino = 1-columnas;        
   pVecino++;
   /* vecino 6*/
   *pVecino = 1;        
   pVecino++;
   /* vecino 7*/
   *pVecino = 1+columnas;        
   pVecino++;
   /* vecino 8 == vecino 0 (utilizado para tener puntos contiguos) */
   *pVecino = columnas;
   pVecino++;
   /* vecino 9 ==1*/
   *pVecino = columnas-1;        
#else  //_OLD_ANGLES_
   /* Calcula los vecinos, que dependen de la anchura de la
    * imagen */
   vecino[0][0] =    columnas;        
   vecino[0][1] =  1+columnas;        
   vecino[0][2] =  1;        
   vecino[0][3] =  1-columnas;        
   vecino[0][4] =   -columnas;        
   vecino[0][5] = -1-columnas;        
   vecino[0][6] = -1;        
   vecino[0][7] = -1+columnas;        
   vecino[0][8] =    columnas;        
   vecino[0][9] =  1+columnas;        
#endif  //_OLD_ANGLES_

   /* Direcciones contrarias */
   {
     int i=_N_VECINOS_1_;
     for (;i;) {
       i--;
      vecino[1][i] = -vecino[0][i];
     }
   }
}

/*********************************************************************************/

void FAST_UMBRALIZA_HISTERESIS ( byte *pic, byte *soluc, byte **almac,
			                           int min_media,int threshold, int min_len,
                                 int columnas,int filas)
{
 int x, y, min_suma = min_media*min_len;
 byte *punto, **pPtoEstudiado, **pFinAlmac, **pSegmentoAnterior;
 
 /* Inicializacin de datos */
 pSegmentoAnterior=almac;
 
 //Copy original pic in solution 
 memcpy(soluc,pic,columnas*filas);

 for (y=(filas-1)*columnas; y -= columnas; ) {
   x=columnas-1;
   punto = soluc+x+y;
   for (;--x;) {
     punto--;
     if ( *punto > threshold ) {
        register byte* ptoVecino;
        int suma=*punto;
        *punto=0;
        pFinAlmac=pPtoEstudiado=pSegmentoAnterior;
	      *pFinAlmac=punto;
        pFinAlmac++;

        while (pPtoEstudiado<pFinAlmac) {
          ptoVecino = *pPtoEstudiado + 1;
          /* (offset+1) ;          Vecino 0 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          } 
          ptoVecino -= columnas;      /* (offset+1-columnas) ; Vecino 1 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          ptoVecino--;                /* (offset  -columnas) ; Vecino 2 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          ptoVecino--;                /* (offset-1-columnas) ; Vecino 3 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          ptoVecino += columnas;      /* (offset-1)          ; Vecino 4 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          ptoVecino += columnas;      /* (offset-1+columnas) ; Vecino 5 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          ptoVecino++;                /* (offset  +columnas) ; Vecino 6 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          ptoVecino++;                /* (offset+1+columnas) ; Vecino 7 */
          if ( *ptoVecino ) { 
            *pFinAlmac=ptoVecino; pFinAlmac++;
            suma += *ptoVecino; *ptoVecino=0; 
          }
          pPtoEstudiado++;
        }
        TRACE("Elemento extraido \n");
        /* Update pointers depending if it's segment */
        {
          int len = (pFinAlmac-pSegmentoAnterior); 
	        /*if ((len > min_len) && (suma/len > min_media))*/ 
          if (suma > min_suma)
            /* Acept segment */
            pSegmentoAnterior=pFinAlmac;
        }
	   }
   }
 }

  /* Clear solution (to erase lonely points with [1..threshold] values) */  
  memset((void*) soluc, 0, filas*columnas);

  /* Copy segments detected to solution */
  {
    int OffsetSolToPic = pic-soluc;
    int contador = pFinAlmac-almac;
    for ( ; contador ; ) {
      contador--;
      pFinAlmac--;
      /* pointer to orig point in pic: (*pFinAlmac+OffsetSolToPic)
       * value of orig point in pic  : *(*pFinAlmac+OffsetSolToPic)
       * pointer to dest point       : *pFinAlmac
       * value of point to be marked : **pFinAlmac
       */
      **pFinAlmac=*(*pFinAlmac+OffsetSolToPic);
    }
  }
}   /* end of FAST_UMBRALIZA_HISTERESIS */

/*********************************************************************************/

void ERASE_LONELY_POINTS_OLD ( byte *pic, int columnas,int filas)
{
 int x,y;
 register byte *punto, *offset;
 
 for (y=(filas-1)*columnas; y -= columnas; ) {
   x=columnas-1;
   punto = pic+x+y;
   for (;--x;) {
     punto--;
     if ( *punto ) {
        offset = punto+1;        /* (offset+1) ;          Vecino 0 */
        if ( *offset ) continue; 
        offset += columnas;      /* (offset+1+columnas) ; Vecino 1 */
        if ( *offset ) continue; 
        offset--;                /* (offset  +columnas) ; Vecino 2 */
        if ( *offset ) continue; 
        offset--;                /* (offset-1+columnas) ; Vecino 3 */
        if ( *offset ) continue; 
        offset -= columnas;      /* (offset-1)          ; Vecino 4 */
        if ( *offset ) continue; 
        offset -= columnas;      /* (offset-1-columnas) ; Vecino 5 */
        if ( *offset ) continue; 
        offset++;                /* (offset  -columnas) ; Vecino 6 */
        if ( *offset ) continue; 
        offset++;                /* (offset+1-columnas) ; Vecino 7 */
        if ( *offset ) continue; 
        *punto=0;
     }
   }
 }
}

/*********************************************************************************/

void ERASE_LONELY_POINTS ( byte *pic, byte *tita, int columnas,int filas)
{
 int x,y, SolucToTita= tita-pic;
 register byte *punto;
 
inicializa_vecinos( filas, columnas);

  /* 
   *   borra los puntos contorno que no tengan ningn otro punto contorno
   *   contiguo y aliniado en direccin perpendicular a la direccin del gradiente
   *   (hay 2 direcciones perpendiculares de sentido contrario -> 
   *   hay 2 segmentos ptos1 y ptos2 a partir de una semilla).
   *   Efecto: borra los puntos contorno aislado que provienen del ruido
   */

 for (y=(filas-1)*columnas; y -= columnas; ) {
   x=columnas-1;
   punto = pic+x+y;
   for (;--x;) {
     punto--;
     if ( *punto ) {
        /* aumento un poco tita para que los dos vecinos estudiados estn ms
         * centrados respecto la direccin del gradiente */
        byte nvecino = (*(SolucToTita + punto)+8) >> 5;
        /* Primer vecino */
        if ( punto[vecino[0][nvecino]] ) continue; 
        /* Segundo vecino */;
        if (  punto[vecino[1][nvecino]]  ) continue; 
        /* Tercer vecino */;
        if (  punto[vecino[0][nvecino+1]]  ) continue; 
        /* Tercer vecino */;
        if (  punto[vecino[1][nvecino+1]]  ) continue; 
        *punto=0;
     }
   }
 }
}