/*=====================================================================*/
/*                      RUTINAS DE VISION POR ORDENADOR                */
/*                      POR MIGUEL ANGEL MARTINEZ ESTEBAN              */
/*=====================================================================*/

/***********************************************************************/
/* Modificado y ampliado por Joaqun Mur ( abril 1997)                 */
/* jmur@robot3.cps.unizar.es                                           */
/***********************************************************************/

#include <io.h>
#include <stdio.h>
#include <math.h>
#include <memory.h>

#define IMPLEMENTATION_MIGUELAN  
#include "miguelan.h"

#define BLOCK_SIZE 4096

/*=====================================================================*/
/*                      RUTINAS DE TRABAJO IMAGENES                    */
/*=====================================================================*/

/******************************************************************************/
/* GRABA UNA IMAGEN EN DISCO EN FORMA BINARIA (tipo RAW).
 * Devuelve 0 si no ha habido fallos. */
int graba_imagen_raw(byte *picture, unsigned long tamano, char nombre[])
{
 FILE * input_file;
 int ngrabados, nbytes;

 /* Abrir fichero solo ESCRITURA */ 
 if ((input_file = fopen( nombre,"wb" )) == (FILE*) -1) {
   return -1;
 }

 /* Escribe el fichero */
  do {
     nbytes = MIN(tamano,BLOCK_SIZE);   
     ngrabados = fwrite((void *) picture, 1 , nbytes ,input_file);
     if ( ngrabados != nbytes ) {
            fclose( input_file );
            return -2;
     }
     /* update pointer to next value */
     picture += ngrabados;
     tamano -= ngrabados;
  } while (tamano>0);
 
/* Cierra el fichero de lectura */
 return (fclose(input_file));
}

/******************************************************************************/
/* LEE UNA IMAGEN DEL DISCO EN FORMA BINARIA (tipo RAW). 
 * Devuelve el nmero de bytes ledos o un valor negativo en caso de error.*/
int lee_imagen_raw(unsigned char *picture, unsigned long tamano,char nombre[])
{
 int bytes_readed;
 FILE * input_file;

 /* Abrir fichero solo ESCRITURA */ 
 if ((input_file = fopen( nombre,"wb" )) == (FILE*) -1) {
   return -1;
 }

 /* Lee el fichero */
 bytes_readed = fread((void *) picture, 1 , tamano,input_file);
 
/* Cierra el fichero de lectura y vuelve */
 fclose( input_file );
 return bytes_readed;
}

/******************************************************************************/
/* GRABA UNA IMAGEN EN DISCO EN FORMA DE TEXTO ASCII.
 * Devuelve 0 si no ha habido fallos. */
int graba_imagen_txt(unsigned char *picture,
		      int columnas,int filas,
		      char nombre[])            /* GRABA UNA IMAGEN EN DISCO   */
{
 FILE *salida;
 unsigned char color;
 int i,j;

 if ( (salida=fopen(nombre,"wb"))== (FILE*) -1)
   return -1;
 fprintf(salida,"A=[");
 for (i=0;i<filas;i++)
   {
    for (j=0 ; j<columnas ; j++)
       {
	color=*(picture+j+i*columnas);
	fprintf(salida,"%d",color);
	if (j != columnas) fprintf(salida,"\t");
       }
    if (i<filas-1) fprintf(salida,"\n");
   }
 fprintf(salida,"];");
 return fclose(salida);
}

/******************************************************************************/

void ConvFloatToByte(float *original,unsigned char *copia,int columnas,int filas)
{
  
    register long i;
    register unsigned char *byte_ptr=copia;
    register float *flt_ptr=original;

    /* convert float to byte*/
    for(i=(long) filas * (long) columnas ; i; i--, flt_ptr++) 
      *byte_ptr++=  ( *flt_ptr > 255.0) ? 255 :
                    ( *flt_ptr <   0.0) ?   0 :
                    (unsigned char)  *flt_ptr  ;
}

/******************************************************************************/

void ConvByteToFloat(unsigned char *original,float *copia,int columnas,int filas)
{
    register long i;
    register unsigned char *byte_ptr=original;
    register float *flt_ptr=copia;

    /* convert byte to float */
    for(i=(long) filas * (long) columnas ; i; i--) 
      *flt_ptr++ = *byte_ptr++;
}

/******************************************************************************/

void histograma(unsigned char *pt_imagen,
		unsigned long tamano,
		char file_out[])            /* CALCULA EL HISTOGRAMA DE UNA IMAGEN*/
{
 FILE *salida;
 int color_num,i;
 unsigned long indice;
 unsigned long vector_his[256];


 for (i=0;i<256;i++)
   vector_his[i]=0;
 for (indice=0;indice<(tamano);indice++)
   {
    color_num=*(pt_imagen+indice);
    vector_his[color_num]+=1;
   }

   salida=fopen(file_out,"wb");
 for (i=0;i<256;i++)
   fprintf(salida,"%ld\n",vector_his[i]);
 fclose(salida);

}

/******************************************************************************/

void superpone(unsigned char *original,
	  unsigned char *mascara,unsigned long tamano)
{
 unsigned long contador;
 for (contador=0;contador<tamano;contador++)
    if (*(mascara+contador)!=0)
       *(mascara+contador)=*(original+contador);
}

/******************************************************************************/

unsigned char * busca_memoria(unsigned long espacio)
{
 unsigned char *posicion;

 /* Tamao de la imagen en BYTES y busqueda del espacio en memoria */
 if ((posicion=(char *)malloc(espacio))==NULL) {
   TRACE("Fuera de memoria.\n");
   exit(0);
 }
 return posicion;
}

/******************************************************************************/

float* busca_memoria_Float(unsigned long espacio)
{
 float *posicion;

 /* Tamao de la imagen en BYTES y busqueda del espacio en memoria */
 if ((posicion=(float *)malloc(espacio*sizeof(float)))==NULL)
 {
    TRACE("Fuera de memoria.\n");
    exit(0);
 }
 return posicion;
}

/*********************************************************************************/

void CopySpacedIntsToBuffer ( int* src, int* dst, int total, int space)
{
   for(;total;total--) {
     *dst=*src;
     dst++;
     src+=space;
   }
}

/******************************************************************************/

void CopyBufferToSpacedInts ( int* src, int* dst, int total, int space)
{
   for(;total;total--) {
     *dst=*src;
     src++;
     dst+=space;
   }
}

/******************************************************************************/
/* Get the maximum value of an array or matrix of dimension elements */
int MaximumInt (int *src, int contador)
{
   register int *ptr=src, scale=*src;
   register int  i=contador; 
   
   /* Normalize results */
   for (; (--i); ){ /* compute max value */
     ptr++;
     if (*ptr>scale) scale = *ptr;
   }   
   return scale;
}

/******************************************************************************/

void FitIntInByte(int *src, byte *dst, int contador, float factor)
{
   register int *ptr=src;
   register int  i;
   int max=*src, scale;
   
   /* Normalize results */
   for (i=contador; i; i--, ptr++) {/* compute max value */
       if (*ptr>max) max = *ptr; }     
   
   scale = (int) (max / ( 255*factor));
  
   if (scale==0) {
       for (i=contador; i; i--)
         *dst++=0;
       return;
   }
  
  ptr=src;
  
  i=contador;
  if (factor > (float) 1.0) {
    for (; i;) {
      i--;
      dst[i] =  (*ptr<=0) ? 0 : ( *ptr < max) ? (byte) (*ptr/scale) : 255;
      ptr++;
    }
  } else
    for(;i;) {
      i--;
      /* Optimized to speed dividing ceros by scale */
      dst[i] = (ptr[i]) ? (byte) (ptr[i]/scale) : 0;
    }
}

/*********************************************************************************/

/******************************************************************************/
/* Get the maximum value of an array or matrix of dimension elements */
float MaximumFlt (float *src, int contador)
{
   register float *ptr=src, scale=*src;
   register int  i=contador; 
   
   /* Normalize results */
   for (; (--i); ){ /* compute max value */
     ptr++;
     if (*ptr>scale) scale = *ptr;
   }   
   return scale;
}

/******************************************************************************/

void FitFltInByte(float *src, byte *dst, int contador, float factor)
{
   float scale=*src;
   
   /* Normalize results */
   {
     register float *ptr=src;
     register int i=contador;
     for (; i; i--, ptr++) {/* compute max value */
         if (*ptr>scale) scale = *ptr; }     
   }
   scale /= ( (float) 255.0*factor);
  
   if (scale==0) {
       register int i=contador;
       for (; i; i--)
         *dst++=0;
       return;
   }
  
  if (factor > (float) 1.0) {
    register float num;
    register int i=contador;
    float max=scale*(float)254.5;
    for (; i;) {
      i--;
      num=src[i];
      dst[i] =  (num<=0) ? 0 : ( num < max) ? (byte) (num/scale) : 255;
    }
  } else {
    register int i=contador;
    for(;i;) {
      i--;
      /* Optimized to speed dividing ceros by scale */
      dst[i] = (src[i]>0) ? (byte) (src[i]/scale) : 0;
    }
  }
}

/*********************************************************************************/

void FitFltInByteThreshold(float *src, byte *dst, int contador, float factor,
                           byte threshold)
{
   float min,scale=*src;
   
   /* Normalize results */
   {
     register float *ptr=src;
     register int i=contador;
     for (; i; i--, ptr++) {/* compute max value */
         if (*ptr>scale) scale = *ptr; }     
   }
   scale /= ( (float) 255.0*factor);
  
   if (scale==0) {
       register int i=contador;
       for (; i; i--)
         *dst++=0;
       return;
   }

   min = scale*((float)threshold);
    
 if (factor > (float) 1.0) {
    register float num;
    register int i=contador;
    float max=scale*(float)254.5;
    for (; i;) {
      i--;
      num=src[i];
      dst[i] =  (num<=min) ? 0 : ( num < max) ? (byte) (num/scale) : 255;
    }
  } else {
    register int i=contador;
    for(;i;) {
      i--;
      /* Optimized to speed dividing ceros by scale */
      dst[i] = (src[i]>min) ? (byte) (src[i]/scale) : 0;
    }
  }
}

/*********************************************************************************/

void ConvByteToInt(byte *src, int *dst,int contador)
{
  for(;contador;){
    contador--;
    dst[contador]=src[contador];
  }
}

/*********************************************************************************/

void ConvIntToByte(int *src, byte *dst, int contador, int factor)
{
  if ( factor>1 ){
    int max = 255*factor;
    for(;contador;contador--) {
      *dst =  (*src<=(int)0) ? 0 : ( *src < max) ? (byte)((int) (*src/factor)) : 255;
      src++; dst++;
    }
  } else 
    for(;contador;contador--) {
      *dst= (*src>=(int)255) ? 255 : (*src>(int)0) ? ((byte) (*src)) : 0;
      src++; dst++;
    }
}

/******************************************************************************/
/******************************************************************************/

/* Static variables for FastBlur */
static int longitud_buffer;
static int longitud_ventana;
static int *fila_original, *fila_destino;

/* Functions that use this variables */
void FastBlurRow ( void );
void FastBlurRowWindow2 ( void );

/******************************************************************************/
void FastBlurGrey( int *image, int columns, int rows, int n)
{
  /*       U L T R A     F A S T !!!                                   */
  /* convolves with an n*n array, consisting of only 1's.  
     Operates on rectangular region 'selx,sely,selw,selh' (in pic coords)
     Region is guaranteed to be completely within pic boundaries
     'n' can be odd or even.
     2*columns*rows sums, rows+columns multiplications
     The image contained in aux is overwritten.
  */

  int *buffer1,*buffer2;    /* Auxiliar image for separating components */
  int index;
  void (*RowProcess)( void );
  /* In case longitud_ventana == 2, use special filter to avoid displacements */
  if ( n == 2) 
      RowProcess=FastBlurRowWindow2;
  else
      RowProcess=FastBlurRow;
  
  /* MEMORY ALLOCATION */
  buffer1 = (int*) malloc (MAX(rows,columns)*sizeof(int));
  if (!buffer1) return;
  
  buffer2 = (int*) malloc (MAX(rows,columns)*sizeof(int));
  if (!buffer2) { free (buffer1); return; }
  
  /* SET FastBlurRow static variables */
  longitud_buffer = columns;
  longitud_ventana = n;
  fila_destino=buffer2;
  
  /* BLUR X (calculated with a "mirror image" beyond the border )*/
  for ( index=rows*columns ; index ;) {     
     
     index-=columns;
     
     /* Blur row */
     fila_original=image+index;
     RowProcess ();
 
     /* overwrite row image with buffer content */
     memcpy ((byte*)fila_original,(byte*)buffer2,columns*sizeof(int));
  }

  /* SET FastBlurRow static variables */
  longitud_buffer = rows;
  fila_original=buffer1;
  
  /* BLUR Y (calculated with a "mirror image" beyond the border )*/
  for ( index=columns ; index ; ) {
     
     /* copy row to buffer */
     index--;
     CopySpacedIntsToBuffer ( image+index, buffer1, rows, columns);

     /* Blur row */
     RowProcess();

     /* overwrite row image with buffer content */
     CopyBufferToSpacedInts ( buffer2, image+index, rows, columns);    

  }

  /* FREE MEMORY */
  free(buffer1);
  free(buffer2);
}

/******************************************************************************/

void FastBlur24( int *image, int columns, int rows,int n)
{
  /* convolves with an n*n array, consisting of only 1's.  
     Operates on rectangular region 'selx,sely,selw,selh' (in pic coords)
     Region is guaranteed to be completely within pic boundaries
     'n' can be odd or even.
     2*columns*rows sums, rows+columns multiplications
     The image contained in aux is overwritten.
  */
  int *buffer1,*buffer2;    /* Auxiliar image for separating components */
  int index, IntsPerRow=columns*3,color;
  void (*RowProcess)( void );
  
  /* In case longitud_ventana == 2, use special filter to avoid displacements */
  if (n == 2) 
      RowProcess=FastBlurRowWindow2;
  else
      RowProcess=FastBlurRow;

  /* MEMORY ALLOCATION */
  buffer1 = (int*) malloc (MAX(rows,columns)*sizeof(int));
  if (!buffer1) return;
  
  buffer2 = (int*) malloc (MAX(rows,columns)*sizeof(int));
  if (!buffer2) { free (buffer1); return; }
  
  
  /* SET FastBlurRow static variables */
  fila_original=buffer1;
  fila_destino=buffer2;
  longitud_buffer = columns;
  longitud_ventana = n;
  
  /* BLUR X (calculated with a "mirror image" beyond the border )*/
  for ( index=rows*IntsPerRow ; index ;){
    index-=IntsPerRow;
    for ( color=0; color<3; color++) {     
     
     /* copy row to buffer */     
     CopySpacedIntsToBuffer ( image+index+color, buffer1, columns, 3 );

     /* Blur row */
     RowProcess();

     /* overwrite row image with buffer content */
     CopyBufferToSpacedInts ( buffer2, image+index+color, columns, 3 );    
    }
  }
  
  /* BLUR Y (calculated with a "mirror image" beyond the border )*/
  longitud_buffer = rows;
  for ( index=IntsPerRow ; index ; ) {
     
     index--;
     /* copy row to buffer */
     CopySpacedIntsToBuffer ( image+index, buffer1, rows, IntsPerRow);

     /* Blur row */
     RowProcess();

     /* overwrite row image with buffer content */
     CopyBufferToSpacedInts ( buffer2, image+index, rows, IntsPerRow);    

  }

  /* FREE MEMORY */
  free(buffer1);
  free(buffer2);
}

/******************************************************************************/

void FastBlurRow ( void )
{
  int *dst,*viejo,*nuevo;    
  register int valor,indice;

   /* starting value */
   /* si es impar, toma un punto ms */
   viejo=fila_original + longitud_buffer - 1;
   valor = *viejo;
   viejo--;
   for ( indice = ((longitud_ventana-1)>>1); indice ; indice--,viejo--)
     valor += *viejo;
   /* efecto espejo en el borde -> punto en la imagen y reflejado */
   valor <<= 1; 
   /* si la anchura del filtro es impar, el punto fronterizo no se refleja */
   if (longitud_ventana&1) valor -= fila_original[longitud_buffer-1];
   /* fijo el primer punto de la derecha */
   fila_destino[longitud_buffer-1] = valor;
   
   /* right border */
   viejo = fila_original + longitud_buffer - 1 - (longitud_ventana&1);
   nuevo = fila_original + longitud_buffer - longitud_ventana;
   dst = fila_destino + longitud_buffer - (longitud_ventana>>1);
   for (indice = ((longitud_ventana)>>1) ; indice>0 ; ) {
     indice--;
     valor += nuevo[indice]-viejo[-indice];
     dst[indice]=valor;
   }

   /* middle */
   viejo = fila_original + longitud_ventana-1;
   nuevo = fila_original - 1;
   dst   = fila_destino + ((longitud_ventana-1)>>1);
   for (indice = longitud_buffer-longitud_ventana; indice ; indice--) {
     valor += nuevo[indice] - viejo[indice];
     dst[indice]=valor;
   }

   /* left value */
   nuevo=fila_original+(longitud_ventana&1);
   indice=((longitud_ventana-1)>>1)+1;
   viejo=fila_original+longitud_ventana-indice;
   for (; indice ;) {
     indice--;
     valor += *nuevo - viejo[indice];
     fila_destino[indice]=valor;
     nuevo++;
   } 
} /* End of FastBlurRow */


/******************************************************************************/

void FastBlurRowWindow2 ( void )
{
/* This function avoids the displacement that a blur of window 2 causes because
 * it is a odd number (the center is in the middle of two points. The solution
 * in case window is 2, is use half of right point, half of left point respect
 * center point. To increase speed, we take right point, left point and twice
 * central point, so the result has a factor of 4*WindowWidth
 */

  int *dst,*pto_centro,*pto_izda;    
  register int valor,indice;

   /* starting value */
   indice = longitud_buffer;
   indice--;
   pto_izda = pto_centro = fila_original + indice;
   pto_izda--;
   dst = fila_destino + indice;
   *dst = ((*pto_centro + *pto_izda)<<1);
   dst--;
   
   /* middle */
   for (; (--indice) ;) {
     valor = *pto_centro;
     pto_centro=pto_izda;
     valor += ((*pto_centro)<<1);
     pto_izda--;
     valor += *pto_izda;
     *dst = valor;
     dst--;
   }

   /* Last point */
   *dst = ((*pto_centro + *pto_izda)<<1);

} /* End of FastBlurRowWindow2 */

/******************************************************************************/
