// dibdoc.cpp : implementation of the CDibDoc class
//

/* Proyecto Fin de Carrera
 * Departamento de Informatica e Ingeniera de Sistemas
 * Universidad de Zaragora
 *
 * rea:      Robtica y Automatizacin Avanzada.
 * Tema:      Visin por Computador.
 * Ttulo:    Diseo de un entorno de visin para 
 *            el tratamiento de imgenes.
 * Director:  Dr. Carlos Sags Blzquiz.
 * Autor:     Joaqun Mur Amada.
 * Promocin: 1991-1996.
*/

#include <limits.h>
#include "dibdoc.h"
#include "dibview.h"
#include "..\res\resource.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDibDoc

IMPLEMENT_DYNCREATE(CDibDoc, CDocument)

BEGIN_MESSAGE_MAP(CDibDoc, CDocument)
	//{{AFX_MSG_MAP(CDibDoc)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_EDIT_CUT, OnEditCut)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_COMMAND(ID_EDIT_UNDO, OnEditUndo)
	ON_COMMAND(ID_SAVE_MEMORY, OnSaveMemory)
 	ON_COMMAND(ID_FORMAT_BWDITHER, OnFormatBwdither)
	ON_COMMAND(ID_FORMAT_GREYSCALE, OnFormatGreyscale)
	ON_COMMAND(ID_FORMAT_REDUCED, OnFormatReduced)
	ON_COMMAND(ID_FORMAT_FULLCOLOR, OnFormatFullcolor)
	ON_UPDATE_COMMAND_UI(ID_SAVE_MEMORY, OnUpdateSaveMemory)
	ON_UPDATE_COMMAND_UI(ID_EDIT_UNDO, OnUpdateEditUndo)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CUT, OnUpdateEditCut)
	ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, OnUpdateEditCopy)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDibDoc construction/destruction

// Default: use palettes
int CDibDoc::UsePalette = (int) RC_PALETTE;



//---------------------------------------------------------------
// CDibDoc : : I n i t  D i b  D a t a
// Create a device dependent bitmap from the IOfile's pic on
// the currentdisplay device.
// This routine is used to accelerate the display speed.

void CDibDoc::InitDibData( BOOL bSetModifiedFlag, CDC* pDC )
{
	TRACE("\nCDibDoc::InitDibData\n");
  // Check document size
	if (GetHeight() > INT_MAX || GetWidth() > INT_MAX)
	{
		Warning ( "La imagen es demasiado grande o no est" );
		return;
	}

  /* Do only clear screen if no image is in memory */
  if (pic==NULL || GetHeight()==0 || GetWidth()==0) {
    SetHBITMAP(NULL);
    SetHPALETTE(NULL);
    m_redraw=TRUE;
    UpdateAllViews(NULL);
    NumCols=0;
    return;
  }
    
  HDIB hDIB = GetHDIB();
  if (hDIB==NULL) {
    Error("InitDibData:\n"
          "No puedo obtener una imagen dib del objeto IOfile\n"
          "(Pic est vacio, no se ha cargado la imagen o no hay memoria.)");
    //GlobalFree(hDIB);
    return;
  }
#ifdef _DEBUG
  MessageBeep((WORD) -1);
#endif //_DEBUG
  // Do a bitmap for such a display.
  SetHBITMAP(NULL);
  SetHPALETTE(NULL);
  SetHBITMAP( BITMAPfromHDIB(hDIB, pDC));

  // Release DIB
  GlobalFree(hDIB);

  // If the image has been changed since it was first open, mark it for ask for
  // save before closing it and update all views and redraw all views.
  if ( bSetModifiedFlag )
  {
      SetModifiedFlag(TRUE);
      m_redraw=TRUE;
      UpdateAllViews(NULL);
  }

  // Release DIB memory from pic2bmp
  GlobalFree(hDIB);
}

//---------------------------------------------------------------
// CDibDoc : : C r e a t e  H P A L E T T E
/* Function placed in Con24to8.cpp and Quant.cpp to computer colortable */

extern int quick_check_map(const byte *pic24, int w,int h, byte *rmap,
                       byte *gmap,byte *bmap, int maxcol);
extern int    slow_quant_map  (const byte *pic24, int w, int h, byte *rm,
                           byte *gm, byte *bm, int descols);

extern int    GetSlowMapFromPIC24(const byte *pic24,int w,int h,int nc,
                           byte *rm,byte *gm,byte *bm);

/* Auxiliar function */
void SortColormap(const byte *pic, int pwide, int phigh, u_int *pnumcols,
                  const byte *rmap, const byte *gmap,const byte *bmap, 
                  byte *order, byte *trans);

HPALETTE CDibDoc::CreateHPALETTE()
{
    TRACE("\nCDibView::CreateHPALETTE\n");
	
	// Don't do nothing if size is zero
	if (   pic== NULL || GetWidth() <= 0 || GetHeight() <= 0 ) return 0;

  // Try to catch palette from DIB
	if ( m_hBITMAP != NULL ) {
    LPBITMAPINFO pIMAGE = (LPBITMAPINFO) GlobalLock(m_hBITMAP);
	  if (!pIMAGE) {
      GlobalUnlock(m_hBITMAP);
		  return m_hPALETTE;
    }
	  if ( DIBNumColors((LPBITMAPINFOHEADER)pIMAGE) ) { 
		  HPALETTE hPALETTE = CreateBIPalette((LPBITMAPINFOHEADER) pIMAGE);
		  SetHPALETTE(hPALETTE);
		  GlobalUnlock(m_hBITMAP);
		  return m_hPALETTE;
		}
		GlobalUnlock(m_hBITMAP);
	}


  // Try to catch greyscale palette or images with a lot of colors
  // from colormapped IOfile image
  if ( IOfile::numcols>64 ) {
	  byte red[256],green[256],blue[256];
    byte table[256];
    byte trans[256];
    byte order[256];   /* order to allocate cols */

    u_int i, nc = IOfile::numcols;
    
    // copy original colormap ( to avoid changing IOfile's one )
    xvbcopy((char *)rMap, (char *)red, 256);
    xvbcopy((char *)gMap, (char *)green,256);      
    xvbcopy((char *)bMap, (char *)blue,256);

    // Calculate HPALETTE with IOfile image
    ::SortColormap(IOfile::pic,pWIDE,pHIGH,&nc,red,
          green,blue,order,trans);
    
    /* Copy ordered colormap in LOGPALETTE structure*/    
    for (i=0; i<nc; i++)
           table[order[i]]=i;
    for (i=0; i<256;i++)
        if(trans[i]!=i) table[i]=trans[i];      
    {
      size_t sizeLOGPALETTE = 4 + nc * sizeof(PALETTEENTRY);
      HGLOBAL hpal =GlobalAlloc(GHND,sizeLOGPALETTE);
      LPLOGPALETTE p = (LPLOGPALETTE)GlobalLock(hpal);
      p->palVersion    = PALVERSION; 
      p->palNumEntries = nc; 
      for (i=0; i<NumCols; i++) {
        p->palPalEntry[i].peRed   = red  [order[i]];
        p->palPalEntry[i].peGreen = green[order[i]];
        p->palPalEntry[i].peBlue  = blue [order[i]];
        p->palPalEntry[i].peFlags = 0;
      }
      SetHPALETTE( CreatePalette(p) );
      GlobalFree(hpal);
	    return m_hPALETTE;
    }
  }

  // Try to catch palette from colormapped IOfile image 
  // Images with few colors (less than 64 colors)
  // (an unshorted colormap is not so bad and faster! )
  if ( IOfile::numcols ) {
    // Copy HPALETTE from IOfile image
    size_t sizeLOGPALETTE = 4 + NumCols * sizeof(PALETTEENTRY);
    HGLOBAL hpal =GlobalAlloc(GHND,sizeLOGPALETTE);
    LPLOGPALETTE p = (LPLOGPALETTE)GlobalLock(hpal);
    p->palVersion    = PALVERSION; 
    p->palNumEntries = IOfile::numcols; 
    int i;
    for (i=0; i<NumCols; i++) {
        p->palPalEntry[i].peRed   = rMap[i];
        p->palPalEntry[i].peGreen = gMap[i];
        p->palPalEntry[i].peBlue  = bMap[i];
        p->palPalEntry[i].peFlags = 0;
    }
    SetHPALETTE( CreatePalette(p) );
    GlobalFree(hpal);
	  return m_hPALETTE;
  }

  // Compute palette from RGB image only in 256 colors displays!!
	if ( picType==PIC24 && UsePalette && IOfile::pic )	{
	  byte red[256],green[256],blue[256];
	  int nc = MAX_ALLOCABLE_COLS;
	  // Pointer to BITMAPINFOHEADER
    /* computes colormap of nc number of color in 
     * (red map) , (green map) and (blue map)
     * uses slow method */
  
     if (!quick_check_map(IOfile::pic, pWIDE, pHIGH,
                        red, green, blue, nc)) 
      if (slow_quant_map(IOfile::pic, pWIDE, pHIGH,
                        red, green, blue, nc))
       return m_hPALETTE; /* unable to compute colormap */
  

    size_t sizeLOGPALETTE = 4 + MAX_ALLOCABLE_COLS*sizeof(PALETTEENTRY);
    HGLOBAL hpal =GlobalAlloc(GHND,sizeLOGPALETTE);
    LPLOGPALETTE p = (LPLOGPALETTE)GlobalLock(hpal);
    p->palVersion    = PALVERSION; 
    p->palNumEntries = MAX_ALLOCABLE_COLS; 
    int i;
    for (i=0; i<MAX_ALLOCABLE_COLS; i++) {
        p->palPalEntry[i].peRed   = red  [i];
        p->palPalEntry[i].peGreen = green[i];
        p->palPalEntry[i].peBlue  = blue [i];
        p->palPalEntry[i].peFlags = 0;
    }
    SetHPALETTE( CreatePalette(p) );
    GlobalFree(hpal);
    return m_hPALETTE;
	} 

  /* if we are in this point, this function returns NULL palette */
  return m_hPALETTE /* ==NULL */;

}

//---------------------------------------------------------------
// CDibDoc : : I n s t a l l N e w P i c
void CDibDoc::RedrawPic (void)
{ // This function is called every time the dispayed function must be
  // updated. This fuction must be overriden in a viewer.
  InitDibData(TRUE);
}

//---------------------------------------------------------------
// CDibDoc : : N u m  C o l o r s 
u_int CDibDoc::NumColors(HPALETTE hPALETTE)
{
    if (m_hPALETTE == NULL)
        return 0;
    return ( GetPaletteEntries(m_hPALETTE,0,0,NULL));
}


//---------------------------------------------------------------
// CDibView : : G e t H B I T M A P 
// Return handle to member bitmap if exist. If the bitmap does not
// exist but there is a image no empty, constructs and returns one new.
HBITMAP CDibDoc::GetHBITMAP(CDC* pDC)
{
  if (m_hBITMAP == NULL && GetWidth()>0 && GetHeight()>0 && pic!=NULL)
        InitDibData(FALSE,pDC);
  TRACE("\nCDibView::GetHBITMAP\n");
  return m_hBITMAP;
}

//---------------------------------------------------------------
// CDibView : : S e t I M A G E  (from BITMAP)

int CDibDoc::SetIMAGE(HBITMAP hBITMAP)
{
    #ifdef _DEBUG
    TRACE("\nCDibView::SetIMAGE\n");
    #endif 
    HDIB hDIB = HDIBfromBITMAP ( hBITMAP );
    if (hDIB == NULL) {
      Error("CDibDoc::SetIMAGE( hBITMAP) ha fallado convirtiendo BITMAP a HDIB"
            "\nHDIBfromBITMAP ( hBITMAP )"); 
      return 0;
    }
    SetIMAGE( hDIB );
    ::GlobalFree((HGLOBAL) hDIB);   // Free DIB memory
    return 1;
}



//---------------------------------------------------------------
// CDibView : :   B I T M A P f r o m I M A G E
// Create a device dependent bitmap from the DIB on the current
// display device.
// This routine is used to accelerate the display speed.
// CCD pDC: DC to do output to
HBITMAP CDibDoc::BITMAPfromHDIB( HDIB hDIB, CDC* pDC)
{
    //Stuff for create a pDC if none is given
    BOOL FreeDC = FALSE;
    CView* pFirstView = NULL;

    // Set up a handle to IMAGE
    if ( /*pDC==NULL ||*/ hDIB == NULL)
        return m_hBITMAP;
    
    TRACE("\n****************\n"
          "\nCDibDoc::BITMAPfromHDIB"
          "\n************\n");

    // Create a BITMAP if there is a IMAGE!
    if ((hDIB == NULL) || (GetSize() == CSize(0,0))) {
      return NULL;
    }
    
    // Pointer to BITMAPINFOHEADER
    LPBITMAPINFO pIMAGE = (LPBITMAPINFO) GlobalLock(hDIB);
    
#ifdef _DEBUG
    if (pIMAGE == NULL)
    {   
        Error("BITMAPfromHDIB\nImagen vaca");
        GlobalUnlock(hDIB);
        return NULL;
    }
#endif  //_DEBUG

    //(pDC==NULL) -> get a device context from the firs view  
    // Get a DC if none is given and a palette should be used
    if ( pDC==NULL ) {
      POSITION pos = GetFirstViewPosition();
      if (pos!=NULL) {
        pFirstView = GetNextView( pos );
        if (pFirstView!=NULL) {
          FreeDC = TRUE;
          pDC = pFirstView->GetDC();
        }
      }
    }

    // Realize palette, if there is one. Note that this does not do
    // anything on the standard 16-color VGA driver because that
    // driver does not allow changing the palette, but the new palette
    // should work on Super VGA displays.
    // The function OnDoRealize handles the palettes. Realizing the
    // palette twice can lead to a circle of funtion calls.

    HPALETTE PalOld = NULL;    
    // If there is a DC, realize the new palette
    
    SetHPALETTE(NULL);

    HPALETTE hPALETTE = CreateBIPalette((LPBITMAPINFOHEADER) pIMAGE);
  #ifdef _DEBUG
	  if (UsePalette && hPALETTE == NULL && pIMAGE!=NULL && NumCols) {
		    // DIB must have a palette
        Warning("BITMAPfromHDIB:\n"
                "Imposible generar una paleta con imagen no nula.\n"
                "La aplicacin se puede volver inestable!.");
	  }  
  #endif //_DEBUG

    SetHPALETTE(hPALETTE);
    if( hPALETTE!=NULL && pDC != NULL) {
            PalOld = SelectPalette(pDC->GetSafeHdc(), hPALETTE, TRUE );
            RealizePalette(pDC->GetSafeHdc());
    }

  
    // Get a bitmap for the device context of the first CDibView if
    // none is given.
    // To get the first view in the list of views:

    // Pointer to DIB bits
    LPSTR lpDIBBits = ::FindDIBBits( (LPSTR) pIMAGE);

    // Build the device-dependent bitmap:    
    // Convert the DIB into a DDB (device dependent bitmap) and 
    // block transfer (blt) it to the device context.
    HBITMAP hBITMAP = CreateDIBitmap  ( pDC->GetSafeHdc(),
                             (LPBITMAPINFOHEADER) pIMAGE,
                             CBM_INIT,
                             lpDIBBits,
                             (LPBITMAPINFO) pIMAGE,
                             DIB_RGB_COLORS );

    // Don't need the palette once the bitmap is converted to BITMAP
    // format.
    if( UsePalette && PalOld != NULL && pDC!=NULL)
        SelectPalette(pDC->GetSafeHdc(), PalOld, TRUE );
    
    // Release the DC if it was created inside this function
    if ( FreeDC) 
        pFirstView->ReleaseDC(pDC);
  
    GlobalUnlock(hDIB);
    return hBITMAP;
}


//---------------------------------------------------------------
// CDibView : : D I B f r o m B I T M A P
// Create a device independent image from the displayed BITMAP
// specified by CDibView::m_hBITMAP and store in CDibDoc's IMAGE
HDIB CDibDoc::HDIBfromBITMAP( HBITMAP hBITMAP, CDC* pDC )
{
    BITMAP bm;
        
    // Do nothing if the BITMAP does not exist or if the IMAGE exists
    if( hBITMAP == NULL ) {
        Warning("CDibView::HDIBfromBITMAP\nEl bitmap est vacio");
        return NULL;
    }

    // OnDoRealize take care of upgrading the palete.
    // There is nothing to worry about because if there is no palette one is
    // automatic created when is needed.
    if (!GetObject( hBITMAP, sizeof(bm), (LPVOID) &bm)) {
        Warning("CDibView::HDIBfromBITMAP\nEl bitmap no es vlido");
        return NULL;
    }

// Compute bytes per line, rounding up to align at a 4-byte boundary
    long bytes_per_line = ((long)bm.bmWidth * 
                         (long)(bm.bmPlanes * bm.bmBitsPixel) + 31L) / 32 * 4;

// Determine number of colors in the palette
    short ncolors;
    switch(bm.bmPlanes * bm.bmBitsPixel)
    {
        case 1: 
            ncolors = 2;
            break;
        case 4: 
            ncolors = 16;
            break;
        case 8: 
            ncolors = 256;
            break;
        default: 
            ncolors = 0;
    }
    
// Compute total size of DIB
    long dibsize = (long) ( (long) sizeof(BITMAPINFOHEADER) +
                            (long) ncolors * (long) sizeof(RGBQUAD) +
                            (long) bm.bmWidth * bytes_per_line);
#ifdef _DEBUG
   char cadena[255];
   sprintf(cadena,"CDibView::HDIBfromBITMAP\n"
                "Tamao del dibujo = %lu + %lu * %lu + %lu = %lu",
                (long) sizeof(BITMAPINFOHEADER), (long)ncolors,
                (long)sizeof(RGBQUAD),
                (long) bm.bmWidth * bytes_per_line, dibsize);
   Warning(cadena);
#endif

// Allocate memory for the new IMAGE
    HDIB hNewIMAGE = (HDIB) ::GlobalAlloc(GHND, dibsize);
    if( hNewIMAGE == NULL) 
    {   
        char cadena[255];
        sprintf(cadena,"Fallo al requerir memoria en CDibView::SetIMAGEfromBitmap.\n"
              "No se copia el bitmap del display a la representacin interna de la IMAGEN.\n"
              "Memoria requerida: %lu",dibsize);
        Warning(cadena);
        ::GlobalFree((HGLOBAL)hNewIMAGE);
        return NULL;
    }

// Get a pointer to the IMAGE ( equal to the pointer to info header )
    LPBITMAPINFOHEADER pDIB = (LPBITMAPINFOHEADER)::GlobalLock ((HGLOBAL) hNewIMAGE );
    
//  Calculate the adress of the pointer to PALETTE
    LPPALETTEENTRY pPALETTE = (LPPALETTEENTRY) ((LPSTR) pDIB 
                               + sizeof(BITMAPINFOHEADER));

//  Calculate the adress of DIBBits pointer
    LPSTR pDIBBits = (LPSTR) pPALETTE + ncolors * sizeof(RGBQUAD);
    
// Copy BITMAPINFO structure bh into beginning of DIB.
// Set up a BITMAPINFOHEADER for the DIB ( IMAGE's real format)
    pDIB->biSize        = sizeof(BITMAPINFOHEADER);
    pDIB->biWidth       = bm.bmWidth;
    pDIB->biHeight      = bm.bmHeight;
    pDIB->biPlanes      = 1;
    pDIB->biBitCount    = (WORD)(bm.bmPlanes * bm.bmBitsPixel);
    pDIB->biCompression = BI_RGB;

    pDIB->biSizeImage   = (DWORD) bm.bmHeight * (DWORD) bytes_per_line;
    pDIB->biXPelsPerMeter  = 0;
    pDIB->biYPelsPerMeter  = 0;

    pDIB->biClrUsed        = ncolors;
    pDIB->biClrImportant   = 0;

// Copy PALETTEENTRY structure into lpDIB. Yet copied in GetDIBits.
  /*
   if (hPALETTE == NULL) hPALETTE = m_hPALETTE;  //Default
   if (hPALETTE != NULL)
     GetPaletteEntries(hPALETTE, 0, ncolors,pPALETTE);
   */

// Call GetDIBits to get the image and fill the palette indices
// into a BITMAPINFO structure. Also copies PALETTEENTRY structure
// into lpDIB.
   GetDIBits(pDC->GetSafeHdc(), hBITMAP, 0, (UINT) bm.bmHeight,
              pDIBBits, (LPBITMAPINFO) pDIB, DIB_RGB_COLORS);

// All done. Clean up and return.
    GlobalUnlock ((HGLOBAL) hNewIMAGE );
    return hNewIMAGE;    
}




/////////////////////////////////////////////////////////////////////////////
// CDibDoc diagnostics

#ifdef _DEBUG
void CDibDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CDibDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}

#endif //_DEBUG

//////////////////////////////////////////////////////////////////////////


void CDibDoc::OnSettings() 
{
  char buf[64]="0\0";
  InputDialog("Introduce el nivel de depurado (0 para no monstrar "
              "mensajes de comprobacin en la ventana OUTPUT).",
              buf,64,"0123456789",1
              );
  IOfile::DEBUG=atoi(buf);

}
void CDibDoc::OnSaveMemory() 
{
	// TODO: Add your command handler code here
  SaveMemory(!SaveMemory());	
}

void CDibDoc::OnUpdateSaveMemory(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
  pCmdUI->SetCheck(SaveMemory());
}

void CDibDoc::OnUpdateEditUndo(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
  pCmdUI->Enable(pic!=NULL);	
}

void CDibDoc::OnUpdateEditPaste(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
  pCmdUI->Enable(::IsClipboardFormatAvailable(CF_DIB) ||
           ::IsClipboardFormatAvailable(CF_BITMAP) );	
}

void CDibDoc::OnUpdateEditCut(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
  pCmdUI->Enable(pic!=NULL);	
}

void CDibDoc::OnUpdateEditCopy(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
  pCmdUI->Enable(pic!=NULL);	
}

/************************************************/
/* structure and routine used in SortColormap() */
/************************************************/

typedef struct thing {
  byte r,g,b, n;          /* actual value of color + alignment */
  int oldindex;           /* its index in the old colormap */
  int use;                /* # of pixels of this color */
  int mindist;            /* min distance to a selected color */
} CMAPENT;

/***********************************/

void SortColormap(const byte *pic, int pwide, int phigh, u_int *pnumcols,
                  const byte *rmap, const byte *gmap,const byte *bmap, 
                  byte *order, byte *trans)
{
  /* operates on 8-bit images.  sorts the colormap into 'best' order
   * 'order' is the 'best' order to allocate the colors.  'trans' is a 
   * transformation to be done to pic (in PIC8 mode) to compress the colormap
   */

  byte *p;
  u_int   i, j;
  int mdist, entry, d, hist[256];
  static CMAPENT c[256], c1[256], *cp, *cj, *ck;

  /* trans mappes the actual colormap in one without any repeated or unuded color 
     the old entry is the index, and the return number is the new equivalent entry
     new_entry_to_use = trans[old_entry] */


  /* init some stuff */
  for (i=0; i<256; i++) order[i]=i;

  /* initialize histogram and compute it */
  for (i=0; i<256; i++) hist[i]=0;
  for (i=pwide*phigh, p=(byte*)pic; i; i--, p++) hist[*p]++;  
  
  /* put the actually-used colors into the 'c' array in the order they occur
     also, while we're at it, calculate ncols, and close up gaps in
     colortable */
  
  /* we may have duplicate colors in the colormap, and we'd prefer not to.
  * order[0..numcols-1] contains the indices corresponding to the minimum colormap
  * set( r/g/bmap[order[0]] color must be the first color allocated). 
  * Trans array maps 'pic8' values (0-256) into corresponding non-repeated
  * colormap. 
  * To reflect the new colormap, note that order[updated_pix] entry and  
  * trans[old_pix] (both entrys in the old colormap) represent the same RGB value.
  */
  
  for (*pnumcols=i=0; i<256; i++) {
    if (hist[i]) {
      /* see if color #i is a duplicate used color */
      for (j=0; j<i; j++) { /* if color #j not used trans[j]==0 */
	      if ( trans[j] && rmap[i] == rmap[j] && gmap[i] == gmap[j] && 
	        bmap[i] == bmap[j]) break;
      }
      trans[i]    = (byte) j;
      if (j==i) {  /* unique color */
        cp = &c[*pnumcols];
        cp->r = rmap[i];    cp->g = gmap[i];    cp->b = bmap[i];
        cp->use = hist[i];  cp->oldindex = i;
        cp->mindist = 1000000; /* 255^2 * 3 = 195075 */
        (*pnumcols)++;
      }else{ /* color is repeated -> merge repeated entry in the first one*/
        /* i = repeated entry, j = entry first happened */
        hist[j] += hist[i];
        hist[i] = (byte) 0;
      }
    }
    else //(hist[i])
      trans[i]  =(byte) 0;
  }

  /* Update c with the histogram without repeated colors */
  for (i=0; i<*pnumcols; i++) {
      c[i].use = hist[c[i].oldindex];
  }

  /* find most-used color, put that in c1[0] */
  entry = -1;  mdist = -1;
  for (i=0; i<*pnumcols; i++) {
    if (c[i].use > mdist) { mdist = c[i].use;  entry=i; }
  }
  xvbcopy((char *) &c[entry], (char *) &c1[0], sizeof(CMAPENT));
  c[entry].use = 0;   /* dealt with */
  
  
  /* sort rest of colormap.  Half of the entries are allocated on the
     basis of distance from already allocated colors, and half on the
     basis of usage.  (NB: 'taxicab' distance is used throughout this file.)

     Mod:  pick first 10 colors based on maximum distance.  pick remaining
     colors half by distance and half by usage   -- JHB

     To obtain O(n^2) performance, we keep each unselected color
     (in c[], with use>0) marked with the minimum distance to any of
     the selected colors (in c1[]).  Each time we select a color, we
     can update the minimum distances in O(n) time. 

     mod by Tom Lane   Tom.Lane@g.gp.cs.cmu.edu */

  for (i=1; i<*pnumcols; i++) {
    int ckR, ckG, ckB;
    /* Get RGB of color last selected  and choose selection method */
    ck = &c1[i-1];            /* point to just-selected color */
    ckR = ck->r; ckG = ck->g; ckB = ck->b;

    /* Now find the i'th most different color */

    if (i&1 || i<10) {
      /* select the unused color that has the greatest mindist */
      entry = -1;  mdist = -1;
      for (j=0, cj=c; j<*pnumcols; j++,cj++) {
	      if (cj->use) {      /* this color has not been marked already */
	        /* update mindist */
                d = (cj->r - ckR)*(cj->r - ckR) + 
	            (cj->g - ckG)*(cj->g - ckG) + 
	            (cj->b - ckB)*(cj->b - ckB);
                if (cj->mindist > d) cj->mindist = d;
	        if (cj->mindist > mdist) { mdist = cj->mindist;  entry = j; }
	      }
      }
    }
    else {
      /* select the unused color that has the greatest usage */
      entry = -1;  mdist = -1;
      for (j=0, cj=c; j<*pnumcols; j++,cj++) {
	      if (cj->use) {  /* this color has not been marked already */
	        /* update mindist */
                d = (cj->r - ckR)*(cj->r - ckR) + 
	            (cj->g - ckG)*(cj->g - ckG) + 
    	            (cj->b - ckB)*(cj->b - ckB);
                if (cj->mindist > d) cj->mindist = d;
	        if (cj->use > mdist) { mdist = cj->use;  entry = j; }
	      }
      }
    }


    /* c[entry] is the next color to put in the map.  do so */
    xvbcopy((char *) &c[entry], (char *) &c1[i], sizeof(CMAPENT));
    c[entry].use = 0;
  }
  

  for (i=0; i<*pnumcols; i++) order[i] = (byte) c1[i].oldindex;
}

/////////////////////////////////////////////////////////////////////////

BOOL CDibDoc::OnSaveDocument(LPCTSTR lpszPathName)
{
  strcpy(IOfile::filename, lpszPathName);
  char* suffix = strrchr(IOfile::filename,'.');
  // KNOWN BUG: for some extensions, ::GetSaveFileName
  // doesn't recognize some extensions!!!.
  // At this moment, I must check the bug and 
  // I desestimate added extension
  if (suffix != NULL && *suffix=='.' && *(suffix-4)=='.') {
    *suffix='\0';
    SetPathName( (LPCTSTR) IOfile::filename, FALSE );
  }
  BOOL retval = !IOfile::SavePic(REWRITEPIC);
  SetModifiedFlag(FALSE);
  return retval; 
}

BOOL CDibDoc::OnOpenDocument(LPCTSTR lpszPathName)
{
  strcpy(IOfile::filename, lpszPathName);
  m_redraw=TRUE;
  return (BOOL) !IOfile::OpenPic(RELOAD);
}
