// dibdoc.h : interface of the CDibDoc class
//

/* Proyecto Fin de Carrera
 * Departamento de Informatica e Ingeniera de Sistemas
 * Universidad de Zaragora
 *
 * rea:      Robtica y Automatizacin Avanzada.
 * Tema:      Visin por Computador.
 * Ttulo:    Diseo de un entorno de visin para 
 *            el tratamiento de imgenes.
 * Director:  Dr. Carlos Sags Blzquiz.
 * Autor:     Joaqun Mur Amada.
 * Promocin: 1991-1996.
*/

#ifndef _DIBDOC_H
#define _DIBDOC_H

#include "dibapi.h"
#include "..\ImgBasic\ImgBasic.h"

class CDibView;

class CDibDoc : public CDocument, virtual public IOfile
{
friend CDibView;

protected:   //DibDoc specific variables
	HPALETTE m_hPALETTE;   // Palette of m_hBITMAP
  HBITMAP  m_hBITMAP;    // Device Dependent BitMap
  u_int    NumCols;      // Number of colors in palette

public:
	void OnProperties();
  static int CDibDoc::UsePalette; 
  // Nonzero if we must use a palette for displaying bitmaps
  //Note: UsePalette is setted once (when the splash windows is showed)
  int m_redraw;       // True if the bitmap must be redrawn

protected: // create from serialization only
	CDibDoc() : m_hPALETTE(NULL),m_hBITMAP(NULL), NumCols(0),m_redraw(TRUE) {}
           // Constructor
	DECLARE_DYNCREATE(CDibDoc)
  virtual ~CDibDoc()     // Destructor
       { SetHBITMAP(NULL); SetHPALETTE(NULL); }

// Palette stuff
public:
  HPALETTE GetHPALETTE();
  HBITMAP GetHBITMAP(CDC* pDC = NULL);
  HPALETTE CreateHPALETTE();
  inline void SetHPALETTE( HPALETTE hPALETTE );

  u_int NumColors()const { return NumCols; }
  u_int NumColors (HPALETTE hPALETTE);

	CSize GetSize()
		{ return CSize(IOfile::pWIDE,IOfile::pHIGH); }
  u_int GetHeight() const { return IOfile::pHIGH; }
	u_int GetWidth() const { return IOfile::pWIDE; }
	
// Image stuff
public:
  void RedrawPic  (void); 
	void InitDibData(BOOL bSetModifiedFlag = FALSE, CDC* pDC  = NULL);
  int SetIMAGE(HBITMAP hBITMAP); 
  inline void SetHBITMAP( HBITMAP hBITMAP );
  void FreeDocData() 
     { SetHPALETTE(NULL); SetHBITMAP(NULL);  }
  HDIB HDIBfromBITMAP ( HBITMAP hBITMAP, CDC* pDC=NULL );
  HBITMAP BITMAPfromHDIB(HDIB hDIB, CDC* pDC=NULL);

// Cursor handler functions
public:
  virtual void WaitCursor (void) ;
  virtual void ArrowCursor (void) ;

// Error handler functions
public:
  virtual void Warning ( LPCTSTR strMsg );
  virtual void Error ( LPCTSTR strMsg );
  virtual void FatalError( LPCTSTR strMsg );

  // Needed to actually override IOfile Error function
  virtual void Error      ( char* strMsg ) { Error ( (LPCTSTR) strMsg ); };
  virtual void Warning  ( char* strMsg = NULL)  { Warning(( LPCTSTR ) strMsg);};
  virtual void FatalError (char* strMsg) { FatalError ( (LPCTSTR) strMsg ); };
  virtual int  printf(const char *format, ...);
  virtual int  scanf (const char *format, ...);


//Dialog user interface overridabled functions
public:
      virtual void ChangeImageFormat(int type);

/**************************** INTERWIN.CPP ***************************/
/* This functions are pure virtual in IOFILE, I mean to say, they are only 
 * declarared, and implementation is giving in this class to provide
 * graphic user interface.
 */
public:   
     virtual int  PopUp          (char *, char **, int);
     virtual int  InputDialog    (char *, char *, int,char *, int);

     virtual void OpenAlert      (char *);
     virtual void CloseAlert     (void);

     virtual void JPEGDialog     (void);
     virtual void TIFFDialog     (void);

     virtual void drawStrings    (void);
     virtual int  PopMenu        (char *txt, char **labels, int n);

     virtual void ProgressMeter  (int min, int max, int val, char *str = NULL);
     virtual void TextView       (char *str);

     virtual char *GetFileName   (int bOpenFileDialog = TRUE) ;
     // bOpenFileDialog: TRUE for FileOpen, FALSE for FileSaveAs

/**************************** YESNO.CPP ***************************/
/* This functions are pure virtual in IOFILE, I mean to say, they are only 
 * declarared, and implementation is giving in this class to provide
 * graphic user interface.
 */
public:
     virtual int YesNo(char *str);
     virtual int YesNoCancel(char *str);
     virtual int OkCancel(char *str);
     virtual int RetryCancel(char *str);
     virtual int AbortRetryIgnore(char *str);

// Implementation
protected:
	BOOL OnSaveDocument(LPCTSTR lpszPathName);
  BOOL OnOpenDocument(LPCTSTR lpszPathName);
  
protected:

#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif



/**************************** PIC2BMP.C ***************************/
/* This built-in functions exports/imports images to/from memory PIC
 * and Windows Bitmap formats.
 * (Usefull for display/retrieve images in Microsoft Windows Programs).
 */
public:
    int  SetIMAGE (HDIB hDIB);
    HDIB GetHDIB  (byte *pic824, int ptype, u_int w, u_int h,
                   byte *rmap, byte *gmap,byte *bmap,
                   u_int numcols);
    HDIB GetHDIB  (); 
      
// Generated message map functions
protected:
	//{{AFX_MSG(CDibDoc)
	afx_msg void OnEditCopy();
	afx_msg void OnEditCut();
	afx_msg void OnEditPaste();
	afx_msg void OnEditPasteDdb();
	afx_msg void OnEditPasteDib();
  afx_msg void OnEditUndo() { RestorePic(); }
	afx_msg void OnSettings();
	afx_msg void OnSaveMemory();
	afx_msg void OnUpdateSaveMemory(CCmdUI* pCmdUI);
	afx_msg void OnUpdateEditUndo(CCmdUI* pCmdUI);
	afx_msg void OnUpdateEditPaste(CCmdUI* pCmdUI);
	afx_msg void OnUpdateEditCut(CCmdUI* pCmdUI);
	afx_msg void OnUpdateEditCopy(CCmdUI* pCmdUI);
	afx_msg void OnFormatBwdither() {ChangeImageFormat(F_BWDITHER);};
	afx_msg void OnFormatGreyscale(){ChangeImageFormat(F_GREYSCALE);};
	afx_msg void OnFormatReduced() {ChangeImageFormat(F_REDUCED);};
	afx_msg void OnFormatFullcolor(){ChangeImageFormat(F_FULLCOLOR);};
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

inline void CDibDoc::SetHPALETTE( HPALETTE hPALETTE )
{ 
    TRACE("\nCDibDoc::SetHPALETTE\n");
    if ( m_hPALETTE!= NULL && !DeleteObject(m_hPALETTE)){
        Error("CDibDoc::SetHPALETTE.\nPaleta anterior en uso.");
        if ( hPALETTE!= NULL ) DeleteObject( hPALETTE );
        return;
    }
    m_hPALETTE = hPALETTE;
}

//---------------------------------------------------------------
// CDibView : : S e t H B I T M A P 
// Copy bitmap handle to member bitmap if exist.
inline void CDibDoc::SetHBITMAP( HBITMAP hBITMAP )
{
    TRACE("\nCDibView::SetHBITMAP\n");    
    if ( m_hBITMAP != NULL && !DeleteObject(m_hBITMAP)) {
        Error("CDibView::SetHBITMAP\nBitmap antiguo en uso");
        if ( hBITMAP!= NULL ) DeleteObject( hBITMAP );
        return;
    }
    m_hBITMAP = hBITMAP;
}

//---------------------------------------------------------------
// CDibDoc : : G e t  H P A L E T T E
inline HPALETTE CDibDoc::GetHPALETTE()
{ 
    if ( m_hPALETTE == NULL )
      CreateHPALETTE();
    return m_hPALETTE; 
}



#endif  //#define _DIBDOC_H
