// dibview.h : interface of the CDibView class

#ifndef _DIBVIEW_H
#define _DIBVIEW_H

#include "DibDoc.h"

class CDibView : public CScrollView
{
protected: // create from serialization only
	CDibView() : m_ScrollSize(0,0),m_Scale(100,100),m_autoadjust(FALSE) {};
	DECLARE_DYNCREATE(CDibView)
  ~CDibView();

// Internal data    
private:
    CSize m_ScrollSize; // Size of the region to paint the bitmap
    CSize m_Scale;      // Scale : 100 
    BOOL  m_autoadjust; // True if autoadjust ZOOM is set.

// Implementation
protected:

// Attributes
public:
#ifdef _DEBUG	
    CDibDoc* GetDocument()
		{
		  //ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CDibDoc)));
		  return (CDibDoc*) m_pDocument;
		}
#else
    CDibDoc* GetDocument() const {return (CDibDoc*) m_pDocument;	};
#endif
    	
    CSize GetSize() const
    {   return m_ScrollSize; }
    
    HBITMAP GetHBITMAP(CDC* pDC = NULL) 
    {return GetDocument()->GetHBITMAP(pDC);}
        
    void SetHBITMAP( HBITMAP hBITMAP )
    {  GetDocument()->SetHBITMAP( hBITMAP ); }
    
    HPALETTE GetHPALETTE()
        { return ( GetDocument()->GetHPALETTE() ); }
    
    void SetHPALETTE( HPALETTE hPALETTE )
        { GetDocument()->SetHPALETTE( hPALETTE ); }

    void InitViewData( void );
    void FreeViewData( void );


// Implementation
public:
	void OnDraw(CDC* pDC);      // new version of OnDraw that dispays HDIB
                              // using a context bitmap   
                                    
	virtual void OnInitialUpdate(void);
	virtual void OnActivateView(BOOL bActivate, CView* pActivateView,
				                      CView* pDeactiveView);
  virtual void OnQueryNewPalette() { OnInitialUpdate(); return;}
  virtual void Warning (LPCTSTR strMsg );
  virtual void Error(LPCTSTR strMsg);
  virtual void FatalError(LPCTSTR strMsg);

// Operations
public:        
 void OnZoom(u_int factor); 

// Printing support
protected:
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
  virtual afx_msg void OnInitMenu( CMenu* pMenu );

// Create a device dependent bitmap from the DIB  and Reverse
public:
    void InitDibData( BOOL bSetModifiedFlag = FALSE, CDC* pDC=NULL);

    virtual void SetScrollSizes( int nMapMode,
                                 SIZE sizeTotal,
                                 const SIZE& sizePage = sizeDefault,
                                 const SIZE& sizeLine = sizeDefault );
    int NumColors()
      { return GetDocument()->NumColors(); };


// Generated message map functions
protected:
	//{{AFX_MSG(CDibView)
	afx_msg LRESULT OnDoRealize(WPARAM wParam, LPARAM lParam);  // user message
	afx_msg void OnImageInformation();
	afx_msg void OnZoom25()  {OnZoom(25);}
  afx_msg void OnZoom50()  {OnZoom(50);}
	afx_msg void OnZoom100();
	afx_msg void OnZoom200() {OnZoom(200);}
	afx_msg void OnZoom400() { OnZoom(400); }
	afx_msg void OnZoomSizetoFit();
	afx_msg LRESULT OnCommandHelp(WPARAM wParam, LPARAM lParam);
	afx_msg LRESULT OnHelpHitTest(WPARAM wParam, LPARAM lParam);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

#endif //#define _DIBVIEW_H
