/*********************************************************************************\
*     File:                 I N T E R W I N  . C P P                              *
*                                                                                 *
*     IOfile routines for display messages (character output for compatibility)*
*     Standard   C++                                                              *
\*********************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <afxwin.h>
#include "..\res\resource.h"
#include "dibdoc.h"
#include "askint.h"
#include "MainFrm.h"
#include "VisioApp.h"
#include "Multifil.h"

/***************************************************/
//------------------------------------------------------------------------------
// Displays message info.
//------------------------------------------------------------------------------
static char *fname[NISTR] = { "Informacion: ", "Ultimo aviso: ", "Nombre: ",
    "Formato: ", "Resolucion: ","Recortado: ","Expandido: ","Seleccionado: ",
    "Color: "};

void CDibDoc::drawStrings()
{
  int i;
  char auxstr[NISTR*ISTRLEN], auxs[ISTRLEN+20];  
  auxstr[0]='\0';
  // Windows based equivalent funciton:
  for (i=0; i<NISTR; i++) {
        sprintf(auxs,"%s %s\n",fname[i],GetISTR(i));
        strcat(auxstr,auxs);
  }
  //Get the first view in the list of views:
  POSITION pos = GetFirstViewPosition();
  CView* pFirstView = GetNextView( pos );
  //Show the message attached to the first view of the document
  pFirstView->MessageBox( (LPCTSTR)auxstr,(LPCTSTR)"CDibDoc::drawStrings",
        MB_ICONEXCLAMATION  | MB_OK ); 
}

/**************************************************/

int CDibDoc::PopMenu(char *txt, char **labels, int n)    
{
  char buffer[512];
  int i,j;
  char *begin;
  i=sprintf(buffer,txt);
  if (n) {
      i+=sprintf(buffer+i,"\n");
      for (j=0;j<n;j++) {
        for (begin=labels[j]; *begin <32 && *begin;begin++);            
        i+=sprintf(buffer+i,"%i) %s\n",j,begin);
      }
      i+=sprintf(buffer+i,"\nIntroduce una opcion:");
      CAskInt Dlg(buffer,"0");
      do {
           Dlg.DoModal();
           i=atoi(Dlg.m_numero);
      }while ((i<0) || (i>n));
  }
  return i;
}

/***************************************************/

int CDibDoc::PopUp(char *txt, char *labels[], int n)
{
     return PopMenu( txt,  labels, n);
}


/**************************************************/

int CDibDoc::InputDialog(char *txt, char *buf,
                           int buflen, char *filstr, int allow)
{
    /* pops up a window with a prompt string, a 1-line editable
     text thingy, and a row of buttons.  'txt' is the prompt
     string, 'buf' is the buffer displayed and edited
     in the window, buflen is its length, filstr is a filter string, of
     characters to block from entry (or, if 'allow' is '1', a list
     of the *only* characters allowed for entry)

     Also note that the filter string should only contain normal printable
     characters (' ' through '\177'), as ctrl chars are pre-filtered 
     (ie, interpreted as emacs-like commands) */

  int i;
  
  CAskInt Dlg(txt,buf);
  
	// Llamamos al dilogo modal.
	i=Dlg.DoModal();

	// Si salimos con OK, tomamos los datos del dilogo.
	if (i==IDOK)
	{
      strncpy(buf,Dlg.m_numero,buflen);
      //FilterString( buf, filstr, allow);
      i=atoi(buf);
  }
	return i;
}

static int IsWaitCursor=0;

void CDibDoc::WaitCursor     (void) 
{
  /*
  if (!IsWaitCursor) {
      BeginWaitCursor();
  }   // Call once BeginWaitCursor
  IsWaitCursor++;
  */
  BeginWaitCursor();
}

void CDibDoc::ArrowCursor     (void)
{
  /*
  if (IsWaitCursor==1) {
      EndWaitCursor();
  }   // Call once EndWaitCursor
  else
    IsWaitCursor--;
  */
  EndWaitCursor();
} 

/***************************************************/

void CDibDoc::JPEGDialog(void)
{
	char buffer[20];
  int  porcentaje;
  itoa(GetJPEGquality(),buffer,10);
  CAskInt Dlg("Calidad en % para guardar la imagen en formato JPEG:",
              buffer); 
  do {
	    // Llamamos al dilogo modal.
	    Dlg.DoModal();
      porcentaje=atoi(Dlg.m_numero);
  } while ((porcentaje > 100) || (porcentaje <0));
  SetJPEGquality(porcentaje);
  Dlg.m_texto = _T("Suavizado en % para guardar la imagen en formato JPEG:");
  itoa(GetJPEGsmooth(),buffer,10);
  Dlg.m_numero=_T(buffer);
  do {
	    // Llamamos al dilogo modal.
	    Dlg.DoModal();
      porcentaje=atoi(Dlg.m_numero);
  } while ((porcentaje > 100) || (porcentaje <0));
  SetJPEGsmooth(porcentaje);
}

/**************************************************/  
static char * tiffBut[] = { "Ninguno",
   "LZW",
   "PackBits (Machintosh RLE)", 
   "CCITT 3",
   "CCITT 4","NEXT 2-bit RLE",
   "ThunderScan 4 bit RLE",
   "PIXARFILM"
   #ifdef ALLOW_JPEG
   "JPEG",
   #endif
   };

void CDibDoc::TIFFDialog(void)
{
  /* returns the kind of compression scheme used in tiffwrite algorithm */
  int comp;
  comp = PopMenu("Tipo de algoritmo de compresin para guardar la imagen TIFF",
          tiffBut,sizeof(tiffBut)/sizeof(tiffBut[0]));
  switch (comp) {
  case 0: SetTIFFcompression(COMPRESSION_NONE);      break;
  case 1: SetTIFFcompression(COMPRESSION_LZW);       break;
  case 2: SetTIFFcompression(COMPRESSION_PACKBITS);  break;
  case 3: SetTIFFcompression(COMPRESSION_CCITTFAX3); break;
  case 4: SetTIFFcompression(COMPRESSION_CCITTFAX4); break;
  case 5: SetTIFFcompression(COMPRESSION_NEXT);      break;
  case 6: SetTIFFcompression(COMPRESSION_CCITTRLEW); break;
  case 7: SetTIFFcompression(COMPRESSION_THUNDERSCAN); break;
  case 8: SetTIFFcompression(COMPRESSION_PIXARFILM); break;
  #ifdef ALLOW_JPEG
  case 9: SetTIFFcompression(COMPRESSION_JPEG);      break;
  #endif
  }
}


/**************************************************/  
//---------------------------------------------------------------
// M E S S A G E   H A N D L E R S
void CDibDoc::Warning (LPCTSTR strMsg )
{ 
  char cadena[60];
  sprintf(cadena,"Aviso en %s", GetTitle()); 
   //To get the first view in the list of views:
  POSITION pos = GetFirstViewPosition();
  CView* pFirstView = GetNextView( pos );
  //Show the message attached to the first view of the document
  pFirstView->MessageBox( 
    (strMsg != NULL) ? (LPCTSTR)strMsg :  (LPCTSTR)GetISTR(ISTR_WARNING), 
    (LPCTSTR)cadena, 
    MB_ICONEXCLAMATION  | MB_OK ); 
}

void CDibDoc::Error(LPCTSTR strMsg )
{ 
  char cadena[60];
  sprintf(cadena,"Error en %s", GetTitle()); 
   //To get the first view in the list of views:
  POSITION pos = GetFirstViewPosition();
  CView* pFirstView = GetNextView( pos );
  //Show the message attached to the first view of the document
  pFirstView->MessageBox((LPCTSTR)strMsg, (LPCTSTR)cadena,
     MB_ICONEXCLAMATION  | MB_OK ); 
}

void CDibDoc::FatalError(LPCTSTR strMsg )
{
  char cadena[60];
  sprintf(cadena,"Error Fatal!!! Imagen: %s, clase CDibDoc", GetTitle());
  //To get the first view in the list of views:
  POSITION pos = GetFirstViewPosition();
  CView* pFirstView = GetNextView( pos );
  //Show the message attached to the first view of the document
  pFirstView->MessageBox( (LPCTSTR)strMsg, (LPCTSTR)cadena, MB_ICONSTOP | MB_OK ); 
  exit(-2);
}


/***************************************************/
void CDibDoc::ProgressMeter(int min, int max, int val, char *str /* ==NULL*/)
{
  /* called during 'long' operations (algorithms, smoothing, etc.) to
     give some indication that the program will ever finish.  Draws a 
     progress gauge in either mainW (if not useRoot), or ctrlW (Updates
     the progress meter each second).
     Tries to be clever:  only draws gauge if it looks like the operation is
     going to take more than a few seconds.  Calling with val == max removes
     the gauge. */

  static int    waiting  = 0;
  static time_t lastTime = 0;
  time_t        nowTime;

  if (min >= max) return;
  if (val>=max && lastTime==0) return;  /* already erased meter */


  time(&nowTime);

  if (lastTime == 0) {   /* first time we're calling for this meter */
    lastTime = nowTime;
    waiting = 1;                           
  }
  else if (waiting) {    /* haven't drawn meter yet... */
    double doneness = ((double) (val-min)) / ((double) (max - min));
    if ((nowTime - lastTime >= 2) && doneness < .5)
      waiting = 0;
  }

  if (!waiting) {     /* not waiting (or not waiting any longer) */
    if (nowTime == lastTime && val<max) return;  /* max one draw per second */
    lastTime = nowTime;
    CMainFrame* pFrame = (CMainFrame*) AfxGetApp()->m_pMainWnd;
    ASSERT_KINDOF(CMainFrame, pFrame);
    if (val >= max) {            /* remove temp gauge */
      pFrame->m_wndStatusBar.HideProgress();
      /* Show Clock */
  	  CTime CurTime(time(NULL));
	    pFrame->m_wndStatusBar.SetPaneText(4,CurTime.Format(" %I:%M %p"));
      lastTime = 0;
    } else {                     /* draw temp gauge */
      pFrame->m_wndStatusBar.ProgressMeter(min,max,val,str);
      //pFrame->m_wndStatusBar.ShowProgress();
    }
  }
  
  
}

/**********************************************************************/  
class CAlertDlg : public CDialog
{
// Construction
public:
	CAlertDlg (LPSTR texto,CWnd* pParent = NULL);   // standard constructor

// Dialog Data
	//{{AFX_DATA(CAlertDlg)
	enum { IDD = IDD_ALERTDLG };
	CString	m_texto;
	//}}AFX_DATA


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAlertDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:

	// Generated message map functions
	//{{AFX_MSG(CAlertDlg)
	virtual void OnCancel();
	virtual void OnOK();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};
/**********************************************************************/  

CAlertDlg::CAlertDlg(LPSTR texto /*=NULL*/, CWnd* pParent /*=NULL*/)
	: CDialog(CAlertDlg::IDD, pParent), m_texto(_T(texto))
{
	//{{AFX_DATA_INIT(CAlertDlg)
	//}}AFX_DATA_INIT
}

void CAlertDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAlertDlg)
  DDX_Text(pDX, IDC_TEXTO, m_texto);
	//}}AFX_DATA_MAP
}

void CAlertDlg::OnCancel() // not really a message handler
{
    // Do not call base class OnCancel
    UpdateData(TRUE);
}

void CAlertDlg::OnOK() // not really a message handler
{
    // Do not call base class OnOK
    UpdateData(TRUE);
}

BEGIN_MESSAGE_MAP(CAlertDlg, CDialog)
	//{{AFX_MSG_MAP(CAlertDlg)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()



/**********************************************************************/  

static CAlertDlg* m_pAlertDlg = NULL;

/**********************************************************************/  
   
void CDibDoc::OpenAlert (char *str)
{
  //To get the first view in the list of views:
  POSITION pos = GetFirstViewPosition();
  CView* pFirstView = GetNextView( pos );
  /*Show the message attached to the first view of the document
  pFirstView->MessageBox( (LPCTSTR)str,(LPCTSTR) "CDibDoc::OpenAlert", 
    MB_ICONEXCLAMATION | MB_OK ); 
  // creates the dialog if not already created */
  
  // Constructor (move to Main Frame constructror)
  if (m_pAlertDlg == NULL)
    m_pAlertDlg = new CAlertDlg(str, pFirstView);

  m_pAlertDlg->m_texto = str;
  // NEW
  if (m_pAlertDlg->GetSafeHwnd() == 0) {
    m_pAlertDlg->Create(CAlertDlg::IDD); // displays the dialog window
    }
}

void CDibDoc::CloseAlert (void)
{
  /*
  //To get the first view in the list of views:
  POSITION pos = GetFirstViewPosition();
  CView* pFirstView = GetNextView( pos );
  //Show the message attached to the first view of the document
  pFirstView->MessageBox( (LPCTSTR) "Fin de la alerta",
    (LPCTSTR) "CDibDoc::CloseAlert",
    MB_ICONEXCLAMATION | MB_OK ); 
  */
  m_pAlertDlg->DestroyWindow();
  // no problem if window was already destroyed
  delete m_pAlertDlg; // destroys window if not already destroyed
  m_pAlertDlg = NULL;
}

/**********************************************************************/  

void CDibDoc::TextView       (char *str) 
{
    //Open file in text view
    ((CVisioApp*)AfxGetApp())->m_pTemplateTexto->OpenDocumentFile(str);
}

/**********************************************************************/  
/* Return a filename from file dialogbox*/

char *CDibDoc::GetFileName(int bOpenFileDialog) 
// bOpenFileDialog: TRUE for FileOpen, FALSE for FileSaveAs
{
  static char FileName[MAXPATHLEN];
  CVisioApp* pApp = (CVisioApp*) AfxGetApp();

  // prompt the user (with all document templates)
	CMultiFileDialog dlgFile( bOpenFileDialog );
	CString title,strFilter,strDefault;
  VERIFY(title.LoadString((bOpenFileDialog) ? AFX_IDS_OPENFILE : AFX_IDS_SAVEFILE ));
	
  if (bOpenFileDialog ) {
	  dlgFile.m_ofn.Flags |= OFN_HIDEREADONLY | OFN_FILEMUSTEXIST ;
    //dlgFile.m_ofn.Flags &= ~OFN_SHOWHELP;
  }

  dlgFile.m_ofn.nFilterIndex = 1;
	// strDefExt is necessary to hold onto the memory from GetExtFromType
	dlgFile.m_ofn.lpstrDefExt = NULL;
		
	// do for all doc template
	AppendFilterSuffix (strFilter, dlgFile.m_ofn, pApp->m_pTemplateIOfile,
		&strDefault);

	// append the "*.*" all files filter
	CString allFilter;
	VERIFY(allFilter.LoadString(AFX_IDS_ALLFILTER));
	strFilter += allFilter;
	strFilter += (TCHAR)'\0';   // next string please
	strFilter += _T("*.*");
	strFilter += (TCHAR)'\0';   // last string
	dlgFile.m_ofn.nMaxCustFilter++;

	dlgFile.m_ofn.lpstrFilter = strFilter;
	dlgFile.m_ofn.lpstrTitle = title;
	dlgFile.m_ofn.hwndOwner = AfxGetMainWnd()->GetSafeHwnd();

  CString strPath;

	if (dlgFile.DoModal() == IDOK)
	{
		POSITION pos = dlgFile.GetStartPosition();
		if (pos != NULL)
		{
			strPath = dlgFile.GetNextPathName(pos);
			if (strPath.Find(":\\\\") == 1 && strPath.GetLength() > 4)
			{
				// this means we have an invalid path that looks like this:
				// C:\\foo.bmp
				// We need to cut out the extra slash
				CString temp;
				temp = strPath.Left(3);
				temp += strPath.Mid(4);
				strPath = temp;
			}
		}
  }
  strncpy(FileName,(LPCTSTR) strPath,MAXPATHLEN); 
	// the default PostNcDestroy handler will delete this object when destroyed
  return FileName;
}
/**********************************************************************/  
