#include "stdafx.h"
#include "..\res\resource.h"
#include "Thread.h"
#include "Registry.h"
#include "MainFrm.h"
#include "VisioApp.h"



static CEvent* m_pTermThreadsEvent = NULL;
static THREADINFO ThreadInfo;

/////////////////////////////////////////////////////////////////////////////
// CLaunchDlg message handlers

void CMainFrame::OnCapture() 
{
	// TODO: Add your command handler code here

#ifdef _DEBUG
  CString str;
  str.Format("Programa que captura las imgenes:%s\n"
    "Fichero donde se guarda la imagen capturada: %s\n"
    "El programa pide confirmacin para cargar el fichero: %s",
    (LPCTSTR) CCaptureDlg::strCapturador, (LPCTSTR) CCaptureDlg::strFichero,
    (CCaptureDlg::bPreguntar) ? "S" : "No");;
	MessageBox((LPCTSTR)str);
#endif

  // Don't launch capture twice at the same time
  if (m_pTermThreadsEvent != NULL) return;

	// Create event object. This event object can be used to terminate 
	// all monitoring threads at any time. It is signaled when the main
	// application is about to terminate, thus preventing any memory
	// leaks with the allocated CWinThread objects.
	m_pTermThreadsEvent = new CEvent(FALSE, TRUE);
	ASSERT_VALID(m_pTermThreadsEvent);

  // Set up setting if no program is given
  if (CCaptureDlg::strCapturador.IsEmpty())
    OnCaptureSettings();

	// If select file, allocate memory to hold path name. This is very
	// important since the thread function will access this address
	// after this function has terminated. It is the responsibility of
	// the thread function to free the memory for the path name.
	if ( !CCaptureDlg::strCapturador.IsEmpty())	
	{
		// Allocate structure that will pass relevant information
		// to the thread function. The thread function is responsible
		// for freeing the structure.
		// fill in members of structure
		ThreadInfo.pTermThreadEvent = m_pTermThreadsEvent;
		ThreadInfo.hThread = NULL;
    ThreadInfo.hWnd = GetSafeHwnd();
		ThreadInfo.strPathName = CCaptureDlg::strCapturador;

		// Create CWinThread object. The thread function will 
		// retrieve and store the handle to this thread in the 
		// list box. Then we can make sure the thread is 
		// terminated before the main application is terminated.
		AfxBeginThread(LaunchAndWait, & ThreadInfo);
	}
}


void CMainFrame::OnDestroy() 
{

	// TODO: Add your message handler code here
	::WinHelp(GetSafeHwnd(),"HLP.HLP",HELP_QUIT,0);
  //Eliminar el temporizador PROG_TIMER (igual a 1, ver Mainfrm.cpp)
	KillTimer( PROG_TIMER );

  if (m_pTermThreadsEvent != NULL ) {
	  // Terminate all the active threads. This causes the threads that 
	  // are still monitoring launched application to terminate because 
	  // they terminate when the launched application terminates OR 
	  // when this event becomes signaled. 
	  m_pTermThreadsEvent->SetEvent();

	  // We signaled the event, which terminates all active monitoring
	  // threads. Now sit back and wait for all of the threads to 
	  // terminate. This allows all of the allocated CWinThread objects
	  // to terminate and clean up properly. Just deleting the 
	  // CWinThread objects will cause memory leaks.
	  WaitForSingleObject( ThreadInfo.hThread,  INFINITE);

	  // delete the CEvent object that was allocated eariler
    if (m_pTermThreadsEvent != NULL) {
	    delete m_pTermThreadsEvent;
      m_pTermThreadsEvent = NULL;
    }
    // mark the finished thread
    if (ThreadInfo.hThread != NULL)
      ThreadInfo.hThread = NULL;
  }
	CMDIFrameWnd::OnDestroy();
}

void CMainFrame::OnCaptureSettings() 
{
	// TODO: Add your command handler code here
	CCaptureDlg dlg(this);
  dlg.DoModal();
}

LRESULT CMainFrame::OnEndCapture(WPARAM wParam, LPARAM lparam)
// Carga la imagen capturada, despus de que el 
// programa de captura haya terminado
{
  // mark the capture finished
  if (m_pTermThreadsEvent != NULL) {
	  delete m_pTermThreadsEvent;
    m_pTermThreadsEvent = NULL;
  }
  // mark the finished thread
  if (ThreadInfo.hThread != NULL)
    ThreadInfo.hThread = NULL;

  // Carga la imagen capturada
  if (CCaptureDlg::bPreguntar) {
    if (!CCaptureDlg::strFichero.IsEmpty()) {
      SetCurrentDirectory(
         (LPCTSTR) GetDirectory(CCaptureDlg::strFichero) );
    }
    // Pregunta el fichero
    ((CVisioApp*)AfxGetApp())->OnFileOpen();
  }
  else
    // Toma el fichero por defecto
    ((CVisioApp*)AfxGetApp())->OpenDocumentFile(CCaptureDlg::strFichero);

  return 0;
}


UINT LaunchAndWait(LPVOID pParam)
{
	// assign address of parameter
	LPTHREADINFO pThreadInfo = (LPTHREADINFO)pParam;

	// process structure that will receive info on the new process
	PROCESS_INFORMATION	stProcessInfo;

	// launch application
	if (LaunchApplication(pThreadInfo->strPathName, &stProcessInfo))
	{
		// Need to store thread to list to prevent memory leaks with
		// CWinThread object. If our main app is requested to terminate 
		// before this thread terminates, then we need to terminate 
		// this thread first.
    if ( pThreadInfo->hThread == NULL) {
        pThreadInfo->hThread = AfxGetThread()->m_hThread;
    }

		// Wait for either the process to terminate or the event
		// to be signaled. If the event is signaled, then that indicates
		// the main application is terminating and all open threads need
		// to terminate also.
		HANDLE hThreads[2];
		hThreads[0] = pThreadInfo->pTermThreadEvent->m_hObject;
		hThreads[1] = stProcessInfo.hProcess;

		// wait for application to terminate or event to be signaled
		DWORD dwIndex = WaitForMultipleObjects(
					2, hThreads, FALSE, INFINITE); 

    // Carga la imagen nicamente si la aplicacin termin,
    // no elimina de la lista si se sealiz un evento
    // (me comunico con la hebra principal a travs de un
    // mensaje de Windows)
    if (dwIndex) {
      ::PostMessage((HWND) pThreadInfo->hWnd, WM_ENDCAPTURE,0,0);
    }

		// close handles to process and its main thread
		CloseHandle(stProcessInfo.hThread);
		CloseHandle(stProcessInfo.hProcess);
	}
	return 0;
}


BOOL LaunchApplication(LPCTSTR pCmdLine, PROCESS_INFORMATION* pProcessInfo)
{
	// startup structure for new process
	STARTUPINFO	stStartUpInfo;

	// since a lot of members are reserved or ignored, we can just set
	// all members to 0 and then reset the members we care about
	memset(&stStartUpInfo,0,sizeof(STARTUPINFO));

	stStartUpInfo.cb = sizeof(STARTUPINFO); 	// size of structure
	stStartUpInfo.dwFlags = STARTF_USESHOWWINDOW;	// use wShowWindow member
	stStartUpInfo.wShowWindow = SW_SHOWDEFAULT;	// default value

	// create new process, process creates primary 
	// thread and launches application
	return CreateProcess(NULL, (LPTSTR)pCmdLine, NULL, NULL, FALSE,
			HIGH_PRIORITY_CLASS, NULL, 
			(LPCTSTR) GetDirectory (pCmdLine), &stStartUpInfo, pProcessInfo);
}


char* GetDirectory (const char* filename)
{
  static char directory[128];
  char *backslash;
  backslash = (char *) strrchr(filename, '\\');
  if (!backslash) {  /* add a '.' at the end of the name */
    return NULL;
  }
  // copy directory name in directory
  int len = backslash-filename;
  strncpy(directory,filename,len);
  // return volatile string (next funcion call 
  // directory may change).
  // put and end of string instead last '\'
  directory[len++]='\0';
  return directory;
}
